--- constants ---
-- PAL resolution
NES_WIDTH = 256
NES_HEIGHT = 240

-- to reduce the amount of decimals used, I decided to put player/object coordinates in terms of subpixels
PX = 16
SSPX = 1 / PX       -- size of a sub-subpixel compared to our coordinate units
SSSPX = SSPX * SSPX -- size of a sub-subsubpixel compared to our coordinate units
BLOCK = PX * PX

--- rendering ---

-- called on load and on every frame.
-- sets the game resolution
function setup_game_size()
    -- coop's screen size
    screenWidth = djui_hud_get_screen_width()
    screenHeight = djui_hud_get_screen_height()

    -- make it a bit larger, since the screenWidth is a bit lower than expected
    screenWidth = screenWidth + (screenWidth * 5 // screenHeight)

    -- size of the actual render area (always 16:15)
    gameWidth = math.min(screenWidth, screenHeight * 16 / 15)
    gameHeight = math.min(screenHeight, screenWidth * 15 / 16)

    -- coordinates of the screen edges
    leftEdge = (screenWidth - gameWidth) / 2
    rightEdge = (screenWidth + gameWidth) / 2
    topEdge = (screenHeight - gameHeight) / 2
    bottomEdge = (screenHeight + gameHeight) / 2

    -- how big the game is compared to the NES resolution
    gameScaleX = gameWidth / NES_WIDTH
    gameScaleY = gameHeight / NES_HEIGHT
end

-- renders a specific texture using the nes resolution
function render_texture_nes(tex, x, y, scaleX, scaleY, fixed)
    -- enforce pixel-perfect
    x = math.floor(x)
    y = math.floor(y)

    -- adjust for camera
    if not fixed then
        local smb = gSMBStates[0]
        x = x - smb.cameraX
        y = y + smb.cameraY
    end

    -- don't render out of view
    if x < -tex.width * math.abs(scaleX) or x > NES_WIDTH or y < -tex.height * math.abs(scaleY) or y > NES_HEIGHT then
        return
    end

    -- allow flipping
    if scaleY < 0 then
        x = x - tex.width * scaleX * (tex.width / tex.height)
        y = y - tex.height * scaleY + (tex.height - tex.width)
    end

    local renderX = leftEdge + x * gameScaleX
    local renderY = topEdge + y * gameScaleY
    local renderScaleX = scaleX * gameScaleX
    local renderScaleY = scaleY * gameScaleY
    djui_hud_render_texture(tex, renderX, renderY, renderScaleX, renderScaleY)
end

-- renders a texture from a sheet using the nes resolution
-- it also considers the camera unless fixed is true
-- unlike the nes, this supports any-sized sprite instead of just 8x8 or 8x16
function render_texture_nes_sheet(tex, x, y, scaleX, scaleY, texWidth, texHeight, frame, spaceBetweenFrames, allowFlip, fixed, adjustRender)
    local sheetWidth = tex.width
    -- sheets that allow flipping are formatted to have the flipped version on the other half
    if allowFlip then
        sheetWidth = sheetWidth // 2
    end

    -- adjust for camera
    if not fixed then
        local smb = gSMBStates[0]
        x = x - smb.cameraX
        y = y + smb.cameraY
    end

    -- don't render out of view
    local testScaleX = scaleX
    if not adjustRender then
        testScaleX = testScaleX * (texWidth / texHeight)
    end
    if x < -texWidth * math.abs(testScaleX) or x > NES_WIDTH or y < -texHeight * math.abs(testScaleX) or y > NES_HEIGHT then
        return
    end

    -- get where on the sheet to render
    local framesPerLine = sheetWidth // (texWidth + spaceBetweenFrames)
    local sheetX = (texWidth + spaceBetweenFrames) * (frame % framesPerLine)
    local sheetY = (texHeight + spaceBetweenFrames) * (frame // framesPerLine)

    if scaleX < 0 then
        if scaleY < 0 then -- horizontal and vertical flip (same as rotating 180)
            x = x - texWidth * scaleX * (texWidth / texHeight)
            y = y - texHeight * scaleY + (texHeight - texWidth)
        elseif allowFlip then -- horizontal flip
            scaleX = -scaleX
            sheetX = sheetWidth * 2 - sheetX - texWidth
        end
    elseif scaleY < 0 and allowFlip then -- vertical flip (do horizontal flip AND rotate 180)
        scaleX = -scaleX
        sheetX = sheetWidth * 2 - sheetX - texWidth
        x = x - texWidth * scaleX * (texWidth / texHeight)
        y = y - texHeight * scaleY + (texHeight - texWidth)
    end
    -- mario's big sprites need these adjustment's fsr
    if adjustRender then
        scaleX = scaleX * (texHeight / texWidth)
        y = y - texHeight + texWidth
    end

    -- enforce pixel-perfect
    x = math.floor(x) + 0.5 -- seems to look better like this? idk
    y = math.floor(y)

    local renderScaleX = scaleX * gameScaleX
    local renderScaleY = scaleY * gameScaleY
    local renderX = leftEdge + x * gameScaleX
    local renderY = topEdge + y * gameScaleY

    -- fix some rendering issues???? I have no idea
    --[[if math.floor(gameScaleX) == gameScaleX then
        renderX = math.floor(renderX) - 2 / texWidth
        renderY = math.floor(renderY) - 2 / texHeight
    end]]

    djui_hud_render_texture_tile(tex, renderX, renderY, renderScaleX, renderScaleY, sheetX, sheetY, texWidth, texHeight)
end

-- renders a texture using the nes resolution, but which texture is based on the frame
-- this is used for the tileset to prevent tile bleeding issues (other sprites get around this by just having space around them)
-- it also considers the camera unless fixed is true
-- unlike the nes, this supports any-sized sprite instead of just 8x8 or 8x16
function render_texture_nes_collection(texName, x, y, scaleX, scaleY, frame, framesPerLine, allowFlip, fixed)
    local do180Flip = false
    if allowFlip then
        frame = frame % framesPerLine + truncate(frame, framesPerLine)
        if scaleX < 0 then
            if scaleY < 0 then -- horizontal and vertical flip (same as rotating 180)
                do180Flip = true
            elseif allowFlip then -- horizontal flip
                scaleX = -scaleX
                frame = frame + framesPerLine
            end
        elseif scaleY < 0 and allowFlip then -- vertical flip (do horizontal flip AND rotate 180)
            scaleX = -scaleX
            frame = frame + framesPerLine
            do180Flip = true
        end
    end

    local frameName = texName..string.format("%04d", frame)
    --if not mod_file_exists("textures/"..frameName..".png") then return end
    
    local tex = get_texture_info(frameName)

    -- enforce pixel-perfect
    x = math.floor(x)
    y = math.floor(y)

    -- adjust for camera
    if not fixed then
        local smb = gSMBStates[0]
        x = x - smb.cameraX
        y = y + smb.cameraY
    end

    -- don't render out of view
    if x < -tex.width * math.abs(scaleX) or x > NES_WIDTH or y < -tex.height * math.abs(scaleY) or y > NES_HEIGHT then
        return
    end

    -- allow flipping
    if do180Flip then
        x = x - tex.width * scaleX * (tex.width / tex.height)
        y = y - tex.height * scaleY + (tex.height - tex.width)
    end

    local renderX = leftEdge + x * gameScaleX
    local renderY = topEdge + y * gameScaleY
    local renderScaleX = scaleX * gameScaleX
    local renderScaleY = scaleY * gameScaleY
    djui_hud_render_texture(tex, renderX, renderY, renderScaleX, renderScaleY)
end

-- prints text using the NES resolution
local FONT_OVERRIDE = {
    ["-"] = 36,
    ["@"] = 37, -- x uses for lives count
    ["!"] = 38,
    ["."] = 39,
    [":"] = 40,             -- copyright symbol
    [">"] = 41,             -- mushroom
    ["$"] = { 42, 43, 44 }, -- coin (animated)
    ["*"] = 45,             -- crown
    [" "] = 47,             -- space
}

---@param text string
function print_text_nes(text, x, y, scale, fixed, otherFont)
    -- enforce pixel-perfect
    x = math.floor(x)
    y = math.floor(y)

    -- adjust for camera
    if not fixed then
        local smb = gSMBStates[0]
        x = x - smb.cameraX
        y = y + smb.cameraY
    end

    -- don't render out of view
    if x < -8 * #text * math.abs(scale) or x > NES_WIDTH or y < -8 * math.abs(scale) or y > NES_HEIGHT then
        return
    end

    local renderX = leftEdge + x * gameScaleX
    local renderY = topEdge + y * gameScaleY
    local renderScaleX = scale * gameScaleX
    local renderScaleY = scale * gameScaleY

    if otherFont then
        djui_hud_set_font(FONT_SPECIAL)
        renderScaleX = renderScaleX / 4
        local origColor = djui_hud_get_color()

        -- outline
        local darkColor = {r = origColor.r, g = origColor.g, b = origColor.b, a = origColor.a}
        darkColor.r = darkColor.r * 0.25
        darkColor.g = darkColor.g * 0.25
        darkColor.b = darkColor.b * 0.25
        local outlineThickness = 2
        djui_hud_set_color(darkColor.r, darkColor.g, darkColor.b, darkColor.a)
        djui_hud_print_text(text, renderX - outlineThickness * renderScaleX, renderY, renderScaleX)
        djui_hud_print_text(text, renderX + outlineThickness * renderScaleX, renderY, renderScaleX)
        djui_hud_print_text(text, renderX, renderY - outlineThickness * renderScaleX, renderScaleX)
        djui_hud_print_text(text, renderX, renderY + outlineThickness * renderScaleX, renderScaleX)

        djui_hud_set_color(origColor.r, origColor.g, origColor.b, origColor.a)
        djui_hud_print_text(text, renderX, renderY, renderScaleX)
        return
    end

    for point = 1, #text do
        local tex = FONT_SMB
        local char = text:sub(point, point)
        char = char:upper()
        local charNum = FONT_OVERRIDE[char]
        if not charNum then
            charNum = (string.byte(char) - 48)
            if charNum > 9 then
                charNum = charNum - 7
            end
        elseif type(charNum) == "table" then -- animated character
            local timer = get_global_timer() % 24
            if #charNum < 3 or timer < 12 then
                charNum = charNum[1]
            elseif timer >= 16 and timer < 20 then
                charNum = charNum[3]
            else
                charNum = charNum[2]
            end
        end

        -- get where on the sheet to render
        local sheetX = (8 * charNum) % tex.width
        local sheetY = ((8 * charNum) // tex.width) * 8
        djui_hud_render_texture_tile(tex, renderX, renderY, renderScaleX, renderScaleY, sheetX, sheetY, 8, 8)
        renderX = renderX + 8 * renderScaleX
    end
end

-- renders a texture from a sheet using the nes resolution, but also rotates based off of given rotation value
-- it also considers the camera unless fixed is true
-- unlike the nes, this supports any-sized sprite instead of just 8x8 or 8x16
function render_texture_nes_sheet_with_rotation(tex, x, y, scaleX, scaleY, texWidth, texHeight, frame, spaceBetweenFrames, rotation, fixed, adjustRender)
    local sheetWidth = tex.width

    -- adjust for camera
    if not fixed then
        local smb = gSMBStates[0]
        x = x - smb.cameraX
        y = y + smb.cameraY
    end

    -- don't render out of view
    if x < -texWidth * math.abs(scaleX) or x > NES_WIDTH or y < -texHeight * math.abs(scaleY) or y > NES_HEIGHT then
        return
    end

    -- get where on the sheet to render
    local framesPerLine = sheetWidth // (texWidth + spaceBetweenFrames)
    local sheetX = (texWidth + spaceBetweenFrames) * (frame % framesPerLine)
    local sheetY = (texHeight + spaceBetweenFrames) * (frame // framesPerLine)
    -- mario's big sprites need these adjustment's fsr
    if adjustRender then
        scaleX = scaleX * (texHeight / texWidth)
        y = y - texHeight + texWidth
    end

    -- enforce pixel-perfect
    x = math.floor(x)
    y = math.floor(y)

    local renderScaleX = scaleX * gameScaleX
    local renderScaleY = scaleY * gameScaleY
    local renderX = leftEdge + x * gameScaleX
    local renderY = topEdge + y * gameScaleY

    -- set rotation
    djui_hud_set_rotation(rotation, 0.5, 0.5)
    djui_hud_render_texture_tile(tex, renderX, renderY, renderScaleX, renderScaleY, sheetX, sheetY, texWidth, texHeight)
    djui_hud_set_rotation(0, 0, 0)
end

-- unused functions to render a flipped texture (very jank)
function djui_hud_render_texture_allow_flip(tex, x, y, scaleX, scaleY)
    if scaleX >= 0 and scaleY >= 0 then
        djui_hud_render_texture(tex, x, y, scaleX, scaleY) -- normal
    elseif scaleX < 0 and scaleY < 0 then                  -- this actually works, we just need to adjust x and y so it renders in the right spot
        x = x - tex.width * scaleX
        y = y - tex.height * scaleY
        djui_hud_render_texture(tex, x, y, scaleX, scaleY)
    elseif scaleX < 0 then -- x flip; render 1 pixel at a time left to right
        scaleX = -scaleX
        for i = 0, tex.width - 1 do
            djui_hud_render_texture_tile(tex, x - scaleX * (i + 1 / tex.width - tex.width + 1), y, scaleX * tex.width,
                scaleY, i, 0, 1, tex.height)
        end
    else
        scaleY = -scaleY -- y flip; render 1 pixel at a time top to bottom
        for i = 0, tex.height - 1 do
            djui_hud_render_texture_tile(tex, x, y - scaleY * (i + 1 / tex.height) + tex.height - 1, scaleX / tex.height,
                scaleY, 0, i, tex.width, 1)
        end
    end
end

function djui_hud_render_texture_tile_allow_flip(tex, x, y, scaleX, scaleY, tileX, tileY, tileW, tileH)
    if scaleX >= 0 and scaleY >= 0 then
        djui_hud_render_texture_tile(tex, x, y, scaleX, scaleY, tileX, tileY, tileW, tileH) -- normal
    elseif scaleX < 0 and scaleY < 0 then                                                   -- this actually works, we just need to adjust x and y so it renders in the right spot
        local relScale = tileW / tileH
        x = x - tileW * scaleX * relScale
        y = y - tileH * scaleY
        djui_hud_render_texture_tile(tex, x, y, scaleX, scaleY, tileX, tileY, tileW, tileH)
    elseif scaleX < 0 then -- x flip; render 1 pixel at a time left to right
        local relScale = tileW / tileH
        scaleX = -scaleX
        for i = 0, tileW - 1 do
            djui_hud_render_texture_tile(tex, x - scaleX * relScale * (i + 1 / tileW - tileW + 1), y, scaleX * tileW,
                scaleY, i + tileX, tileY, 1, tileH)
        end
    else -- y flip; render 1 pixel at a time top to bottom
        scaleY = -scaleY
        for i = 0, tileH - 1 do
            djui_hud_render_texture_tile(tex, x, y - scaleY * (i + 1 / tileH - tileH + 1), scaleX / tileH, scaleY, tileX,
                i + tileY, tileW, 1)
        end
    end
end

-- renders a player to the screen
---@param m MarioState
function render_smb_player(m, status, overrideX)
    local i = m.playerIndex
    local np = gNetworkPlayers[i]
    local smb = gSMBStates[i]

    local scaleX = 1
    local player_texture
    local animFrame = 0
    local texHeight = 16
    local pTexTable = PLAYER_TEX[m.character.type + 1] or PLAYER_TEX[1]
    local skipRendering = false
    local forcePalette = 0
    local doRainbow = false
    if smb.isBig then
        texHeight = 32
        player_texture = pTexTable[2]
    else
        player_texture = pTexTable[1]
    end
    if m.faceAngle.y < 0 then
        scaleX = -scaleX
    end

    local renderBubble = false
    if status then
        animFrame = 0
        texHeight = 16
        player_texture = pTexTable[1]
    elseif (m.action == ACT_SMB_TITLE_SCREEN and m.playerIndex ~= 0) or m.action == ACT_SMB_LOADING then
        smb.fadeTimer = 20
        skipRendering = true
    elseif m.action == ACT_SMB_BUBBLED then
        animFrame = 14
        texHeight = 16
        player_texture = pTexTable[1]
        if m.actionArg ~= 0 then
            renderBubble = true
        end
    elseif m.action == ACT_SMB_DEAD then
        if m.actionArg == 1 then
            smb.fadeTimer = 20
            skipRendering = true
        else
            animFrame = 14
            texHeight = 16
            player_texture = pTexTable[1]
        end
    elseif m.invincTimer > 5 and get_global_timer() % 3 == 0 then
        skipRendering = true
    elseif m.action == ACT_SMB_BONUS_TRANSITION then
        skipRendering = true
    elseif m.action == ACT_SMB_IN_PIPE then
        if smb.crouch and smb.isBig then
            animFrame = 14
        elseif m.actionArg == 0 then
            animFrame = smb.prevAnimFrame
        else
            animFrame = 0
        end
    elseif m.action == ACT_SMB_HIT_AXE then
        animFrame = smb.prevAnimFrame
    elseif m.action == ACT_SMB_DAMAGE or m.action == ACT_SMB_POWER_UP then
        if smb.powerup < 2 then
            animFrame = 5
        else
            animFrame = smb.prevAnimFrame
        end
        if m.actionTimer >= 16 then
            animFrame = 8
            if (not smb.isBig) then -- damage anim
                if m.actionTimer % 8 > 4 then
                    texHeight = 32
                    player_texture = pTexTable[2]
                end
            elseif smb.powerup >= 2 then
                animFrame = smb.prevAnimFrame
                if localOptionTable.nonFlashStar then
                    doRainbow = true
                else
                    forcePalette = (get_global_timer() % 4)
                end
            else -- grow anim
                if m.actionTimer < 30 then
                    if m.actionTimer % 8 > 4 then
                        animFrame = 0
                        texHeight = 16
                        player_texture = pTexTable[1]
                    else
                        animFrame = 16
                    end
                elseif m.actionTimer % 12 > 8 then
                    animFrame = 0
                    texHeight = 16
                    player_texture = pTexTable[1]
                elseif m.actionTimer % 12 > 4 then
                    animFrame = 16
                else
                    animFrame = 0
                end
            end
        end
    elseif m.action & ACT_FLAG_ON_POLE ~= 0 then
        animFrame = smb.walkCycleFrame + 6
    elseif m.action == ACT_SMB_SWIMMING then
        animFrame = (smb.walkCycleFrame * 2) + 8
        if m.marioObj and m.marioObj.oTimer % 4 > 1 then
            animFrame = animFrame + 1
        end
    elseif smb.crouch and smb.isBig then
        animFrame = 14
    elseif smb.fireBallTimer ~= 0 then
        animFrame = 15
        texHeight = 32
        player_texture = pTexTable[2]
    elseif m.action & ACT_FLAG_CONTROL_JUMP_HEIGHT ~= 0 then
        animFrame = 5
    elseif m.action & ACT_FLAG_AIR ~= 0 then
        animFrame = 3
    elseif m.action & ACT_FLAG_STATIONARY ~= 0 then
        animFrame = 0
    elseif floor_or_ceil(m.vel.x) ~= 0 and (m.faceAngle.y > 0) ~= (m.vel.x > 0) then
        animFrame = 4
    else
        animFrame = smb.walkCycleFrame + 1
    end

    if not skipRendering then
        smb.prevAnimFrame = animFrame
        
        -- convert mario's coordinates to screen pos
        local renderX = overrideX or (m.pos.x * SSPX - 8)
        local renderY = NES_HEIGHT - m.pos.y * SSPX - 15
        local smb0 = gSMBStates[0]
        if status then
            renderX = 0x60
            renderY = 0x69
        elseif m.playerIndex ~= 0 and (not overrideX) and smb0.loopScreenRight ~= 0 then
            -- rendering trickery; if a loop point is active, render players within loop thrice
            if m.pos.x > smb0.loopScreenRight * 16 * BLOCK or m.pos.x <= smb0.loopScreenLeft * 16 * BLOCK then
                return false -- don't render players beyond loop point
            else
                render_smb_player(m, false, renderX + (smb0.loopScreenRight - smb0.loopScreenLeft) * 16 * PX)
                render_smb_player(m, false, renderX - (smb0.loopScreenRight - smb0.loopScreenLeft) * 16 * PX)
            end
        end

        if smb.starTimer ~= 0 then
            if localOptionTable.nonFlashStar then
                doRainbow = true
            else
                forcePalette = (get_global_timer() % 4)
            end
        end

        if renderBubble then
            djui_hud_set_color(255, 255, 255, 255)
            render_texture_nes(TEX_BUBBLE, renderX - 8, renderY - 8, 1, 1)
        end

        -- render in 3 parts for palette
        local palette_parts = { SKIN, HAIR, CAP }
        if m.character.type == CT_TOAD then
            palette_parts[2] = GLOVES
        end
        for i, part in ipairs(palette_parts) do
            local frame = 0
            if player_texture.width == 256 then
                frame = animFrame + (i - 1) * 21 -- 16x16 format
            else
                frame = animFrame + (i - 1) * 30 -- 16x32 format
            end
            local color
            if forcePalette ~= 0 then
                color = STAR_PALETTES[forcePalette][i]
            elseif smb.powerup == 2 and part ~= SKIN and (not status) then
                if part == CAP then
                    if m.character.type == CT_TOAD then
                        color = { r = 0xB5, g = 0x31, b = 0x20 } -- red
                    elseif m.character.type == CT_LUIGI then
                        color = { r = 0xFF, g = 0xFE, b = 0xFF } -- white (to match luigi's normal NES palette)
                    else
                        color = { r = 0xF7, g = 0xD8, b = 0xA5 } -- reddish white
                    end
                    local capColor = network_player_get_override_palette_color(np, part)
                    color_override_set(capColor)

                    -- if the cap color is similar, switch to either red or white
                    if colors_are_similar(color, capColor, 40) then
                        if m.character.type ~= CT_TOAD then
                            color = { r = 0xB5, g = 0x31, b = 0x20 } -- red
                        else
                            color = { r = 0xFF, g = 0xFE, b = 0xFF } -- white
                        end
                    end
                elseif m.character.type == CT_TOAD and colors_are_similar(network_player_get_override_palette_color(np, CAP), { r = 0xFF, g = 0xFE, b = 0xFE }, 45) then
                    -- for toad, don't change glove color if the cap color is close to white (that way, toadette-like palettes don't look like normal toad)
                    color = network_player_get_override_palette_color(np, part)
                else
                    color = network_player_get_override_palette_color(np, CAP)
                end
            else
                color = network_player_get_override_palette_color(np, part)
            end
            color_override_set(color)
            -- if the hair (cap for toad) is similar to the skin, darken the hair/cap
            if forcePalette == 0 and (part == HAIR or (part == CAP and m.character.type == CT_TOAD)) then
                while (color.r ~= 0 or color.g ~= 0 or color.b ~= 0) and colors_are_similar(color, network_player_get_override_palette_color(np, SKIN), 90) do
                    color.r = math.max(color.r - 50, 0)
                    color.g = math.max(color.g - 50, 0)
                    color.b = math.max(color.b - 50, 0)
                end
            end

            local alpha = 255
            if m.playerIndex ~= 0 and smb.fadeTimer ~= 0 then
                alpha = clamp(255 - smb.fadeTimer * 15, 0, 255)
                smb.fadeTimer = smb.fadeTimer - 1
            end
            djui_hud_set_color(color.r, color.g, color.b, alpha)
            render_texture_nes_sheet(player_texture, renderX, renderY, scaleX, 1, 16, texHeight, frame, 1, true, status,
                true)
            
            if doRainbow then
                local r, g, b = hue_shift_over_time(m.marioObj.oTimer, 30)
                djui_hud_set_color(r, g, b, math.max(alpha - 155, 0))
                render_texture_nes_sheet(player_texture, renderX, renderY, scaleX, 1, 16, texHeight, frame, 1, true, status,
                    true)
            end
        end
    end

    -- debug hitboxes and colliders
    if localOptionTable.debugHitboxes then
        djui_hud_set_color(0, 255, 0, 180)
        render_rect_nes((m.pos.x - m.marioObj.hitboxRadius) * SSPX,
            NES_HEIGHT - (m.pos.y + m.marioObj.hitboxHeight - PX) * SSPX, (m.marioObj.hitboxRadius * 2) * SSPX,
            (m.marioObj.hitboxHeight) * SSPX)
        djui_hud_set_color(255, 255, 255, 255)
    end
    if localOptionTable.debugColliders then
        djui_hud_set_color(255, 0, 0, 180)
        local collider_offsets = get_collider_offsets(m)
        for i, collider in ipairs(collider_offsets) do
            render_rect_nes(m.pos.x * SSPX + collider.x, NES_HEIGHT - m.pos.y * SSPX - collider.y, 1, 1)
        end
        djui_hud_set_color(255, 255, 255, 255)
    end
    return (not skipRendering)
end

-- renders an object to the screen
---@param o Object
function render_smb_object(o, overrideX)
    local o_id = get_id_from_behavior(o.behavior)
    local spriteInfo = OBJECT_SPRITE_INFO[o_id]
    -- convert the object's coordinates to screen pos
    local texWidth, texHeight = spriteInfo[2], spriteInfo[3]
    local renderX = overrideX or (o.oPosX * SSPX - texWidth / 2)
    local renderY = NES_HEIGHT - o.oPosY * SSPX - texHeight + 1
    local scaleX = texHeight / texWidth

    -- rendering trickery; if a loop point is active, render objects within loop thrice
    local smb0 = gSMBStates[0]
    if (not overrideX) and smb0.loopScreenRight ~= 0 then
        if o.oPosX > smb0.loopScreenRight * 16 * BLOCK or o.oPosX <= smb0.loopScreenLeft * 16 * BLOCK then
            return -- don't render objects beyond loop point
        else
            render_smb_object(o, renderX + (smb0.loopScreenRight - smb0.loopScreenLeft) * 16 * PX)
            render_smb_object(o, renderX - (smb0.loopScreenRight - smb0.loopScreenLeft) * 16 * PX)
        end
    end

    local renderType = spriteInfo[5]
    if renderType == SPRITE_INFO_BOWSER_SPECIAL then
        local world = ((gNetworkPlayers[0].currActNum-1) // 4) + 1
        renderType = SPRITE_INFO_FLIP
        -- render something else when fake bowser is defeated
        if o.oFaceAnglePitch > 0x4000 and world <= #FakeBowserData then
            spriteInfo = OBJECT_SPRITE_INFO[FakeBowserData[world][1]]
            texWidth, texHeight = spriteInfo[2], spriteInfo[3]
            renderX = o.oPosX * SSPX - texWidth / 2
            renderY = NES_HEIGHT - o.oPosY * SSPX - texHeight + 1
            scaleX = texHeight / texWidth
            renderType = spriteInfo[5]
        end
    end

    if renderType < SPRITE_INFO_ROTATE then
        local scaleY = 1
        if o.oFaceAnglePitch > 0x4000 then
            scaleY = -1
            if renderType == SPRITE_INFO_NO_FLIP then
                scaleX = -scaleX
            end
        end
        if o.oFaceAngleYaw < 0 and renderType ~= SPRITE_INFO_NO_FLIP then
            scaleX = -scaleX
        end

        if o.header.gfx.scale.x == 1 then
            render_texture_nes_sheet(spriteInfo[1], renderX, renderY, scaleX, scaleY, texWidth, texHeight, o.oAnimState, spriteInfo[4], renderType ~= SPRITE_INFO_NO_FLIP)
        else -- used for platforms
            renderX = renderX - texWidth * ((o.header.gfx.scale.x - 1) / 2)
            for i = 1, o.header.gfx.scale.x do
                render_texture_nes_sheet(spriteInfo[1], renderX, renderY, scaleX, scaleY, texWidth, texHeight, o.oAnimState, spriteInfo[4], renderType ~= SPRITE_INFO_NO_FLIP)
                renderX = renderX + texWidth
            end
        end
    elseif renderType == SPRITE_INFO_ROTATE then
        local rotation = o.oFaceAngleYaw
        render_texture_nes_sheet_with_rotation(spriteInfo[1], renderX, renderY, scaleX, 1, texWidth, texHeight, o.oAnimState, spriteInfo[4], rotation)
    elseif renderType == SPRITE_INFO_1_FRAME then
        local scaleY = 1
        if o.oFaceAnglePitch > 0x4000 then
            scaleY = -1
            scaleX = -1
        end

        if o.header.gfx.scale.x == 1 then
            render_texture_nes(spriteInfo[1], renderX, renderY, scaleX, scaleY)
        else -- used for platforms
            renderX = renderX - texWidth * ((o.header.gfx.scale.x - 1) / 2)
            for i = 1, o.header.gfx.scale.x do
                render_texture_nes(spriteInfo[1], renderX, renderY, scaleX, scaleY)
                renderX = renderX + texWidth
            end
        end
    elseif renderType == SPRITE_INFO_FIREBALL_SPECIAL then
        local frame = o.oAnimState
        if o.oAnimState < 4 then
            if frame >= 2 then
                frame = frame + 2
            end
        else
            texWidth = texWidth * 2
            texHeight = texHeight * 2
            frame = frame - 3
            renderX = o.oPosX * SSPX - texWidth / 2
            renderY = NES_HEIGHT - o.oPosY * SSPX - texHeight + 1
        end
        render_texture_nes_sheet(spriteInfo[1], renderX, renderY, scaleX, 1, texWidth, texHeight, frame, spriteInfo[4])
    elseif renderType == SPRITE_INFO_TILED then
        local frame = o.oAnimState
        if frame ~= -1 then
            render_texture_nes_collection("smb_tileset", renderX, renderY, scaleX, 1, frame)
        end
    elseif renderType == SPRITE_INFO_VINE_SPECIAL then
        -- keep rendering vine tiles until we get to home
        local testY = o.oPosY
        while testY >= o.oHomeY + PX do
            local frame = o.oAnimState
            if testY == o.oPosY then
                frame = frame - 1 -- render top
            end
            render_texture_nes_sheet(spriteInfo[1], renderX, renderY, scaleX, 1, texWidth, texHeight, frame, spriteInfo[4])
            renderY = renderY + texHeight
            testY = testY - BLOCK
        end
    end

    if localOptionTable.debugHitboxes then
        djui_hud_set_color(0, 255, 0, 180)
        render_rect_nes((o.oPosX - o.hitboxRadius - o.oHitboxLeftOffset) * SSPX,
            NES_HEIGHT - (o.oPosY + o.hitboxHeight - o.hitboxDownOffset - PX) * SSPX, (o.hitboxRadius * 2) * SSPX,
            (o.hitboxHeight) * SSPX)
        if o.hurtboxRadius ~= 0 then
            render_rect_nes((o.oPosX - o.hurtboxRadius - o.oHurtboxLeftOffset) * SSPX,
                NES_HEIGHT - (o.oPosY + o.hurtboxHeight - o.oHurtboxDownOffset - PX) * SSPX, (o.hurtboxRadius * 2) * SSPX,
                (o.hitboxHeight) * SSPX)
        end
        djui_hud_set_color(255, 255, 255, 255)
    end
end

-- renders a fake object to the screen (used with the editor)
function render_fake_smb_object(x, y, spriteInfo)
    if not spriteInfo[1] then
        -- use text instead
        local texWidth, texHeight = 16, 16
        local renderX = x * SSPX - texWidth / 2
        local renderY = NES_HEIGHT - y * SSPX - texHeight + 8
        return print_text_nes(spriteInfo[2], renderX, renderY, 1)
    end

    -- convert the object's coordinates to screen pos
    local texWidth, texHeight = spriteInfo[2], spriteInfo[3]
    local renderX = x * SSPX - texWidth / 2
    local renderY = NES_HEIGHT - y * SSPX - texHeight + 1
    local scaleX = texHeight / texWidth

    local renderType = spriteInfo[5]
    if renderType >= SPRITE_INFO_BOWSER_SPECIAL then
        renderType = SPRITE_INFO_FLIP
    end
    
    if renderType <= SPRITE_INFO_ROTATE then
        local scaleY = 1
        if renderType ~= 0 then
            scaleX = -scaleX
        end
        render_texture_nes_sheet(spriteInfo[1], renderX, renderY, scaleX, scaleY, texWidth, texHeight, 0, spriteInfo[4], renderType == 1)
    elseif renderType == SPRITE_INFO_1_FRAME then
        render_texture_nes(spriteInfo[1], renderX, renderY, scaleX, 1)
    elseif renderType == SPRITE_INFO_FIREBALL_SPECIAL then
        render_texture_nes_sheet(spriteInfo[1], renderX, renderY, scaleX, 1, texWidth, texHeight, 0, spriteInfo[4])
    elseif renderType == SPRITE_INFO_TILED then
        render_texture_nes_collection(spriteInfo[1], renderX, renderY, scaleX, 1, 0, spriteInfo[4])
    end
end

-- renders a rectangle with the given x and y coordinates, and size
-- note the size is in pixels
function render_rect_nes(x, y, width, height, fixed)
    -- enforce pixel-perfect (actually no, it looks better this way)
    -- x = math.floor(x)
    -- y = math.floor(y)

    -- adjust for camera
    if not fixed then
        local smb = gSMBStates[0]
        x = x - smb.cameraX
        y = y + smb.cameraY
    end

    -- don't render out of view
    if x < -width or x > NES_WIDTH or y < -height or y > NES_HEIGHT then
        return
    end

    local renderX = leftEdge + x * gameScaleX
    local renderY = topEdge + y * gameScaleY
    local renderWidth = (width * gameScaleX)
    local renderHeight = (height * gameScaleY)

    djui_hud_render_rect(renderX, renderY, renderWidth, renderHeight)
end

-- checks if a color is similar to some color in COLOR_INPUT, and if so, replaces it with the respective color in COLOR_OVERRIDE
-- this is so using the default Mario palette makes Mario appear as he does in the original game.
-- (Luigi gets his SMB DX apperance, but with his NES green color. Other characters also have specific palettes)
local COLOR_INPUT = {
    { 0xFF, 0x00, 0x00 }, -- mario's cap/overalls (also toad's spots)
    { 0x73, 0x06, 0x00 }, -- mario/luigi's hair
    { 0xfe, 0xc1, 0x79 }, -- mario's skin (also includes all but toad)
    { 0x00, 0xFF, 0x00 }, -- luigi's cap/overalls
    { 0xfe, 0xd5, 0xa1 }, -- toad's skin
    { 0x73, 0x53, 0x00 }, -- wario/waluigi's hair
    -- { 0xFF, 0xBD, 0x00 }, -- wario's cap/overrals (removed because the nes yellow is kinda gross)
}
local COLOR_OVERRIDE = {
    { 0xB5, 0x31, 0x20 }, -- mario's cap/overalls (also toad's spots)
    { 0x6B, 0x6D, 0x00 }, -- mario/luigi's hair
    { 0xEA, 0x9E, 0x22 }, -- mario's skin (also includes all but toad)
    { 0x38, 0x87, 0x00 }, -- luigi's cap/overalls
    { 0xEA, 0x9E, 0x22 }, -- toad's skin (maps to the same as mario)
    { 0x6B, 0x6D, 0x00 }, -- wario/waluigi's hair (maps to the same as mario)
    -- { 0xBC, 0xBE, 0x00 }, -- wario's cap/overalls
}
function color_override_set(color)
    local check_dist = 5
    for i, compare_color in ipairs(COLOR_INPUT) do
        local dist = vec3f_dist({ x = color.r, y = color.g, z = color.b },
            { x = compare_color[1], y = compare_color[2], z = compare_color[3] })
        if dist <= check_dist then
            color.r, color.g, color.b = COLOR_OVERRIDE[i][1], COLOR_OVERRIDE[i][2], COLOR_OVERRIDE[i][3]
            break
        end
    end
end

-- returns TRUE if c1 is brighter than c2 (determined by distance to white)
function first_color_is_brighter(c1, c2)
    local dist1 = vec3f_dist({ x = 255, y = 255, z = 255 }, { x = c1.r, y = c1.g, z = c1.b })
    local dist2 = vec3f_dist({ x = 255, y = 255, z = 255 }, { x = c2.r, y = c2.g, z = c2.b })
    return dist1 <= dist2
end

-- returns TRUE if c1 is similar to c2
function colors_are_similar(c1, c2, checkDist)
    local dist = vec3f_dist({ x = c1.r, y = c1.g, z = c1.b }, { x = c2.r, y = c2.g, z = c2.b })
    return dist <= checkDist
end

-- returns a bright color with its hue based on the given time and max (used for the alt star effect)
function hue_shift_over_time(time, max)
    local h = time % max * (360 / max)
    local s = 0.8
    local v = 1
    -- Now it's time to convert this to RGB, which is very annoying
    local M = 255 * v
    local m = M * (1 - s)
    local z = (M - m) * (1 - math.abs(h / 60 % 2 - 1))
    -- there's SIX CASES
    if h < 60 then
        r = M
        g = z + m
        b = m
    elseif h < 120 then
        r = z + m
        g = M
        b = m
    elseif h < 180 then
        r = m
        g = M
        b = z + m
    elseif h < 240 then
        r = m
        g = z + m
        b = M
    elseif h < 300 then
        r = z + m
        g = m
        b = M
    else
        r = M
        g = m
        b = z + m
    end
    return r, g, b
end

-- objects --

-- 2D dist between objects
function get_dist_between_objects_2d(o1, o2)
    if not (o1 and o2) then return 10000 end

    return math.sqrt(sqr(o2.oPosX - o1.oPosX) + sqr(o2.oPosY - o1.oPosY))
end

-- X dist between objects
function get_horizontal_dist_between_objects(o1, o2)
    if not (o1 and o2) then return 10000 end

    return math.abs(o2.oPosX - o1.oPosX)
end

-- spawns a sync/non-sync object and resets all of its rotation angles
function spawn_object_no_rotate(o_id, model, x, y, z, sync, initFunc)
    local trueInitFunc = function(o)
        o.oFaceAngleYaw, o.oFaceAnglePitch, o.oFaceAngleRoll = 0, 0, 0
        o.oMoveAngleYaw, o.oMoveAnglePitch, o.oMoveAngleRoll = 0, 0, 0
        if initFunc then initFunc(o) end
    end
    if sync then
        return spawn_sync_object(o_id, model, x, y, z, trueInitFunc)
    else
        return spawn_non_sync_object(o_id, model, x, y, z, trueInitFunc)
    end
end

-- check if mario is overlapping with an object hitbox
---@param m MarioState
---@param o Object
function mario_and_object_hitbox_overlap(m, o)
    local result = m.pos.x + m.marioObj.hitboxRadius >= o.oPosX - o.hitboxRadius - o.oHitboxLeftOffset -- left
        and m.pos.x - m.marioObj.hitboxRadius <= o.oPosX + o.hitboxRadius - o.oHitboxLeftOffset        -- right
        and m.pos.y + m.marioObj.hitboxHeight >= o.oPosY - o.hitboxDownOffset                          -- bottom
        and m.pos.y <= o.oPosY - o.hitboxDownOffset + o.hitboxHeight                                   -- top

    -- treat hurtbox as second hitbox if set (used only by Bowser)
    if (not result) and o.hurtboxRadius ~= 0 then
        result = m.pos.x + m.marioObj.hurtboxRadius >= o.oPosX - o.hurtboxRadius - o.oHurtboxLeftOffset -- left
            and m.pos.x - m.marioObj.hurtboxRadius <= o.oPosX + o.hurtboxRadius - o.oHurtboxLeftOffset  -- right
            and m.pos.y + m.marioObj.hurtboxHeight >= o.oPosY - o.oHurtboxDownOffset                    -- bottom
            and m.pos.y <= o.oPosY - o.oHurtboxDownOffset + o.hurtboxHeight                             -- top
    end
    return result
end

-- check if mario is overlapping with another mario
---@param m1 MarioState
---@param m2 MarioState
function mario_and_mario_hitbox_overlap(m1, m2)
    local o = m2.marioObj
    local result = m1.pos.x + m1.marioObj.hitboxRadius >= m2.pos.x - o.hitboxRadius -- left
        and m1.pos.x - m1.marioObj.hitboxRadius <= m2.pos.x + o.hitboxRadius        -- right
        and m1.pos.y + m1.marioObj.hitboxHeight >= m2.pos.y - o.hitboxDownOffset    -- bottom
        and m1.pos.y <= m2.pos.y - o.hitboxDownOffset + o.hitboxHeight              -- top
    return result
end

-- check if two objects have overlapping hitboxes
---@param o1 Object
---@param o2 Object
function object_and_object_hitbox_overlap(o1, o2)
    local result = o1.oPosX + o1.hitboxRadius - o1.oHitboxLeftOffset >=
        o2.oPosX - o2.hitboxRadius -
        o2.oHitboxLeftOffset -- left
        and
        o1.oPosX - o1.hitboxRadius - o1.oHitboxLeftOffset <=
        o2.oPosX + o2.hitboxRadius -
        o2.oHitboxLeftOffset -- right
        and
        o1.oPosY - o1.hitboxDownOffset + o1.hitboxHeight >=
        o2.oPosY - o2.hitboxDownOffset                   -- bottom
        and o1.oPosY - o1.hitboxDownOffset <=
        o2.oPosY - o2.hitboxDownOffset + o2.hitboxHeight -- top

    -- treat hurtbox as second hitbox if set (used only by Bowser)
    -- there will never be two bowsers interacting with each other
    if (not result) then
        if o1.hurtboxRadius ~= 0 then
            result = o1.oPosX + o1.hurtboxRadius - o1.oHurtboxLeftOffset >=
                o2.oPosX - o2.hitboxRadius -
                o2.oHitboxLeftOffset -- left
                and
                o1.oPosX - o1.hurtboxRadius - o1.oHurtboxLeftOffset <=
                o2.oPosX + o2.hitboxRadius -
                o2.oHitboxLeftOffset                             -- right
                and o1.oPosY - o1.oHurtboxDownOffset + o1.hurtboxHeight >=
                o2.oPosY - o2.hitboxDownOffset                   -- bottom
                and o1.oPosY - o1.oHurtboxDownOffset <=
                o2.oPosY - o2.hitboxDownOffset + o2.hitboxHeight -- top
        elseif o2.hurtboxRadius ~= 0 then
            result = o1.oPosX + o1.hitboxRadius - o1.oHitboxLeftOffset >=
                o2.oPosX - o2.hurtboxRadius -
                o2.oHurtboxLeftOffset -- left
                and
                o1.oPosX - o1.hitboxRadius - o1.oHitboxLeftOffset <=
                o2.oPosX + o2.hurtboxRadius -
                o2.oHurtboxLeftOffset                               -- right
                and o1.oPosY - o1.hitboxDownOffset + o1.hitboxHeight >=
                o2.oPosY - o2.oHurtboxDownOffset                    -- bottom
                and o1.oPosY - o1.hitboxDownOffset <=
                o2.oPosY - o2.oHurtboxDownOffset + o2.hurtboxHeight -- top
        end
    end
    return result
end

-- check if mario's hitbox is within a rectangle (starting from the top left)
function mario_rect_hitbox_overlap(m, x, y, width, height)
    local o = m.marioObj
    local result = x + width >= m.pos.x - o.hitboxRadius                -- left
        and x <= m.pos.x + o.hitboxRadius                               -- right
        and y >= m.pos.y - o.hitboxDownOffset                           -- bottom
        and y - height <= m.pos.y - o.hitboxDownOffset + o.hitboxHeight -- top
    return result
end

-- check if an object's hitbox is within a rectangle (starting from the top left)
function object_rect_hitbox_overlap(o, x, y, width, height)
    local result = x + width >= o.oPosX - o.hitboxRadius - o.oHitboxLeftOffset -- left
        and x <= o.oPosX + o.hitboxRadius - o.oHitboxLeftOffset                -- right
        and y >= o.oPosY - o.hitboxDownOffset                                  -- bottom
        and y - height <= o.oPosY - o.hitboxDownOffset + o.hitboxHeight        -- top

    -- treat hurtbox as second hitbox if set (used only by Bowser)
    if (not result) and o.hurtboxRadius ~= 0 then
        result = x + width >= o.oPosX - o.hurtboxRadius - o.oHurtboxLeftOffset -- left
            and x <= o.oPosX + o.hurtboxRadius - o.oHurtboxLeftOffset          -- right
            and y >= o.oPosY - o.oHurtboxDownOffset                            -- bottom
            and y - height <= o.oPosY - o.oHurtboxDownOffset + o.hurtboxHeight -- top
    end
    return result
end

-- gets if an object is on-screen or within the specified distance of the screen
-- if noOthers is true, it will only check the local player
-- also returns which side the object is on
function obj_is_on_screen(o, leftDist, rightDist, noOthers)
    for i = 0, MAX_PLAYERS - 1 do
        local m = gMarioStates[i]
        local smb = gSMBStates[i]
        if is_player_active_and_synced(m) ~= 0 and m.action ~= ACT_SMB_LOADING and o.oPosX + 8 * PX >= (smb.cameraX * PX - leftDist) and o.oPosX - 8 * PX <= ((smb.cameraX + NES_WIDTH) * PX + rightDist) then
            local side = 1
            if o.oPosX <= (smb.cameraX + NES_WIDTH) // 2 * PX and m.prevAction ~= ACT_SMB_LOADING then
                side = -1
            end
            return true, side
        end
        if noOthers then break end
    end
    return false, 0
end

-- gets if the specified X coordinate is on-screen or within the specified distance of the screen
-- if noOthers is true, it will only check the local player
function point_is_on_screen(x, leftDist, rightDist, noOthers)
    for i = 0, MAX_PLAYERS - 1 do
        local m = gMarioStates[i]
        local smb = gSMBStates[i]
        if is_player_active_and_synced(m) ~= 0 and m.action ~= ACT_SMB_LOADING and x + 8 * PX > (smb.cameraX * PX - leftDist) and x - 8 * PX < ((smb.cameraX + NES_WIDTH) * PX + rightDist) then
            local side = 1
            if x <= m.pos.x and m.prevAction ~= ACT_SMB_LOADING then
                side = -1
            end
            return true, side
        end
        if noOthers then break end
    end
    return false, 0
end

-- check if platforms should be 2-wide instead of 3-wide (and also controls other hard mode objects)
-- TRUE if hard mode is enabled, OR if this is 5-3 or later
function do_hard_mode_objs()
    local np = gNetworkPlayers[0]
    return (gGlobalSyncTable.hardMode or (np.currActNum >= 19))
end

-- picks a random value from this table using the current object's position
function cur_obj_random_from_table(randomTable)
    local u16 = position_based_random_u16()
    local value = (u16 % #randomTable) + 1
    return randomTable[value]
end

-- returns the mario state that is the farthest to the right and isn't in a cutscene action
-- it also checks their frenzy state is frenzy is TRUE
function get_rightmost_mario_state(frenzy)
    local farthestPos = 0
    local farthestM

    for g=0,MAX_PLAYERS-1 do
        -- use global index so it's consistent between players
        local i = network_local_index_from_global(g) or 255
        if i ~= 255 then
            local m = gMarioStates[i]
            local smb = gSMBStates[i]
            if is_player_active_and_synced(m) ~= 0 and not (frenzy and smb.frenzyDisabled) then
                if m.action & ACT_GROUP_MASK ~= ACT_GROUP_CUTSCENE and m.pos.x > farthestPos then
                    farthestPos = m.pos.x
                    farthestM = m
                end
            end
        end
    end
    return farthestM
end

-- returns TRUE if no other player's screens are farther right
-- and close to our screen, and our frenzy state is valid,
-- AND our screen is close enough to this object
function frenzy_spawn_is_valid(o, maxDist)
    local m0 = gMarioStates[0]
    local smb0 = gSMBStates[0]
    if smb0.frenzyDisabled or m0.action == ACT_SMB_LOADING then return false end
    local rightScreenPoint0 = (smb0.cameraX + NES_WIDTH) * PX
    if rightScreenPoint0 < o.oPosX - 8 * PX then return false end

    for i=1,MAX_PLAYERS-1 do
        local m = gMarioStates[i]
        local smb = gSMBStates[i]
        local leftScreenPoint = smb.cameraX * PX
        local rightScreenPoint = leftScreenPoint + NES_WIDTH * PX

        -- check if THEIR right side is further or the same distance as OUR right side
        local rightScreenFurther = (is_player_active_and_synced(m) ~= 0 and m0.action ~= ACT_SMB_LOADING and rightScreenPoint0 <= rightScreenPoint)
        -- if our screens are at the exact same point, the one with the lower global index is considered further
        if rightScreenFurther and rightScreenPoint0 == rightScreenPoint then
            rightScreenFurther = (network_global_index_from_local(0) > network_global_index_from_local(i))
        end

        if rightScreenFurther then
            -- return false if THEIR left side is within a certain distance from OUR right side
            if leftScreenPoint <= rightScreenPoint0 + maxDist then
                return false
            end
        end
    end
    return true
end

-- count how many objects with this id are on our screen
function obj_count_on_screen(o_id, leftDist, rightDist)
    local o = obj_get_first_with_behavior_id(o_id)
    local count = 0
    while o do
        if obj_is_on_screen(o, leftDist, rightDist, true) then
            count = count + 1
        end
        o = obj_get_next_with_same_behavior_id(o)
    end
    return count
end

-- get the nearest mario state that isn't in a cutscene action to this object
function nearest_non_cutscene_mario_state_to_object(o)
    local closestDist = -1
    local closestM
    for i=0,MAX_PLAYERS-1 do
        local m = gMarioStates[i]
        if is_player_active(m) ~= 0 and m.action & ACT_GROUP_MASK ~= ACT_GROUP_CUTSCENE then
            local dist = get_dist_between_objects_2d(o, m.marioObj)

            if closestDist == -1 or closestDist > dist then
                closestDist = dist
                closestM = m
            end
        end
    end
    return closestM
end

-- get the nearest mario state HORIZONTALLY to this object (does include cutscenes; it is used for piranha plants)
function nearest_mario_state_to_object_horizontal(o)
    local closestDist = -1
    local closestM
    for i=0,MAX_PLAYERS-1 do
        local m = gMarioStates[i]
        if is_player_active(m) ~= 0 then
            local dist = get_horizontal_dist_between_objects(o, m.marioObj)

            if closestDist == -1 or closestDist > dist then
                closestDist = dist
                closestM = m
            end
        end
    end
    return closestM
end

--- math ---

-- square function
function sqr(value)
    return value * value
end

-- floor positive values and ceil negative values
function floor_or_ceil(value)
    if value >= 0 then return math.floor(value) end
    return math.ceil(value)
end

-- floor based on given multiplier
function truncate(value, multiplier)
    return floor_or_ceil(value / multiplier) * multiplier
end

-- returns 1 or -1 depending on given value's sign
function sign(value)
    if value >= 0 then return 1 end
    return -1
end

-- from extended moveset; limits angle from -0x8000 to 0x8000
function limit_angle(a)
    return (a + 0x8000) % 0x10000 - 0x8000
end

--- misc ---

-- returns the value as a string in hex
function hexString(value)
    if value >= 0 then
        return string.format("%x", value)
    else
        return string.format("-%x", -value)
    end
end

-- copies a table and all tables inside of it, recursively
function deep_copy(oldTable)
    local newTable = {}
    for i, v in pairs(oldTable) do
        if type(v) == "table" then
            newTable[i] = deep_copy(v)
        else
            newTable[i] = v
        end
    end
    return newTable
end

-- does time stop if we're alone, and also sets the timeStopEnabled value
function custom_handle_time_stop(includeMario)
    local flags = TIME_STOP_ENABLED
    if includeMario then
        flags = flags | TIME_STOP_MARIO_AND_DOORS
    end

    local alone = (network_player_connected_count() <= 1)
    if alone then
        set_time_stop_flags(flags)
        timeStopEnabled = true
    else
        clear_time_stop_flags(flags)
        timeStopEnabled = false
    end
end

-- checks if the player is active AND their sync status is valid
---@param m MarioState
function is_player_active_and_synced(m)
    if m.playerIndex == 0 then return 1 end
    if is_player_active(m) == 0 then return 0 end
    local np = gNetworkPlayers[m.playerIndex]
    if not (np.connected and np.currLevelSyncValid and np.currAreaSyncValid) then return 0 end
    return 1
end

-- checks if we can bubble (based on settings, our lives, and if any non-cutscene player is in this area)
---@param m MarioState
function mario_can_bubble_smb(m)
    if gServerSettings.bubbleDeath == 0 then return false end
    if m.numLives <= 0 then return false end
    if localOptionTable.debugBubble then return true end

    for i=0,MAX_PLAYERS-1 do
        local m2 = gMarioStates[i]
        if i ~= m.playerIndex and is_player_active(m2) ~= 0 and m2.action & ACT_GROUP_MASK ~= ACT_GROUP_CUTSCENE then
            return true
        end
    end
    return false
end

-- adds a menu option with a save function attached
-- table is either gGlobalSyncTable for host options or localOptionTable for other options
function add_smb_menu_option(name, option, useTable)
    if mod_storage_exists(option) then
        useTable[option] = mod_storage_load_bool(option)
    end
    hook_mod_menu_checkbox(name, useTable[option], function(index, value)
        useTable[option] = value
        mod_storage_save_bool(option, value)
    end)
end