--- object behavior and movement utils ---
define_custom_obj_fields({
    oRenderBehindFloor = "u32",
    oHurtboxDownOffset = "s32",
    oHitboxLeftOffset = "s32",
    oHurtboxLeftOffset = "s32",
    oSkipObjectInteraction = "u32",
    oHurtsEnemies = "u32",
    oTurnTowardsOnHit = "u32",
    oIntangibleToMarioTimer = "s32",
    oShellCombo = "s32",
    oFireballOverridePoints = "u32",
    oGroupNumber = "u32",
    oSMBTimer2 = "u32",
    oBowserMaxHomeDist = "u32",
    oBowserFlameProgress = "u32",
    oBowserBroJumpTimer = "u32",
    oImmuneToFire = "u32",
})

-- use base sm64 interact types in different contexts
INTERACT_POWERUP = INTERACT_UNKNOWN_08
INTERACT_SPRINGBOARD = INTERACT_WATER_RING

--- utils ---
-- reformatted version of hook_behavior that also calls loop twice per frame
-- this means that objects don't update until Mario has already updated twice
-- thankfully, SMB's object collision is only every OTHER frame,
-- so this only causes inaccuracies with surface objects.
function hook_smb_behavior(name, list, initFunc, loopFunc)
    return hook_behavior(nil, list, true, initFunc, function(o)
        if o.activeFlags & ACTIVE_FLAG_DEACTIVATED ~= 0 then return end
        loopFunc(o)
        o.oTimer = o.oTimer + 1
        if (not localOptionTable.debugHalfSpeed) and o.activeFlags & ACTIVE_FLAG_DEACTIVATED == 0 then
            loopFunc(o)
        end
    end, name)
end

-- handle object movement and collision for enemies; it's way simpler than mario's
---@param o Object
function enemy_move_and_collide(o, turnAtLedge)
    o.oVelY = math.max(-TERMINAL_VEL, o.oVelY - o.oGravity)
    local landed, hitWall = false, false

    -- horizontal
    o.oPosX = o.oPosX + floor_or_ceil(o.oVelX)
    local tile, px = get_tile_at_coordinate_pos(o.oPosX + o.hitboxRadius * sign(o.oVelX), o.oPosY)
    if tile ~= 0 and is_tile_solid(tile) and px < 11 and o.oPosY + BLOCK < (NES_HEIGHT - 32) * PX then
        hitWall = true
    end

    -- vertical
    o.oPosY = o.oPosY + floor_or_ceil(o.oVelY)
    tile, px = get_tile_at_coordinate_pos(o.oPosX, o.oPosY)
    if tile ~= 0 and is_tile_solid(tile) and o.oVelY < 0 then
        -- land on tile
        o.oPosY = truncate(o.oPosY+PX * (PX - px), PX) + o.oPosY % 1
        o.oVelY = 0
        landed = true
    elseif turnAtLedge then
        -- land on tile anyway
        o.oPosY = truncate(o.oPosY+PX * (PX - px), PX) + o.oPosY % 1
        o.oVelY = 0
        landed = true
        hitWall = true
    end
    return landed, hitWall
end

-- move without collision
function enemy_move(o, gravity)
    o.oPosX = o.oPosX + floor_or_ceil(o.oVelX)
    o.oPosY = o.oPosY + floor_or_ceil(o.oVelY)
    if gravity then
        o.oVelY = math.max(-TERMINAL_VEL, o.oVelY - o.oGravity)
    end
end

-- returns frame based on the current level theme
function frame_based_on_level_theme(frame, framesWidth, paletteTotal, skipCastle)
    local palette = loaded_tile_data.palette or loaded_tile_data.theme or 0
    if palette == 0 or palette > 3 then
        return frame
    end
    if skipCastle and palette == 2 then
        palette = palette - 1
    end
    if palette >= paletteTotal then palette = paletteTotal - 1 end
    frame = frame + (framesWidth * palette)
    return frame
end

-- spawns a coin particle and optionally awards coin + points
function do_coin_particle(blockX, blockY, award)
    local coordX = (blockX + 0.5) * BLOCK
    local coordY = blockY * BLOCK
    spawn_object_no_rotate(id_bhvSmbCoinParticle, E_MODEL_NONE, coordX, coordY, 0, false, function(o)
        o.oBehParams = (award and 1) or 0
    end)
    if award then
        gGlobalSyncTable.coins = gGlobalSyncTable.coins + 1
        award_points(gMarioStates[0], 2, false)
    end
end

-- handles collision and displacement for moving platforms
function smb_handle_surface_obj(o)
    -- check if we're overlapping with any marios' feet; if so, make mario land on this platform
    local count = 0
    for i=0,MAX_PLAYERS-1 do
        local m = gMarioStates[i]
        local smb = gSMBStates[i]
        if is_player_active_and_synced(m) ~= 0 and m.action & ACT_GROUP_MASK ~= ACT_GROUP_CUTSCENE then
            if smb.platform == o then -- displacement
                m.pos.x = m.pos.x + floor_or_ceil(o.oVelX)
                m.pos.y = m.pos.y + floor_or_ceil(o.oVelY)
                count = count + 1
            else
                -- check if mario's feet are overlapping with this platform; if so, make him land on it
                local collider_offsets = get_collider_offsets(m)
                local collider = collider_offsets[COLLIDER_LEFT_FOOT]
                local x = m.pos.x + collider.x * PX
                local y = m.pos.y + collider.y * PX
                x = math.floor(x)
                y = math.floor(y)
                local valid = object_rect_hitbox_overlap(o, x, y, 0, 0)
                if not valid then
                    collider = collider_offsets[COLLIDER_RIGHT_FOOT]
                    x = m.pos.x + collider.x * PX
                    y = m.pos.y + collider.y * PX
                    x = math.floor(x)
                    y = math.floor(y)
                    valid = object_rect_hitbox_overlap(o, x, y, 0, 0)
                end
                if valid then
                    local px = ((y - (o.oPosY - o.hitboxDownOffset)) % BLOCK) // PX
                    if px >= 10 then
                        -- land on platform
                        m.pos.y = o.oPosY + o.hitboxHeight - o.hitboxDownOffset
                        m.vel.y = 0
                        smb.platform = o
                        count = count + 1
                    end
                end
            end
        end
    end
    return count
end

-- load all objects at the start of a level
-- sync objects only load if we're the first to enter, while nonsync objects always load unless onlySync is true
function initial_object_load(loadSync, onlySync)
    local width = loaded_tile_data.width
    local height = 13
    for x=0,width-1 do
        for y=0,height-1 do
            local this_obj_data = loaded_obj_data[x] and loaded_obj_data[x][y]
            local o_id = this_obj_data and this_obj_data[1]
            if o_id and ((not this_obj_data[4]) or do_hard_mode_objs()) then
                local spriteInfo = OBJECT_SPRITE_INFO[o_id]
                local sync = true
                if spriteInfo and spriteInfo[6] == true then
                    sync = false
                end
                -- replace goombas with buzzy beetles in hard mode
                if o_id == id_bhvSmbGoomba and gGlobalSyncTable.hardMode then
                    o_id = id_bhvSmbBuzzyBeetle
                end

                local xOffset = spriteInfo[7] or 0
                local yOffset = spriteInfo[8] or 0
                xOffset = xOffset * PX
                yOffset = yOffset * PX

                local valid = true
                if sync and not loadSync then
                    valid = false
                elseif onlySync and not sync then
                    valid = false
                end

                if valid then
                    if this_obj_data[2] <= 1 then
                        spawn_object_no_rotate(o_id, E_MODEL_NONE, (x + 0.5) * BLOCK + xOffset, y * BLOCK + yOffset, 0, sync, function(o)
                            o.oBehParams = this_obj_data[3] or 0
                            o.oBehParams2ndByte = (o.oBehParams >> 16) & 0xFF
                        end)
                    else -- object group
                        local x2 = x
                        for i=1,this_obj_data[2] do
                            spawn_object_no_rotate(o_id, E_MODEL_NONE, (x2 + 0.5) * BLOCK + xOffset, y * BLOCK + yOffset, 0, sync, function(o)
                                o.oBehParams = this_obj_data[3] or 0
                                o.oBehParams2ndByte = (o.oBehParams >> 16) & 0xFF
                                o.oGroupNumber = (this_obj_data[2]-i)
                                -- set home to last of all spawned objects so they all load at once
                                o.oHomeX = (x + (this_obj_data[2]-1) * 1.5) * BLOCK
                                o.oHomeY = y * BLOCK
                                o.oHomeZ = 0
                            end)
                            
                            x2 = x2 + 1.5
                        end
                    end
                end
            end
        end
    end
end

-- try to spawn a hammer. 8 is the max, just like in the original game
-- this limitation is implementing here just to make bowser and hammer bros less op :P
function attempt_spawn_hammer(o)
    -- randomly fail based on hammer amount (more hammers makes hammers less likely to spawn)
    -- it will never fail at zero hammers, and always fail at 8
    local random = math.random(1, 8)
    local count = obj_count_on_screen(id_bhvSmbHammer, 0, 0)
    if count >= random then return end

    return spawn_object_no_rotate(id_bhvSmbHammer, E_MODEL_NONE, o.oPosX, o.oPosY, 0, true, function(hammer)
        hammer.parentObj = o
    end)
end

--- object behavior scripts ---
-- goomba
---@param o Object
function smb_goomba_init(o)
    o.oFlags = OBJ_FLAG_SET_FACE_YAW_TO_MOVE_YAW
    o.oGravity = 4
    o.oMoveAngleYaw = -0x4000
    o.oInteractType = INTERACT_BOUNCE_TOP
    o.hitboxRadius = 5 * PX -- 5 pixels on both sides from center
    o.hitboxHeight = 6 * PX
    o.hitboxDownOffset = -4 * PX
    o.oDeathSound = 0 -- used as "played death sound" flag
    o.oDamageOrCoinValue = 1 -- 100 points by default
    o.oFireballOverridePoints = 1 -- always 100 points when defeated by a fireball
    o.oAnimState = frame_based_on_level_theme(0, 4, 3)
    o.oTurnTowardsOnHit = 1
    if o.oHomeY == 0 then
        cur_obj_set_home_once()
    end
    network_init_object(o, true, {"oHomeX", "oHomeY", "oVelX", "oVelY", "oGroupNumber"})
end

---@param o Object
function smb_goomba_loop(o)
    cur_obj_unhide()

    if o.oAction == 0 then
        if not obj_is_on_screen(o, 4 * BLOCK, 4 * BLOCK) then
            cur_obj_change_action(3)
            return
        end

        o.oFaceAnglePitch = 0
        o.oVelX = sign(o.oMoveAngleYaw) * 8
        -- move faster in hard mode (yes I know goombas don't appear in hard mode)
        if gGlobalSyncTable.hardMode then
            o.oVelX = sign(o.oMoveAngleYaw) * 12
        end
        cur_obj_become_tangible()
        local landed, hitWall = enemy_move_and_collide(o)

        if hitWall then
            o.oMoveAngleYaw = -o.oMoveAngleYaw
        end

        if o.oTimer % 8 == 0 then
            o.oAnimState = frame_based_on_level_theme((o.oAnimState + 1) % 2, 4, 3)
        end

        if o.oInteractStatus & INT_STATUS_INTERACTED ~= 0 and o.oInteractStatus & INT_STATUS_WAS_ATTACKED ~= 0 then
            if o.oInteractStatus & INT_STATUS_ATTACK_MASK == ATTACK_FROM_ABOVE then
                cur_obj_change_action(1)
            else
                o.oVelY = 4 * PX
                o.oVelX = -o.oVelX
                o.oMoveAngleYaw = -o.oMoveAngleYaw
                cur_obj_change_action(2)
            end
        end
    elseif o.oAction == 1 then -- stomp death
        cur_obj_become_intangible()
        play_smb_sound_object(o, "stomp", true)
        o.oAnimState = frame_based_on_level_theme(2, 4, 3)
        if o.oTimer > 30 then
            obj_mark_for_deletion(o)
        end
    elseif o.oAction == 2 then -- other death
        o.oFaceAnglePitch = 0x8000 -- flip sprite
        cur_obj_become_intangible()
        play_smb_sound_object(o, "kick", true)
        o.oAnimState = frame_based_on_level_theme(0, 4, 3)
        enemy_move(o, true)
    elseif o.oAction == 3 then -- off-screen
        cur_obj_hide()
        cur_obj_become_intangible()
        local homeOnScreen, side = point_is_on_screen(o.oHomeX, BLOCK, 4 * BLOCK)
        if o.oSubAction == 0 and not homeOnScreen then
            o.oPosX, o.oPosY = (o.oHomeX - (1.5 * o.oGroupNumber) * BLOCK), o.oHomeY
            o.oSubAction = 1
        elseif o.oSubAction == 1 and homeOnScreen then
            o.oMoveAngleYaw = -0x4000 * side
            cur_obj_change_action(0)
        end
    end

    if o.oPosY < -16 * PX then
        if o.oAction == 2 then
            obj_mark_for_deletion(o)
        else
            cur_obj_change_action(3)
        end
    end

    o.oInteractStatus = 0
end

id_bhvSmbGoomba = hook_smb_behavior("bhvSmbGoomba", OBJ_LIST_GENACTOR, smb_goomba_init, smb_goomba_loop)

-- koopa (both red and green)
---@param o Object
function smb_koopa_init(o)
    o.oFlags = OBJ_FLAG_SET_FACE_YAW_TO_MOVE_YAW
    o.oGravity = 4
    o.oMoveAngleYaw = -0x4000
    o.oInteractType = INTERACT_BOUNCE_TOP
    o.hitboxRadius = 6 * PX -- 6 pixels on both sides from center
    o.hitboxHeight = 12 * PX
    o.hitboxDownOffset = -3 * PX
    o.oDeathSound = 0 -- used as "played death sound" flag
    o.oDamageOrCoinValue = 1 -- 100 points by default

    o.oAnimState = 0
    if o.oBehParams == 0 then
        o.oAnimState = frame_based_on_level_theme(o.oAnimState, 7, 2)
    else
        o.oAnimState = o.oAnimState + 14
    end
    
    o.oTurnTowardsOnHit = 1
    if o.oHomeY == 0 then
        cur_obj_set_home_once()
    end
    network_init_object(o, true, {"oHomeX", "oHomeY", "oVelX", "oVelY", "globalPlayerIndex", "oGroupNumber"})
end

---@param o Object
function smb_koopa_loop(o)
    cur_obj_unhide()

    if o.oIntangibleToMarioTimer > 0 then
        o.oIntangibleToMarioTimer = o.oIntangibleToMarioTimer - 1
    end

    if o.oAction == 0 then
        o.oHurtsEnemies = 0
        o.oTurnTowardsOnHit = 1
        if not obj_is_on_screen(o, 4 * BLOCK, 4 * BLOCK) then
            cur_obj_change_action(3)
            return
        end

        o.oFaceAnglePitch = 0
        o.oInteractType = INTERACT_BOUNCE_TOP
        o.oVelX = sign(o.oMoveAngleYaw) * 8
        -- move faster in hard mode
        if gGlobalSyncTable.hardMode then
            o.oVelX = sign(o.oMoveAngleYaw) * 12
        end
        cur_obj_become_tangible()
        local landed, hitWall = enemy_move_and_collide(o, (o.oBehParams ~= 0 and o.oBehParams2ndByte == 0))

        if hitWall then
            o.oMoveAngleYaw = -o.oMoveAngleYaw
        end

        if o.oTimer % 8 == 0 then
            o.oAnimState = (o.oAnimState%7 + 1) % 2
            if o.oBehParams == 0 then
                o.oAnimState = frame_based_on_level_theme(o.oAnimState, 7, 2)
            else
                o.oAnimState = o.oAnimState + 14
            end
        end

        if o.oInteractStatus & INT_STATUS_INTERACTED ~= 0 and o.oInteractStatus & INT_STATUS_WAS_ATTACKED ~= 0 then
            if o.oInteractStatus & INT_STATUS_ATTACK_MASK == ATTACK_FROM_ABOVE then
                o.oVelX = 0
                o.oIntangibleToMarioTimer = 5
                o.oInteractType = INTERACT_KOOPA_SHELL
                cur_obj_change_action(1)
            elseif o.oInteractStatus & INT_STATUS_ATTACK_MASK == ATTACK_FROM_BELOW then
                o.oVelY = 4 * PX
                o.oFaceAnglePitch = 0x8000
                o.oInteractType = INTERACT_KOOPA_SHELL
                cur_obj_change_action(1)
            else
                o.oVelY = 4 * PX
                o.oVelX = -o.oVelX
                o.oMoveAngleYaw = -o.oMoveAngleYaw
                cur_obj_change_action(2)
            end
        end
    elseif o.oAction == 1 then -- in shell
        if not obj_is_on_screen(o, 4 * BLOCK, 4 * BLOCK) then
            cur_obj_change_action(3)
            return
        end

        o.oTurnTowardsOnHit = 0

        cur_obj_become_tangible()
        if o.oTimer == 1 then
            play_smb_sound_object(o, "stomp")
        end
        
        if o.oSubAction == 0 then -- idle
            o.oInteractType = INTERACT_KOOPA_SHELL
            o.oHurtsEnemies = 0
        else -- kicked
            o.oInteractType = INTERACT_BOUNCE_TOP
            o.oHurtsEnemies = 1
            o.oTimer = 2
        end
        local landed, hitWall = enemy_move_and_collide(o)
        if landed then
            if o.oSubAction == 0 then
                o.oVelX = 0
            else
                o.oVelX = sign(o.oMoveAngleYaw) * 0x30
            end
        elseif o.oSubAction ~= 0 then
            o.oVelX = sign(o.oMoveAngleYaw) * 0x18
        end

        if hitWall and o.oSubAction ~= 0 then
            o.oMoveAngleYaw = -o.oMoveAngleYaw
            o.oVelX = math.abs(o.oVelX) * sign(o.oMoveAngleYaw)
            play_smb_sound_object(o, "bump")
            o.oIntangibleToMarioTimer = 0
        end

        if o.oTimer < 10 * 21 or o.oTimer % 16 >= 8 then
            o.oAnimState = 4
        else
            o.oAnimState = 5
        end

        if o.oBehParams == 0 then
            o.oAnimState = frame_based_on_level_theme(o.oAnimState, 7, 2)
        else
            o.oAnimState = o.oAnimState + 14
        end

        if o.oInteractStatus & INT_STATUS_INTERACTED ~= 0 then
            if o.oInteractStatus & INT_STATUS_WAS_ATTACKED ~= 0 then
                if o.oInteractStatus & INT_STATUS_ATTACK_MASK == ATTACK_FROM_ABOVE then
                    if o.oSubAction ~= 0 then
                        o.oVelX = 0
                        o.oIntangibleToMarioTimer = 5
                        o.oInteractType = INTERACT_KOOPA_SHELL
                        cur_obj_change_action(1)
                    end
                elseif o.oInteractStatus & INT_STATUS_ATTACK_MASK == ATTACK_FROM_BELOW then
                    o.oVelY = 4 * PX
                    o.oFaceAnglePitch = 0x8000
                    o.oInteractType = INTERACT_KOOPA_SHELL
                    cur_obj_change_action(1)
                elseif o.oInteractStatus & INT_STATUS_ATTACK_MASK == ATTACK_KICK_OR_TRIP then
                    if o.oSubAction == 0 then
                        o.oIntangibleToMarioTimer = 20
                        o.oSubAction = 1
                        play_smb_sound_object(o, "kick")
                    end
                else
                    o.oVelY = 4 * PX
                    o.oMoveAngleYaw = -o.oMoveAngleYaw
                    o.oVelX = sign(o.oMoveAngleYaw) * 8
                    cur_obj_change_action(2)
                end
                o.oInteractStatus = 0
                return
            end
        end

        if o.oTimer >= 16 * 21 then
            cur_obj_change_action(0)
            o.oAnimState = 0
            if o.oBehParams == 0 then
                o.oAnimState = frame_based_on_level_theme(o.oAnimState, 7, 2)
            else
                o.oAnimState = o.oAnimState + 14
            end

            -- turn to face nearest mario
            local m = nearest_non_cutscene_mario_state_to_object(o)
            if m then
                local tSign = sign(m.pos.x - o.oPosX)
                o.oMoveAngleYaw = tSign * 0x4000
                o.oVelX = tSign * 8
                -- move faster in hard mode
                if gGlobalSyncTable.hardMode then
                    o.oVelX = tSign * 12
                end
            end
        end
    elseif o.oAction == 2 then -- other death
        o.oFaceAnglePitch = 0x8000 -- flip sprite
        cur_obj_become_intangible()
        play_smb_sound_object(o, "kick", true)
        o.oAnimState = 4
        if o.oBehParams == 0 then
            o.oAnimState = frame_based_on_level_theme(o.oAnimState, 7, 2)
        else
            o.oAnimState = o.oAnimState + 14
        end
        enemy_move(o, true)
    elseif o.oAction == 3 then -- off-screen
        cur_obj_hide()
        cur_obj_become_intangible()
        o.oShellCombo = 0
        local homeOnScreen, side = point_is_on_screen(o.oHomeX, BLOCK, 4 * BLOCK)
        if o.oSubAction == 0 and not homeOnScreen then
            o.oPosX, o.oPosY = (o.oHomeX - (1.5 * o.oGroupNumber) * BLOCK), o.oHomeY
            o.oSubAction = 1
        elseif o.oSubAction == 1 and homeOnScreen then
            o.oMoveAngleYaw = -0x4000 * side
            cur_obj_change_action(0)
        end
    end

    if o.oPosY < -16 * PX then
        if o.oAction == 2 then
            obj_mark_for_deletion(o)
        else
            cur_obj_change_action(3)
        end
    end

    o.oInteractStatus = 0
end

id_bhvSmbKoopa = hook_smb_behavior("bhvSmbKoopa", OBJ_LIST_GENACTOR, smb_koopa_init, smb_koopa_loop)

-- piranha plant
---@param o Object
function smb_piranha_plant_init(o)
    o.oFlags = 0
    o.oGravity = 4
    o.oInteractType = INTERACT_DAMAGE
    o.oRenderBehindFloor = 1
    o.hitboxRadius = 5 * PX -- 5 pixels on both sides from center
    o.hitboxHeight = 6 * PX
    o.hitboxDownOffset = -4 * PX
    o.oDeathSound = 0 -- used as "played death sound" flag
    o.oDamageOrCoinValue = 2 -- 200 points by default (not killable anyway though)
    o.oAnimState = 0
    if o.oBehParams == 0 then
        o.oAnimState = frame_based_on_level_theme(o.oAnimState, 2, 3, true)
    else
        o.oAnimState = o.oAnimState + 6
    end
    o.oSMBTimer2 = 0
    if o.oHomeY == 0 then
        cur_obj_set_home_once()
    end
    network_init_object(o, true, {"oHomeX", "oHomeY", "oVelX", "oVelY", "oSMBTimer2"})
end

---@param o Object
function smb_piranha_plant_loop(o)
    cur_obj_unhide()

    if o.oAction == 0 then -- main action
        if not obj_is_on_screen(o, BLOCK, 7.5 * BLOCK) then
            cur_obj_change_action(1)
            return
        end

        cur_obj_become_tangible()
        if o.oSubAction == 0 then -- going down
            o.oPosY = o.oPosY - 8
            if o.oPosY <= o.oHomeY then
                cur_obj_become_intangible()
                cur_obj_hide()
                o.oPosY = o.oHomeY
                o.oSMBTimer2 = o.oSMBTimer2 + 1
                if o.oSMBTimer2 >= 64 then
                    -- check for player, unless this is a red plant
                    local m = nearest_mario_state_to_object_horizontal(o)
                    if o.oBehParams ~= 0 or m == nil or get_horizontal_dist_between_objects(o, m.marioObj) > 0x21 * PX then
                        o.oSMBTimer2 = 0
                        o.oSubAction = 1
                    end
                end
            end
        else -- going up
            o.oPosY = o.oPosY + 8
            if o.oPosY >= o.oHomeY + 2 * BLOCK then
                o.oPosY = o.oHomeY + 2 * BLOCK
                o.oSMBTimer2 = o.oSMBTimer2 + 1
                if o.oSMBTimer2 >= 64 then
                    o.oSMBTimer2 = 0
                    o.oSubAction = 0
                end
            end
        end

        if o.oTimer % 8 == 0 then
            o.oAnimState = (o.oAnimState + 1) % 2
            if o.oBehParams == 0 then
                o.oAnimState = frame_based_on_level_theme(o.oAnimState, 2, 3, true)
            else
                o.oAnimState = o.oAnimState + 6
            end
        end

        if o.oInteractStatus & INT_STATUS_INTERACTED ~= 0 and o.oInteractStatus & INT_STATUS_WAS_ATTACKED ~= 0 then
            play_smb_sound_object(o, "kick")
            obj_mark_for_deletion(o)
        end
    else -- off-screen
        cur_obj_hide()
        cur_obj_become_intangible()
        local homeOnScreen = point_is_on_screen(o.oHomeX, BLOCK, 7.5 * BLOCK)
        if o.oSubAction == 0 and not homeOnScreen then
            o.oPosX, o.oPosY = o.oHomeX, o.oHomeY
            o.oSubAction = 1
        elseif o.oSubAction == 1 and homeOnScreen then
            cur_obj_change_action(0)
        end
    end

    -- hide and become intangible if the warp zone is active
    if warpZoneActive then
        cur_obj_hide()
        cur_obj_become_intangible()
    end

    o.oInteractStatus = 0
end

id_bhvSmbPiranhaPlant = hook_smb_behavior("bhvSmbPiranhaPlant", OBJ_LIST_GENACTOR, smb_piranha_plant_init, smb_piranha_plant_loop)

-- flying paratroopas (both red and green)
---@param o Object
function smb_flying_paratroopa_init(o)
    o.oFlags = OBJ_FLAG_SET_FACE_YAW_TO_MOVE_YAW
    o.oGravity = 4
    o.oInteractType = INTERACT_BOUNCE_TOP2
    o.hitboxRadius = 6 * PX -- 6 pixels on both sides from center
    o.hitboxHeight = 12 * PX
    o.hitboxDownOffset = -3 * PX
    o.oDeathSound = 0 -- used as "played death sound" flag
    o.oDamageOrCoinValue = 3 -- 400 points by default
    o.oBehParams2ndByte = 1 -- so red paratroopas walk off of ledges when demoted (smb1 quirk)
    if o.setHome == 0 then
        o.oAction = 4
        o.oMoveAngleYaw = -0x4000
    end

    o.oAnimState = 2
    if o.oBehParams == 0 then
        o.oAnimState = frame_based_on_level_theme(o.oAnimState, 7, 2)
    else
        o.oAnimState = o.oAnimState + 14
    end
    
    o.oTurnTowardsOnHit = 0
    if o.oHomeY == 0 then
        cur_obj_set_home_once()
    end
    network_init_object(o, true, {"oHomeX", "oHomeY", "oVelX", "oVelY", "globalPlayerIndex"})
end

---@param o Object
function smb_flying_paratroopa_loop(o)
    -- behave as a regular koopa
    if o.oAction < 3 then
        -- stomp sound
        if o.oAction == 0 and o.oPrevAction == 4 and o.oTimer == 1 then
            play_smb_sound_object(o, "stomp")
        end
        o.oTurnTowardsOnHit = 1
        o.oInteractType = INTERACT_BOUNCE_TOP
        return smb_koopa_loop(o)
    end
    
    o.oTurnTowardsOnHit = 0
    o.oInteractType = INTERACT_BOUNCE_TOP2
    cur_obj_unhide()

    if o.oIntangibleToMarioTimer > 0 then
        o.oIntangibleToMarioTimer = o.oIntangibleToMarioTimer - 1
    end

    if o.oAction == 4 then -- main action
        o.oHurtsEnemies = 0
        if not obj_is_on_screen(o, 4 * BLOCK, 4 * BLOCK) then
            cur_obj_change_action(3)
            return
        end

        o.oFaceAnglePitch = 0
        if o.oBehParams == 0 then -- left-right
            -- wavy flight
            if o.oTimer % 128 >= 64 then
                o.oVelY = 4
            else
                o.oVelY = -4
            end

            if o.oSubAction == 0 then -- going left
                o.oVelX = o.oVelX - 4 * SSPX
                if o.oVelX < -0x13 then
                    o.oVelX = -0x13
                    o.oSubAction = 1
                end
            elseif o.oSubAction == 1 then -- going right
                o.oVelX = o.oVelX + 4 * SSPX
                if o.oVelX > 0x13 then
                    o.oVelX = 0x13
                    o.oSubAction = 0
                end
            end
            o.oMoveAngleYaw = 0x4000 * sign(o.oVelX)
        else -- up-down
            o.oVelX = 0
            if o.oSubAction == 0 then -- going down
                o.oVelY = o.oVelY - 3 * SSPX
                if o.oVelY < -PX then
                    o.oVelY = -PX
                    o.oSubAction = 1
                end
            elseif o.oSubAction == 1 then -- going up
                o.oVelY = o.oVelY + 3 * SSPX
                if o.oVelY > PX then
                    o.oVelY = PX
                    o.oSubAction = 0
                end
            end
        end
        cur_obj_become_tangible()
        enemy_move(o, false)

        if o.oTimer % 8 == 0 then
            o.oAnimState = (o.oAnimState%7 + 1) % 2 + 2
            if o.oBehParams == 0 then
                o.oAnimState = frame_based_on_level_theme(o.oAnimState, 7, 2)
            else
                o.oAnimState = o.oAnimState + 14
            end
        end

        if o.oInteractStatus & INT_STATUS_INTERACTED ~= 0 and o.oInteractStatus & INT_STATUS_WAS_ATTACKED ~= 0 then
            if o.oInteractStatus & INT_STATUS_ATTACK_MASK == ATTACK_FROM_ABOVE then
                -- demoted :(
                o.oVelY = -PX
                o.oIntangibleToMarioTimer = 5
                cur_obj_change_action(0)
                o.oPrevAction = 4
            else
                o.oVelY = 4 * PX
                o.oVelX = -o.oVelX
                o.oMoveAngleYaw = -o.oMoveAngleYaw
                cur_obj_change_action(2)
            end
        end
    elseif o.oAction == 3 then -- off-screen
        cur_obj_hide()
        cur_obj_become_intangible()
        o.oShellCombo = 0
        local homeOnScreen, side = point_is_on_screen(o.oHomeX, BLOCK, 4 * BLOCK)
        if o.oSubAction == 0 and not homeOnScreen then
            o.oPosX, o.oPosY = o.oHomeX, o.oHomeY
            o.oVelX, o.oVelY = 0, 0
            o.oSubAction = 1
        elseif o.oSubAction == 1 and homeOnScreen then
            o.oMoveAngleYaw = -0x4000 * side
            cur_obj_change_action(4)
        end
    end

    if o.oPosY < -16 * PX then
        if o.oAction == 2 then
            obj_mark_for_deletion(o)
        else
            cur_obj_change_action(3)
        end
    end

    o.oInteractStatus = 0
end

id_bhvSmbFlyingParatroopa = hook_smb_behavior("bhvSmbFlyingParatroopa", OBJ_LIST_GENACTOR, smb_flying_paratroopa_init, smb_flying_paratroopa_loop)

-- bouncing paratroopas
---@param o Object
function smb_bouncing_paratroopa_init(o)
    o.oFlags = OBJ_FLAG_SET_FACE_YAW_TO_MOVE_YAW
    o.oGravity = 2
    o.oInteractType = INTERACT_BOUNCE_TOP2
    o.hitboxRadius = 6 * PX -- 6 pixels on both sides from center
    o.hitboxHeight = 12 * PX
    o.hitboxDownOffset = -3 * PX
    o.oDeathSound = 0 -- used as "played death sound" flag
    o.oDamageOrCoinValue = 3 -- 400 points by default
    o.oBehParams2ndByte = 1 -- so red paratroopas walk off of ledges when demoted (smb1 quirk)
    if o.setHome == 0 then
        o.oAction = 4
        o.oMoveAngleYaw = -0x4000
    end

    o.oAnimState = 2
    if o.oBehParams == 0 then
        o.oAnimState = frame_based_on_level_theme(o.oAnimState, 7, 2)
    else
        o.oAnimState = o.oAnimState + 14 -- red bouncing paratroopas are also implemented for no reason
    end
    
    o.oTurnTowardsOnHit = 1
    if o.oHomeY == 0 then
        cur_obj_set_home_once()
    end
    network_init_object(o, true, {"oHomeX", "oHomeY", "oVelX", "oVelY", "globalPlayerIndex"})
end

---@param o Object
function smb_bouncing_paratroopa_loop(o)
    -- behave as a regular koopa
    if o.oAction < 3 then
        -- stomp sound
        if o.oAction == 0 and o.oPrevAction == 4 and o.oTimer == 1 then
            play_smb_sound_object(o, "stomp")
        end
        o.oGravity = 4
        o.oTurnTowardsOnHit = 1
        o.oInteractType = INTERACT_BOUNCE_TOP
        return smb_koopa_loop(o)
    end
    
    o.oTurnTowardsOnHit = 1
    o.oGravity = 2
    o.oInteractType = INTERACT_BOUNCE_TOP2
    cur_obj_unhide()

    if o.oIntangibleToMarioTimer > 0 then
        o.oIntangibleToMarioTimer = o.oIntangibleToMarioTimer - 1
    end

    if o.oAction == 4 then -- main action
        o.oHurtsEnemies = 0
        if not obj_is_on_screen(o, 4 * BLOCK, 4 * BLOCK) then
            cur_obj_change_action(3)
            return
        end

        o.oFaceAnglePitch = 0
        o.oVelX = sign(o.oMoveAngleYaw) * 8
        cur_obj_become_tangible()
        local landed, hitWall = enemy_move_and_collide(o)
        if landed then
            o.oVelY = 0x38
        end
        if hitWall then
            o.oMoveAngleYaw = -o.oMoveAngleYaw
        end

        if o.oTimer % 8 == 0 then
            o.oAnimState = (o.oAnimState%7 + 1) % 2 + 2
            if o.oBehParams == 0 then
                o.oAnimState = frame_based_on_level_theme(o.oAnimState, 7, 2)
            else
                o.oAnimState = o.oAnimState + 14
            end
        end

        if o.oInteractStatus & INT_STATUS_INTERACTED ~= 0 and o.oInteractStatus & INT_STATUS_WAS_ATTACKED ~= 0 then
            if o.oInteractStatus & INT_STATUS_ATTACK_MASK == ATTACK_FROM_ABOVE then
                -- demoted :(
                o.oVelY = -PX
                o.oIntangibleToMarioTimer = 5
                cur_obj_change_action(0)
                o.oPrevAction = 4
            else
                o.oVelY = 4 * PX
                o.oVelX = -o.oVelX
                o.oMoveAngleYaw = -o.oMoveAngleYaw
                cur_obj_change_action(2)
            end
        end
    elseif o.oAction == 3 then -- off-screen
        cur_obj_hide()
        cur_obj_become_intangible()
        o.oShellCombo = 0
        local homeOnScreen, side = point_is_on_screen(o.oHomeX, BLOCK, 4 * BLOCK)
        if o.oSubAction == 0 and not homeOnScreen then
            o.oPosX, o.oPosY = o.oHomeX, o.oHomeY
            o.oVelX, o.oVelY = 0, 0
            o.oSubAction = 1
        elseif o.oSubAction == 1 and homeOnScreen then
            o.oMoveAngleYaw = -0x4000 * side
            cur_obj_change_action(4)
        end
    end

    if o.oPosY < -16 * PX then
        if o.oAction == 2 then
            obj_mark_for_deletion(o)
        else
            cur_obj_change_action(3)
        end
    end

    o.oInteractStatus = 0
end

id_bhvSmbBouncingParatroopa = hook_smb_behavior("bhvSmbBouncingParatroopa", OBJ_LIST_GENACTOR, smb_bouncing_paratroopa_init, smb_bouncing_paratroopa_loop)

-- spiny
---@param o Object
function smb_spiny_init(o)
    o.oFlags = OBJ_FLAG_SET_FACE_YAW_TO_MOVE_YAW
    o.oGravity = 4
    o.oInteractType = INTERACT_DAMAGE
    o.hitboxRadius = 5 * PX -- 5 pixels on both sides from center
    o.hitboxHeight = 6 * PX
    o.hitboxDownOffset = -4 * PX
    o.oDeathSound = 0 -- used as "played death sound" flag
    o.oDamageOrCoinValue = 1 -- 100 points by default (unused)

    if o.oAction == 1 then
        o.oAnimState = 2 -- thrown by lakitu
    else
        o.oAnimState = 0
        o.oMoveAngleYaw = -0x4000
    end

    o.oTurnTowardsOnHit = 1
    if o.oHomeY == 0 then
        cur_obj_set_home_once()
    end
    network_init_object(o, true, {"oHomeX", "oHomeY", "oVelX", "oVelY", "oGroupNumber"})
end

---@param o Object
function smb_spiny_loop(o)
    cur_obj_unhide()
    o.oGravity = 4
    o.oRenderBehindFloor = 0
    o.oTurnTowardsOnHit = 1

    local m = nearest_non_cutscene_mario_state_to_object(o)

    if o.oAction == 0 then -- main action
        if not obj_is_on_screen(o, 4 * BLOCK, 4 * BLOCK) then
            cur_obj_change_action(3)
            return
        end

        o.oFaceAnglePitch = 0
        o.oVelX = sign(o.oMoveAngleYaw) * 8
        -- move faster in hard mode
        if gGlobalSyncTable.hardMode then
            o.oVelX = sign(o.oMoveAngleYaw) * 12
        end
        cur_obj_become_tangible()
        local falling = (o.oVelY < 0)
        local landed, hitWall = enemy_move_and_collide(o)

        if hitWall then
            o.oMoveAngleYaw = -o.oMoveAngleYaw
        end
        if landed and falling and m then
            -- face nearest mario
            o.oMoveAngleYaw = sign(m.pos.x - o.oPosX) * 0x4000
        end

        if o.oTimer % 8 == 0 then
            o.oAnimState = (o.oAnimState + 1) % 2
        end

        if o.oInteractStatus & INT_STATUS_INTERACTED ~= 0 and o.oInteractStatus & INT_STATUS_WAS_ATTACKED ~= 0 then
            o.oVelY = 4 * PX
            o.oVelX = -o.oVelX
            o.oMoveAngleYaw = -o.oMoveAngleYaw
            cur_obj_change_action(2)
        end
    elseif o.oAction == 1 then -- thrown by lakitu
        if not obj_is_on_screen(o, 4 * BLOCK, 4 * BLOCK) then
            cur_obj_change_action(3)
            return
        end

        if o.oTimer <= 5 then
            o.oRenderBehindFloor = 1
        end
        cur_obj_become_tangible()
        
        o.oGravity = 2
        o.oTurnTowardsOnHit = 0
        local landed, hitWall = enemy_move_and_collide(o)
        if hitWall then
            o.oMoveAngleYaw = -o.oMoveAngleYaw
            o.oVelX = -o.oVelX
        end
        if landed then
            if m then
                -- face nearest mario
                o.oMoveAngleYaw = sign(m.pos.x - o.oPosX) * 0x4000
            end
            o.oVelX = 0
            o.oAnimState = 0
            cur_obj_change_action(0)
            return
        end

        if o.oTimer % 8 == 0 then
            o.oAnimState = (o.oAnimState + 1) % 2 + 2
        end

        if o.oInteractStatus & INT_STATUS_INTERACTED ~= 0 and o.oInteractStatus & INT_STATUS_WAS_ATTACKED ~= 0 then
            o.oVelY = 4 * PX
            o.oVelX = -o.oVelX
            o.oMoveAngleYaw = -o.oMoveAngleYaw
            cur_obj_change_action(2)
        end
    elseif o.oAction == 2 then -- other death
        o.oFaceAnglePitch = 0x8000 -- flip sprite
        cur_obj_become_intangible()
        play_smb_sound_object(o, "kick", true)
        o.oAnimState = 1
        enemy_move(o, true)
    elseif o.oAction == 3 then -- off-screen
        if o.oBehParams ~= 0 then
            -- delete spinies thrown by lakitu
            obj_mark_for_deletion(o)
            return
        end
        cur_obj_hide()
        cur_obj_become_intangible()
        local homeOnScreen, side = point_is_on_screen(o.oHomeX, BLOCK, 4 * BLOCK)
        if o.oSubAction == 0 and not homeOnScreen then
            o.oPosX, o.oPosY = (o.oHomeX - (1.5 * o.oGroupNumber) * BLOCK), o.oHomeY
            o.oSubAction = 1
        elseif o.oSubAction == 1 and homeOnScreen then
            o.oMoveAngleYaw = -0x4000 * side
            cur_obj_change_action(0)
        end
    end

    if o.oPosY < -16 * PX then
        if o.oAction == 2 then
            obj_mark_for_deletion(o)
        else
            cur_obj_change_action(3)
        end
    end

    o.oInteractStatus = 0
end

id_bhvSmbSpiny = hook_smb_behavior("bhvSmbSpiny", OBJ_LIST_GENACTOR, smb_spiny_init, smb_spiny_loop)

-- lakitu
---@param o Object
function smb_lakitu_init(o)
    o.oFlags = 0
    o.oGravity = 2
    o.oInteractType = INTERACT_BOUNCE_TOP2
    o.hitboxRadius = 6 * PX -- 6 pixels on both sides from center
    o.hitboxHeight = 12 * PX
    o.hitboxDownOffset = -3 * PX
    o.oDeathSound = 0 -- used as "played death sound" flag
    o.oDamageOrCoinValue = 5 -- 500 points by default
    o.oAnimState = 0
    
    if o.setHome == 0 then
        o.oMoveAngleYaw = -0x4000
        o.setHome = 1
    end
    if o.oHomeY == 0 then
        cur_obj_set_home_once()
    end
    network_init_object(o, true, {"oHomeX", "oHomeY", "oVelX", "oVelY", "oSMBTimer2"})
end

---@param o Object
function smb_lakitu_loop(o)
    cur_obj_unhide()

    local m = get_rightmost_mario_state(true)

    if o.oAction == 0 then -- main action
        o.oVelY = 0 
        local goalPosX = o.oHomeX
        local leftScreenPoint = (m and gSMBStates[m.playerIndex].cameraX * PX) or 0
        local rightScreenPoint = leftScreenPoint + NES_WIDTH * PX
        -- don't include marios that have yet to make it to lakitu
        if rightScreenPoint < goalPosX then
            m = nil
        elseif m then
            goalPosX = m.pos.x
        end

        -- don't run if we're at home and there's no valid target
        if not m and o.oPosX == o.oHomeX then
            o.oVelX = 0
            o.oMoveAngleYaw = -0x4000
            return
        end

        -- this movement code sucked to program. Definitely not SMB accurate. I'm tired boss
        local goalDiff = (goalPosX - o.oPosX)
        local maxGoalDist = 0x3C * PX
        local minSpeed = 0x05
        local maxSpeed = 0x14

        -- chance speed based on distance from goal
        -- speed is fastest when close to goal and slowest when far from goal
        local t = clampf((maxGoalDist - math.abs(goalDiff)) / maxGoalDist, 0, 1)
        local targetSpeed = minSpeed + (maxSpeed - minSpeed) * t
        targetSpeed = math.ceil(targetSpeed)

        if m then
            if goalDiff > maxGoalDist or math.abs(goalDiff) > 16 * BLOCK then
                -- don't allow going too far left
                if o.oVelX < 0 and goalDiff > maxGoalDist then
                    o.oVelX = 0
                end

                o.oMoveAngleYaw = sign(goalDiff) * 0x4000
                if obj_is_on_screen(o, 0, 0) then
                    -- if we're far away from our goal, speed towards it
                    o.oVelX = approach_f32(o.oVelX, MAX_RUN_VEL * 2 * sign(o.oMoveAngleYaw), 1, 1)
                elseif m.pos.x > o.oPosX then -- if lakitu is not on screen, teleport to screen side
                    o.oVelX = 0
                    o.oPosX = leftScreenPoint - 8 * PX
                else -- right side
                    o.oVelX = 0
                    o.oPosX = rightScreenPoint + 8 * PX
                end
            else
                local absMSpeed = math.abs(m.vel.x)
                -- ALT MOVEMENT PATTERN APPLIES WHEN:
                -- We're traveling in a valid direction (only right, unless backwards is enabled) AND
                -- Mario is moving at least as fast as walking speed AND
                -- Lakitu is moving in the same direction as mario
                if (gGlobalSyncTable.allowBackwards or m.vel.x > 0) and absMSpeed >= MAX_WALK_VEL - 1 and sign(o.oVelX) == sign(m.vel.x) then
                    -- repurpose current target speed as additional speed, and set target to mario's speed
                    local extraSpeed = targetSpeed
                    targetSpeed = math.max(0, absMSpeed + PX)
                    targetSpeed = targetSpeed + extraSpeed

                    -- Move Lakitu back if he is getting too far ahead
                    if (sign(goalDiff) ~= sign(m.vel.x) and math.abs(goalDiff) >= maxGoalDist) then
                        o.oVelX = (absMSpeed - 10) * sign(o.oMoveAngleYaw)
                    end
                    
                    o.oVelX = approach_f32(o.oVelX, targetSpeed * sign(o.oMoveAngleYaw), 1, 1)
                else -- loiter above mario
                    o.oVelX = sign(o.oMoveAngleYaw) * targetSpeed
                    if math.abs(goalDiff) > maxGoalDist and sign(o.oMoveAngleYaw) ~= sign(goalDiff) then
                        -- turn around if moving out of range
                        o.oMoveAngleYaw = -o.oMoveAngleYaw
                    end
                end
            end
        else -- move towards home
            if math.abs(goalDiff) > maxGoalDist and sign(o.oMoveAngleYaw) ~= sign(goalDiff) then
                -- turn around if moving out of range
                o.oMoveAngleYaw = -o.oMoveAngleYaw
            end

            if not obj_is_on_screen(o, 0, 0) then
                -- if no mario is using lakitu and he is off-screen, warp to home
                o.oVelX = 0
                o.oPosX = goalPosX
            end
            
            o.oVelX = sign(o.oMoveAngleYaw) * targetSpeed
        end

        o.oFaceAnglePitch = 0
        if o.oVelX ~= 0 then
            o.oFaceAngleYaw = sign(o.oVelX) * -0x4000
        end
        cur_obj_become_tangible()
        enemy_move(o, false)

        -- throw spinies
        if not m then
            o.oSubAction = 0 -- cancel throw if no mario is a target
            o.oAnimState = 0
        end
        if o.oSMBTimer2 ~= 0 then
            o.oSMBTimer2 = o.oSMBTimer2 - 1
        else
            if not m then
                o.oSMBTimer2 = 0x80
            elseif o.oSubAction == 0 then
                -- max of 3 on screen (note this interacts weirdly with syncing but oh well)
                local spinyCount = obj_count_on_screen(id_bhvSmbSpiny, 0, 0)
                if spinyCount >= 3 then
                    o.oSMBTimer2 = 0x80
                else
                    o.oSMBTimer2 = 0x10
                    o.oSubAction = 1
                end
            else
                o.oSMBTimer2 = 0x80
                o.oSubAction = 0
                if m.playerIndex == 0 then
                    spawn_object_no_rotate(id_bhvSmbSpiny, E_MODEL_NONE, o.oPosX, o.oPosY, 0, true, function(spiny)
                        spiny.oVelY = 3 * PX
                        if gGlobalSyncTable.altSpinyEggBehavior then
                            -- also not quite accurate
                            spiny.oVelX = math.random(-0x13, 0x13)
                            if math.abs(m.vel.x) >= 8 then
                                spiny.oVelX = spiny.oVelX + m.vel.x
                            end
                            spiny.oMoveAngleYaw = sign(o.oVelX) * 0x4000
                        else
                            spiny.oVelX = 0
                            spiny.oMoveAngleYaw = 0x4000
                        end
                        spiny.oRenderBehindFloor = 1
                        spiny.oAction = 1
                        spiny.oBehParams = 1
                    end)
                end
            end
        end
        if o.oSubAction == 0 then
            o.oAnimState = 0
        else
            o.oAnimState = 1
        end

        if o.oInteractStatus & INT_STATUS_INTERACTED ~= 0 and o.oInteractStatus & INT_STATUS_WAS_ATTACKED ~= 0 then
            o.oVelX = 0
            if o.oInteractStatus & INT_STATUS_ATTACK_MASK == ATTACK_FROM_ABOVE then
                o.oVelY = -PX
                cur_obj_change_action(1)
            else
                o.oVelY = PX
                cur_obj_change_action(2)
            end
        end
    elseif o.oAction == 1 then -- stomp death
        o.oFaceAnglePitch = 0x8000 -- flip sprite
        cur_obj_become_intangible()
        play_smb_sound_object(o, "stomp", true)
        o.oAnimState = 0
        enemy_move(o, true)
    elseif o.oAction == 2 then -- other death
        o.oFaceAnglePitch = 0x8000 -- flip sprite
        cur_obj_become_intangible()
        play_smb_sound_object(o, "kick", true)
        o.oAnimState = 0
        enemy_move(o, true)
    elseif o.oAction == 3 then -- despawned
        cur_obj_hide()
        cur_obj_become_intangible()
        if o.oTimer > 16 * 60 then
            o.oPosY = o.oHomeY
            -- start at far right of valid player's screen
            local leftScreenPoint = (m and gSMBStates[m.playerIndex].cameraX * PX) or 0
            local rightScreenPoint = leftScreenPoint + NES_WIDTH * PX
            o.oPosX = rightScreenPoint + 8 * PX
            cur_obj_change_action(0)
        end
    end

    if o.oAction ~= 3 and o.oPosY < -16 * PX then
        cur_obj_change_action(3)
    end

    o.oInteractStatus = 0
end

id_bhvSmbLakitu = hook_smb_behavior("bhvSmbLakitu", OBJ_LIST_GENACTOR, smb_lakitu_init, smb_lakitu_loop)

-- buzzy beetles (they're pretty much just koopas)
---@param o Object
function smb_buzzy_beetle_init(o)
    o.oFlags = OBJ_FLAG_SET_FACE_YAW_TO_MOVE_YAW
    o.oGravity = 4
    o.oMoveAngleYaw = -0x4000
    o.oInteractType = INTERACT_BOUNCE_TOP
    o.hitboxRadius = 6 * PX -- 6 pixels on both sides from center
    o.hitboxHeight = 12 * PX
    o.hitboxDownOffset = -3 * PX
    o.oDeathSound = 0 -- used as "played death sound" flag
    o.oDamageOrCoinValue = 1 -- 100 points by default
    o.oImmuneToFire = 1 -- signature ability: immune to fireballs
    
    o.oAnimState = frame_based_on_level_theme(0, 4, 3)
    
    o.oTurnTowardsOnHit = 1
    if o.oHomeY == 0 then
        cur_obj_set_home_once()
    end
    network_init_object(o, true, {"oHomeX", "oHomeY", "oVelX", "oVelY", "globalPlayerIndex", "oGroupNumber"})
end

---@param o Object
function smb_buzzy_beetle_loop(o)
    cur_obj_unhide()

    if o.oIntangibleToMarioTimer > 0 then
        o.oIntangibleToMarioTimer = o.oIntangibleToMarioTimer - 1
    end

    if o.oAction == 0 then
        o.oHurtsEnemies = 0
        o.oTurnTowardsOnHit = 1
        if not obj_is_on_screen(o, 4 * BLOCK, 4 * BLOCK) then
            cur_obj_change_action(3)
            return
        end

        o.oFaceAnglePitch = 0
        o.oInteractType = INTERACT_BOUNCE_TOP
        o.oVelX = sign(o.oMoveAngleYaw) * 8
        -- move faster in hard mode
        if gGlobalSyncTable.hardMode then
            o.oVelX = sign(o.oMoveAngleYaw) * 12
        end
        cur_obj_become_tangible()
        local falling = (o.oVelY < 0)
        local landed, hitWall = enemy_move_and_collide(o, (o.oBehParams ~= 0 and o.oBehParams2ndByte == 0))

        if hitWall then
            o.oMoveAngleYaw = -o.oMoveAngleYaw
        end
        if landed and falling then
            local m = nearest_non_cutscene_mario_state_to_object(o)
            -- face nearest mario
            if m then
                o.oMoveAngleYaw = sign(m.pos.x - o.oPosX) * 0x4000
            end
        end

        if o.oTimer % 8 == 0 then
            o.oAnimState = frame_based_on_level_theme((o.oAnimState + 1) % 2, 4, 3)
        end

        if o.oInteractStatus & INT_STATUS_INTERACTED ~= 0 and o.oInteractStatus & INT_STATUS_WAS_ATTACKED ~= 0 then
            if o.oInteractStatus & INT_STATUS_ATTACK_MASK == ATTACK_FROM_ABOVE then
                o.oVelX = 0
                o.oIntangibleToMarioTimer = 5
                o.oInteractType = INTERACT_KOOPA_SHELL
                cur_obj_change_action(1)
            elseif o.oInteractStatus & INT_STATUS_ATTACK_MASK == ATTACK_FROM_BELOW then
                o.oVelY = 4 * PX
                o.oFaceAnglePitch = 0x8000
                o.oInteractType = INTERACT_KOOPA_SHELL
                cur_obj_change_action(1)
            else
                o.oVelY = 4 * PX
                o.oVelX = -o.oVelX
                o.oMoveAngleYaw = -o.oMoveAngleYaw
                cur_obj_change_action(2)
            end
        end
    elseif o.oAction == 1 then -- in shell
        if not obj_is_on_screen(o, 4 * BLOCK, 4 * BLOCK) then
            cur_obj_change_action(3)
            return
        end

        o.oTurnTowardsOnHit = 0

        cur_obj_become_tangible()
        if o.oTimer == 1 then
            play_smb_sound_object(o, "stomp")
        end
        
        if o.oSubAction == 0 then -- idle
            o.oInteractType = INTERACT_KOOPA_SHELL
            o.oHurtsEnemies = 0
        else -- kicked
            o.oInteractType = INTERACT_BOUNCE_TOP
            o.oHurtsEnemies = 1
            o.oTimer = 2
        end
        local landed, hitWall = enemy_move_and_collide(o)
        if landed then
            if o.oSubAction == 0 then
                o.oVelX = 0
            else
                o.oVelX = sign(o.oMoveAngleYaw) * 0x30
            end
        elseif o.oSubAction ~= 0 then
            o.oVelX = sign(o.oMoveAngleYaw) * 0x18
        end

        if hitWall and o.oSubAction ~= 0 then
            o.oMoveAngleYaw = -o.oMoveAngleYaw
            o.oVelX = math.abs(o.oVelX) * sign(o.oMoveAngleYaw)
            play_smb_sound_object(o, "bump")
            o.oIntangibleToMarioTimer = 0
        end

        o.oAnimState = frame_based_on_level_theme(2, 4, 3)

        if o.oInteractStatus & INT_STATUS_INTERACTED ~= 0 then
            if o.oInteractStatus & INT_STATUS_WAS_ATTACKED ~= 0 then
                if o.oInteractStatus & INT_STATUS_ATTACK_MASK == ATTACK_FROM_ABOVE then
                    if o.oSubAction ~= 0 then
                        o.oVelX = 0
                        o.oIntangibleToMarioTimer = 5
                        o.oInteractType = INTERACT_KOOPA_SHELL
                        cur_obj_change_action(1)
                    end
                elseif o.oInteractStatus & INT_STATUS_ATTACK_MASK == ATTACK_FROM_BELOW then
                    o.oVelY = 4 * PX
                    o.oFaceAnglePitch = 0x8000
                    o.oInteractType = INTERACT_KOOPA_SHELL
                    cur_obj_change_action(1)
                elseif o.oInteractStatus & INT_STATUS_ATTACK_MASK == ATTACK_KICK_OR_TRIP then
                    if o.oSubAction == 0 then
                        o.oIntangibleToMarioTimer = 20
                        o.oSubAction = 1
                        play_smb_sound_object(o, "kick")
                    end
                else
                    o.oVelY = 4 * PX
                    o.oMoveAngleYaw = -o.oMoveAngleYaw
                    o.oVelX = sign(o.oMoveAngleYaw) * 8
                    cur_obj_change_action(2)
                end
                o.oInteractStatus = 0
                return
            end
        end

        if o.oTimer >= 16 * 21 then
            cur_obj_change_action(0)
            o.oAnimState = frame_based_on_level_theme(0, 4, 3)

            -- turn to face nearest mario
            local m = nearest_non_cutscene_mario_state_to_object(o)
            if m then
                local tSign = sign(m.pos.x - o.oPosX)
                o.oMoveAngleYaw = tSign * 0x4000
                o.oVelX = tSign * 8
                -- move faster in hard mode
                if gGlobalSyncTable.hardMode then
                    o.oVelX = tSign * 12
                end
            end
        end
    elseif o.oAction == 2 then -- other death
        o.oFaceAnglePitch = 0x8000 -- flip sprite
        cur_obj_become_intangible()
        play_smb_sound_object(o, "kick", true)
        o.oAnimState = frame_based_on_level_theme(2, 4, 3)
        enemy_move(o, true)
    elseif o.oAction == 3 then -- off-screen
        cur_obj_hide()
        cur_obj_become_intangible()
        o.oShellCombo = 0
        local homeOnScreen, side = point_is_on_screen(o.oHomeX, BLOCK, 4 * BLOCK)
        if o.oSubAction == 0 and not homeOnScreen then
            o.oPosX, o.oPosY = (o.oHomeX - (1.5 * o.oGroupNumber) * BLOCK), o.oHomeY
            o.oSubAction = 1
        elseif o.oSubAction == 1 and homeOnScreen then
            o.oMoveAngleYaw = -0x4000 * side
            cur_obj_change_action(0)
        end
    end

    if o.oPosY < -16 * PX then
        if o.oAction == 2 then
            obj_mark_for_deletion(o)
        else
            cur_obj_change_action(3)
        end
    end

    o.oInteractStatus = 0
end

id_bhvSmbBuzzyBeetle = hook_smb_behavior("bhvSmbBuzzyBeetle", OBJ_LIST_GENACTOR, smb_buzzy_beetle_init, smb_buzzy_beetle_loop)

-- podoboos (lava bubbles)
---@param o Object
function smb_podoboo_init(o)
    o.oFlags = 0
    o.oGravity = 2.8
    o.oInteractType = INTERACT_DAMAGE
    o.hitboxRadius = 5 * PX -- 5 pixels on both sides from center
    o.hitboxHeight = 6 * PX
    o.hitboxDownOffset = -4 * PX
    o.oDamageOrCoinValue = 2 -- 200 points by default (not killable anyway though)
    o.oAnimState = 0
    o.oSMBTimer2 = 0
    o.oSkipObjectInteraction = 1
    if o.setHome == 0 then
        o.oPosY = -16 * PX -- always start at the bottom of the screen
        cur_obj_set_home_once()
    end
    network_init_object(o, true, {"oHomeX", "oHomeY", "oVelX", "oVelY", "oSMBTimer2"})
end

---@param o Object
function smb_podoboo_loop(o)
    cur_obj_unhide()

    -- don't render if we're doing the cutscene
    local localInCutscene = (gMarioStates[0].action == ACT_SMB_HIT_AXE or gMarioStates[0].action == ACT_SMB_WALK_CUTSCENE)

    if o.oAction == 0 then -- waiting action
        if not obj_is_on_screen(o, 2 * BLOCK, 2 * BLOCK) then
            cur_obj_change_action(2)
            return
        end
        
        if o.oSMBTimer2 ~= 0 then
            o.oSMBTimer2 = o.oSMBTimer2 - 1
        else
            o.oSMBTimer2 = 0x07 * 21 -- should only happen on init
        end
        if o.oSMBTimer2 == 0 then
            o.oVelY = 7 * PX + 4 -- in SMB, the height ranges from 7-8 pixels, so just use a bit below average here
            o.oSMBTimer2 = 0x0F * 21
            if o.oTimer % 4 == 0 then -- "randomly" jump sooner (we can't used position-based, since it always starts at the same position)
                o.oSMBTimer2 = 0x07 * 21
            end
            o.oPosX = o.oHomeX
            o.oPosY = o.oHomeY
            cur_obj_change_action(1)
        end

        cur_obj_hide()
        cur_obj_become_intangible()
    elseif o.oAction == 1 then -- active action
        if not obj_is_on_screen(o, 12 * BLOCK, 12 * BLOCK) then
            cur_obj_change_action(2)
            return
        end

        cur_obj_become_tangible()

        if o.oVelY > 0 then
            o.oFaceAnglePitch = 0
        else
            o.oFaceAnglePitch = 0x8000
        end
        enemy_move(o, true)

        if o.oSMBTimer2 ~= 0 then
            o.oSMBTimer2 = o.oSMBTimer2 - 1
        end
        if o.oPosY <= o.oHomeY then
            cur_obj_change_action(0)
        end

        -- probably never gonna get called but let's just be safe
        if o.oInteractStatus & INT_STATUS_INTERACTED ~= 0 and o.oInteractStatus & INT_STATUS_WAS_ATTACKED ~= 0 then
            play_smb_sound_object(o, "kick")
            obj_mark_for_deletion(o)
        end
    else -- off-screen
        cur_obj_hide()
        cur_obj_become_intangible()
        local homeOnScreen = point_is_on_screen(o.oHomeX, 2 * BLOCK, 2 * BLOCK)
        if o.oSubAction == 0 and not homeOnScreen then
            o.oPosX, o.oPosY = o.oHomeX, o.oHomeY
            o.oSubAction = 1
        elseif o.oSubAction == 1 and homeOnScreen then
            cur_obj_change_action(0)
        end
    end
    
    if localInCutscene then
        cur_obj_hide()
        cur_obj_become_intangible()
    end

    o.oInteractStatus = 0
end

id_bhvSmbPodoboo = hook_smb_behavior("bhvSmbPodoboo", OBJ_LIST_GENACTOR, smb_podoboo_init, smb_podoboo_loop)

-- blooper
---@param o Object
function smb_blooper_init(o)
    o.oFlags = OBJ_FLAG_SET_FACE_YAW_TO_MOVE_YAW
    o.oGravity = 1
    o.oInteractType = INTERACT_BOUNCE_TOP2
    o.hitboxRadius = 5 * PX -- 5 pixels on both sides from center
    o.hitboxHeight = 6 * PX
    o.hitboxDownOffset = -8 * PX
    o.oDamageOrCoinValue = 6 -- 1000 points when stomped (not possible in SMB though)
    o.oAnimState = 0
    cur_obj_set_home_once()
    network_init_object(o, true, {"oVelX", "oVelY"})
end

---@param o Object
function smb_blooper_loop(o)
    cur_obj_unhide()

    if o.oAction == 0 then -- swimming
        if not obj_is_on_screen(o, 4 * BLOCK, 4 * BLOCK) then
            cur_obj_change_action(3)
            return
        end

        local m = nearest_non_cutscene_mario_state_to_object(o)
        local marioY = 2 * BLOCK
        if m then
            marioY = m.pos.y
        end

        o.oFaceAnglePitch = 0
        if o.oSubAction == 0 then -- swimming downward
            o.oVelX = 0
            o.oVelY = -8

            if o.oTimer < 21 then
                o.oAnimState = 0
            elseif o.oTimer < 42 then
                -- do other frame for short time after swimming
                o.oAnimState = 1
            else
                o.oAnimState = 0
                -- swim up once we reach mario
                if o.oPosY <= marioY + BLOCK then
                    o.oVelY = 0
                    o.oSubAction = 1
                end
            end
        elseif o.oSubAction == 1 then -- swimming upward start
            o.oAnimState = 0
            if o.oTimer % 8 == 0 then
                o.oVelY = o.oVelY + PX
            end
            o.oVelX = o.oVelY * sign(o.oMoveAngleYaw)
            if o.oVelY >= 2 * PX then
                o.oSubAction = 2
            end
        elseif o.oSubAction == 2 then -- swimming upward end
            o.oAnimState = 0
            if o.oTimer % 8 == 0 then
                o.oVelY = o.oVelY - PX
            end
            o.oVelX = o.oVelY * sign(o.oMoveAngleYaw)
            if o.oVelY <= 0 then
                o.oSubAction = 0
                o.oTimer = 0
            end
        end

        -- handle direction
        local u8 = position_based_random_u16() & 0xFF -- get random u8 by taking lower part of u16
        local blooperMoveFlags = 63 -- binary number
        if do_hard_mode_objs() then -- change direction more often in hard mode
            blooperMoveFlags = 3
        end
        -- if no bits in the u8 are set, change direction
        if u8 & blooperMoveFlags == 0 then
            -- In SMB, the bloopers behave differently depending on what slot they are loaded in.
            -- This sort of replicates that behavior
            if not m then
                -- failsafe if there are no marios found; just flip direction
                o.oMoveAngleYaw = -o.oMoveAngleYaw
            elseif o.oSyncID % 2 == 0 then
                -- set direction towards player
                o.oMoveAngleYaw = sign(m.pos.x - o.oPosX) * 0x4000
            else
                -- set direction to direction player is facing
                o.oMoveAngleYaw = m.faceAngle.y
            end
        end

        cur_obj_become_tangible()
        enemy_move(o, false)
        -- don't go above screen
        if o.oPosY > NES_HEIGHT * PX - 4 * BLOCK then
            o.oPosY = NES_HEIGHT * PX - 4 * BLOCK
        end

        if o.oInteractStatus & INT_STATUS_INTERACTED ~= 0 and o.oInteractStatus & INT_STATUS_WAS_ATTACKED ~= 0 then
            o.oVelX = 0
            if o.oInteractStatus & INT_STATUS_ATTACK_MASK == ATTACK_FROM_ABOVE then
                o.oVelY = -PX
                cur_obj_change_action(1)
            else
                o.oVelY = PX
                cur_obj_change_action(2)
            end
        end
    elseif o.oAction == 1 then -- stomp death (unused in SMB1)
        o.oFaceAnglePitch = 0x8000 -- flip sprite
        cur_obj_become_intangible()
        play_smb_sound_object(o, "stomp", true)
        o.oAnimState = 1
        enemy_move(o, true)
    elseif o.oAction == 2 then -- other death
        o.oFaceAnglePitch = 0x8000 -- flip sprite
        cur_obj_become_intangible()
        play_smb_sound_object(o, "kick", true)
        o.oAnimState = 1
        enemy_move(o, true)
    elseif o.oAction == 3 then -- off-screen
        cur_obj_hide()
        cur_obj_become_intangible()
        local homeOnScreen, side = point_is_on_screen(o.oHomeX, 4 * BLOCK, 4 * BLOCK)
        if o.oSubAction == 0 and not homeOnScreen then
            o.oPosX, o.oPosY = o.oHomeX, o.oHomeY
            o.oSubAction = 1
        elseif o.oSubAction == 1 and homeOnScreen then
            o.oMoveAngleYaw = -0x4000 * side
            cur_obj_change_action(0)
        end
    end

    if o.oPosY < -16 * PX then
        if o.oAction == 1 or o.oAction == 2 then
            obj_mark_for_deletion(o)
        else
            cur_obj_change_action(3)
        end
    end

    o.oInteractStatus = 0
end

id_bhvSmbBlooper = hook_smb_behavior("bhvSmbBlooper", OBJ_LIST_GENACTOR, smb_blooper_init, smb_blooper_loop)

-- bullet bill
---@param o Object
function smb_bullet_bill_init(o)
    o.oFlags = OBJ_FLAG_SET_FACE_YAW_TO_MOVE_YAW
    o.oFaceAngleYaw = o.oMoveAngleYaw
    o.oGravity = 2
    o.oInteractType = INTERACT_BOUNCE_TOP2
    o.hitboxRadius = 5 * PX -- 5 pixels on both sides from center
    o.hitboxHeight = 6 * PX
    o.hitboxDownOffset = -2 * PX
    o.oDamageOrCoinValue = 2 -- 200 points (unused?)
    o.oAnimState = 0 -- no other frames
    o.oImmuneToFire = 1
    cur_obj_set_home_once()
    network_init_object(o, true, {"oVelX", "oVelY"})
end

---@param o Object
function smb_bullet_bill_loop(o)
    cur_obj_unhide()
    o.oRenderBehindFloor = 0

    if o.oAction == 0 then -- main action
        if not obj_is_on_screen(o, 4 * BLOCK, 4 * BLOCK) then
            obj_mark_for_deletion(o)
            return
        end

        if o.oTimer == 1 then
            play_smb_sound_object(o, "boom")
        end

        -- when fired from a cannon
        if o.oBehParams == 0 and o.oTimer <= 10 then
            o.oRenderBehindFloor = 1
        end

        o.oVelX = 24 * sign(o.oMoveAngleYaw)
        o.oVelY = 0
        cur_obj_become_tangible()
        enemy_move(o, false)

        if o.oInteractStatus & INT_STATUS_INTERACTED ~= 0 and o.oInteractStatus & INT_STATUS_WAS_ATTACKED ~= 0 then
            o.oVelX = 0
            if o.oInteractStatus & INT_STATUS_ATTACK_MASK == ATTACK_FROM_ABOVE then
                o.oVelY = -PX
                cur_obj_change_action(1)
            else
                o.oVelY = PX
                cur_obj_change_action(2)
            end
        end
    elseif o.oAction == 1 then -- stomp death
        cur_obj_become_intangible()
        play_smb_sound_object(o, "stomp", true)
        enemy_move(o, true)
    elseif o.oAction == 2 then -- other death (star, mainly)
        cur_obj_become_intangible()
        play_smb_sound_object(o, "kick", true)
        enemy_move(o, true)
    end

    if o.oPosY < -16 * PX then
        obj_mark_for_deletion(o)
    end

    o.oInteractStatus = 0
end

id_bhvSmbBulletBill = hook_smb_behavior("bhvSmbBulletBill", OBJ_LIST_GENACTOR, smb_bullet_bill_init, smb_bullet_bill_loop)

-- fire bar parent object
---@param o Object
function smb_fire_bar_base_init(o)
    o.oFlags = 0
    o.oInteractType = INTERACT_DAMAGE
    o.oSkipObjectInteraction = 1
    o.hitboxRadius = 4 * PX -- 4 pixels on both sides from center
    o.hitboxHeight = 8 * PX
    o.hitboxDownOffset = 0
    o.oDeathSound = 0
    o.oAnimState = 0
    o.oSMBTimer2 = 0
    cur_obj_set_home_once()

    -- spawn children as non-sync objects
    local length = 6
    if o.oBehParams2ndByte ~= 0 then
        length = 12
    end
    local y = o.oPosY
    for i=1,length-1 do
        y = y + o.hitboxHeight
        spawn_object_no_rotate(id_bhvSmbFireBarPart, E_MODEL_NONE, o.oPosX, y, 0, false, function(o2)
            o2.parentObj = o
            o2.oBehParams2ndByte = i
            o2.oBehParams = o.oBehParams
        end)
    end

    network_init_object(o, true, {"oSMBTimer2"})
end

---@param o Object
function smb_fire_bar_base_loop(o)
    cur_obj_unhide()

    if o.oAction == 0 then -- main action
        if not obj_is_on_screen(o, 0, 0) then
            cur_obj_change_action(1)
            return
        end

        cur_obj_become_tangible()

        if o.oTimer % 4 == 0 then
            o.oAnimState = (o.oAnimState + 1) % 4
        end
        -- this is slightly slower than in SMB; 6 and 4 are too fast respectively
        -- I have literally no idea how to time this properly
        local timeToMove = 7
        if (o.oBehParams >> 8) & 0xFF ~= 0 then
            timeToMove = 5
        end
        if o.oTimer % timeToMove == 0 then
            if o.oBehParams & 0xFF == 0 then -- clockwise
                o.oSMBTimer2 = (o.oSMBTimer2 + 1) % 32
            else -- counter-clockwise
                o.oSMBTimer2 = (o.oSMBTimer2 - 1) % 32
            end
        end
    else -- off-screen
        cur_obj_hide()
        cur_obj_become_intangible()
        local homeOnScreen = point_is_on_screen(o.oHomeX, 0, 0)
        if o.oSubAction == 0 and not homeOnScreen then
            o.oPosX, o.oPosY = o.oHomeX, o.oHomeY
            o.oSubAction = 1
        elseif o.oSubAction == 1 and homeOnScreen then
            cur_obj_change_action(0)
        end
    end

    o.oInteractStatus = 0
end

id_bhvSmbFireBarBase = hook_smb_behavior("bhvSmbFireBarBase", OBJ_LIST_GENACTOR, smb_fire_bar_base_init, smb_fire_bar_base_loop)

-- fire bar parts
---@param o Object
function smb_fire_bar_part_init(o)
    o.oFlags = 0
    o.oInteractType = INTERACT_DAMAGE
    o.oSkipObjectInteraction = 1
    o.hitboxRadius = 4 * PX -- 4 pixels on both sides from center
    o.hitboxHeight = 8 * PX
    o.hitboxDownOffset = 0
    o.oDeathSound = 0
    o.oAnimState = 0
    o.oHomeX, o.oHomeY = o.parentObj.oHomeX, o.parentObj.oHomeY

    -- network_init_object(o, true, {})
end

---@param o Object
function smb_fire_bar_part_loop(o)
    -- delete if parent does not exist
    if o.parentObj == o or o.parentObj == nil then
        obj_mark_for_deletion(o)
        return
    end

    cur_obj_unhide()

    o.oAction = o.parentObj.oAction
    o.oTimer = o.parentObj.oTimer
    o.oSMBTimer2 = o.parentObj.oSMBTimer2
    if o.oAction == 0 then -- main action
        cur_obj_become_tangible()

        -- set position based on the value of oSMBTimer2
        -- there are 32 values
        o.oPosX = floor_or_ceil(o.oHomeX + sins(o.oSMBTimer2 * 0x800) * (o.hitboxRadius * 2) * o.oBehParams2ndByte)
        o.oPosY = floor_or_ceil(o.oHomeY + coss(o.oSMBTimer2 * 0x800) * (o.hitboxHeight) * o.oBehParams2ndByte)

        o.oAnimState = o.parentObj.oAnimState
    else -- off-screen
        cur_obj_hide()
        cur_obj_become_intangible()
    end

    o.oInteractStatus = 0
end

id_bhvSmbFireBarPart = hook_smb_behavior("bhvSmbFireBarPart", OBJ_LIST_GENACTOR, smb_fire_bar_part_init, smb_fire_bar_part_loop)

-- bowser

-- used for Bowser to determine when to change direction and when to jump
local PRandomRange = { 0x21, 0x41, 0x11, 0x31 }
-- used for Bowser to determine when to spit fire (16 frames are subtracted when hard mode is active)
local FlameTimerData = {0xBF, 0x40, 0xBF, 0xBF, 0xBF, 0x40, 0x40, 0xBF}
-- determines where bowser's flame's y coordinate is
local FlameYPosData = {0x90, 0x80, 0x70, 0x90}

---@param o Object
function smb_bowser_init(o)
    o.oFlags = 0
    o.oGravity = 1
    o.oMoveAngleYaw = -0x4000
    o.oFaceAngleYaw = o.oMoveAngleYaw
    o.oInteractType = INTERACT_DAMAGE

    o.hitboxRadius = 8 * PX -- 8 pixels wide
    o.hitboxHeight = 18 * PX
    o.hitboxDownOffset = -12 * PX
    o.oHitboxLeftOffset = 8 * PX
    -- Bowser has to be quirky and has TWO hitboxes!
    o.hurtboxRadius = 8 * PX -- 8 pixels wide
    o.hurtboxHeight = 18 * PX
    o.oHurtboxDownOffset = -3 * PX
    o.oHurtboxLeftOffset = -8 * PX

    if o.setHome == 0 then
        o.oBowserFlameProgress = 0
        o.oSMBTimer2 = 0xDF
        o.oBowserBroJumpTimer = cur_obj_random_from_table(PRandomRange)
        o.oHealth = 5 -- 5 fireballs to kill
    end

    o.oDeathSound = 0 -- used as "played death sound" flag
    o.oDamageOrCoinValue = 9 -- 5000 points by default
    o.oFireballOverridePoints = 1 -- always 5000 points when defeated by a fireball
    o.oAnimState = 0
    if o.oBehParams ~= 0 then
        o.oAnimState = o.oAnimState + 6
    end
    o.oTurnTowardsOnHit = 0
    if o.oHomeY == 0 then
        cur_obj_set_home_once()
    end
    network_init_object(o, true, {"oHomeX", "oHomeY", "oVelX", "oVelY", "oBowserMaxHomeDist", "oBowserBroJumpTimer", "oSMBTimer2", "oBowserFlameProgress"})
end

---@param o Object
function smb_bowser_loop(o)
    cur_obj_unhide()

    -- don't render if we're doing the cutscene
    local localInCutscene = (gMarioStates[0].action == ACT_SMB_HIT_AXE or gMarioStates[0].action == ACT_SMB_WALK_CUTSCENE)

    -- change hitboxes based on facing direction
    if o.oFaceAngleYaw <= 0 then -- left
        o.oHitboxLeftOffset = 8 * PX
        o.oHurtboxLeftOffset = -8 * PX
    else -- right
        o.oHitboxLeftOffset = -8 * PX
        o.oHurtboxLeftOffset = 8 * PX
    end

    local m = get_rightmost_mario_state()
    local world = ((gNetworkPlayers[0].currActNum-1) // 4) + 1
    o.oGravity = 1

    if o.oAction == 0 then -- moving/shooting
        if not (obj_is_on_screen(o, 4 * BLOCK, 4 * BLOCK) or point_is_on_screen(o.oHomeX, 4 * BLOCK, 4 * BLOCK)) then
            cur_obj_change_action(3)
            if m then
                gSMBStates[m.playerIndex].frenzyDisabled = false
            end
            return
        end

        if m then
            gSMBStates[m.playerIndex].frenzyDisabled = true
        end
        
        o.oFaceAnglePitch = 0
        if o.oPosX == o.oHomeX then
            -- set max dist from home to random in table
            o.oBowserMaxHomeDist = cur_obj_random_from_table(PRandomRange) * PX
        elseif sign(o.oPosX - o.oHomeX) == sign(o.oMoveAngleYaw) and math.abs(o.oPosX - o.oHomeX) > o.oBowserMaxHomeDist then
            -- change direction if out of range of home
            o.oMoveAngleYaw = sign(o.oMoveAngleYaw) * -0x4000
            -- set max dist from home to random in table
            o.oBowserMaxHomeDist = cur_obj_random_from_table(PRandomRange) * PX
        end
        
        local doHammers = (world > 5)
        local doFireSpit = (not doHammers) or world == 8

        if o.oSubAction == 0 then
            o.oVelX = sign(o.oMoveAngleYaw) * 4
        else
            o.oVelX = 0 -- don't move when shooting fireball
        end
        o.oFaceAngleYaw = -0x4000
        cur_obj_become_tangible()
        enemy_move(o, true)
        local landed = false
        if o.oPosY <= o.oHomeY then
            landed = true
            o.oPosY = o.oHomeY
            o.oVelY = 0
        end

        -- run towards right if mario is past bowser
        if m and m.pos.x > o.oPosX and o.oSubAction == 0 and o.oTimer > 30 then
            cur_obj_change_action(1)
            return
        end

        if o.oSubAction == 0 and o.oTimer % 32 == 0 then
            o.oAnimState = (o.oAnimState + 1) % 2
            if o.oBehParams ~= 0 then
                o.oAnimState = o.oAnimState + 6
            end
        end

        -- various timers
        if landed then
            o.oBowserBroJumpTimer = o.oBowserBroJumpTimer - 1
            if o.oBowserBroJumpTimer == 0 then
                o.oBowserBroJumpTimer = cur_obj_random_from_table(PRandomRange)
                o.oVelY = 2 * PX
            end
        elseif doHammers and o.oTimer % 4 == 0 then
            if m and m.playerIndex == 0 then
                attempt_spawn_hammer(o)
            end
        end
        if doFireSpit then
            o.oSMBTimer2 = o.oSMBTimer2 - 1
            if o.oSMBTimer2 == 0 then
                if o.oSubAction == 0 then
                    o.oSubAction = 1
                    o.oAnimState = o.oAnimState + 3
                    o.oSMBTimer2 = 0x20
                else
                    o.oSubAction = 0
                    o.oAnimState = o.oAnimState - 3
                    if m and m.playerIndex == 0 then
                        local mouthPos = {x = o.oPosX - 16 * PX, y = o.oPosY + 16 * PX}
                        spawn_object_no_rotate(id_bhvSmbBowserFlame, E_MODEL_NONE, mouthPos.x, mouthPos.y, 0, true, function(flame)
                            flame.oBehParams = 1
                        end)
                    end
                    o.oBowserFlameProgress = o.oBowserFlameProgress + 1
                    o.oSMBTimer2 = FlameTimerData[o.oBowserFlameProgress]
                    if do_hard_mode_objs() then
                        o.oSMBTimer2 = o.oSMBTimer2 - 0x10 -- spit fire sooner in hard mode
                    end
                    if o.oBowserFlameProgress >= #FlameTimerData then
                        o.oBowserFlameProgress = 0
                    end
                end
            end
        end

        if o.oInteractStatus & INT_STATUS_INTERACTED ~= 0 and o.oInteractStatus & INT_STATUS_WAS_ATTACKED ~= 0 then
            if o.oHealth <= 1 then
                o.oVelY = 2 * PX
                o.oVelX = 0
                cur_obj_change_action(2)
            else
                o.oHealth = o.oHealth - 1
            end
        end
    elseif o.oAction == 1 then -- desperately run towards right if mario is to the right
        if m then
            gSMBStates[m.playerIndex].frenzyDisabled = true
        end

        o.oMoveAngleYaw = 0x4000
        o.oFaceAngleYaw = o.oMoveAngleYaw
        o.oVelX = sign(o.oMoveAngleYaw) * 8
        enemy_move(o, true)
        if o.oPosY < o.oHomeY then
            o.oPosY = o.oHomeY
            o.oVelY = 0
        end
        if o.oPosX > o.oHomeX + 5 * BLOCK then
            o.oPosX = o.oHomeX + 5 * BLOCK
        end

        if o.oTimer % 32 == 0 then
            o.oAnimState = (o.oAnimState + 1) % 2
            if o.oBehParams ~= 0 then
                o.oAnimState = o.oAnimState + 6
            end
        end

        if ((not m) or m.pos.x <= o.oPosX) and o.oTimer > 30 then
            cur_obj_change_action(0)
        end
    elseif o.oAction == 2 then -- death
        if m and point_is_on_screen(o.oHomeX, 4 * BLOCK, 4 * BLOCK) then
            gSMBStates[m.playerIndex].frenzyDisabled = true
        end

        if world <= #FakeBowserData then
            o.oGravity = FakeBowserData[world][3]
            o.oAnimState = FakeBowserData[world][2]
        else
            o.oGravity = 1
            o.oAnimState = 0
            if o.oBehParams ~= 0 then
                o.oAnimState = o.oAnimState + 6
            end
        end

        o.oFaceAnglePitch = 0x8000 -- flip sprite
        cur_obj_become_intangible()
        if o.oDeathSound == 0 and not localInCutscene then
            play_smb_sound_object(o, "kick")
            play_smb_sound_object(o, "bowserDeath", true)
        end
        
        enemy_move(o, true)
    elseif o.oAction == 3 then -- off-screen
        cur_obj_hide()
        cur_obj_become_intangible()
        local homeOnScreen = point_is_on_screen(o.oHomeX, 4 * BLOCK, 4 * BLOCK)
        if homeOnScreen then
            o.oMoveAngleYaw = -0x4000
            o.oFaceAngleYaw = o.oMoveAngleYaw
            cur_obj_change_action(0)
        end
    elseif o.oAction == 4 then -- dead (remains in existence to manage frenzy objects)
        cur_obj_hide()
        cur_obj_become_intangible()
        if m and point_is_on_screen(o.oHomeX, 4 * BLOCK, 4 * BLOCK) then
            gSMBStates[m.playerIndex].frenzyDisabled = true
        end
    end

    if o.oPosY < -32 * PX then
        if o.oAction == 2 then
            cur_obj_change_action(4)
        else
            cur_obj_change_action(3)
        end
    end

    if localInCutscene then
        cur_obj_become_intangible()
        cur_obj_hide()
    end

    o.oInteractStatus = 0
end

id_bhvSmbBowser = hook_smb_behavior("bhvSmbBowser", OBJ_LIST_GENACTOR, smb_bowser_init, smb_bowser_loop)

-- cutscene bowser (used when a player hits an axe for the purposes of syncing)
---@param o Object
function smb_bowser_cutscene_init(o)
    o.oFlags = 0
    o.oGravity = 1
    o.oMoveAngleYaw = 0x4000
    o.oFaceAngleYaw = 0x4000

    o.oDeathSound = 0 -- used as "played death sound" flag
    o.oAnimState = 0
    if o.oBehParams ~= 0 then
        o.oAnimState = o.oAnimState + 6
    end

    if o.oHomeY == 0 then
        cur_obj_set_home_once()
    end
end

---@param o Object
function smb_bowser_cutscene_loop(o)
    cur_obj_unhide()
    local m = gMarioStates[0]

    if o.oAction == 0 then -- in air
        if m.action ~= ACT_SMB_HIT_AXE or m.actionState ~= 0 then
            o.oVelX = 0
            o.oVelY = 0
            cur_obj_change_action(1)
        end

        if o.oTimer % 4 == 0 then
            o.oAnimState = (o.oAnimState + 1) % 2
            if o.oBehParams ~= 0 then
                o.oAnimState = o.oAnimState + 6
            end
        end
    elseif o.oAction == 1 then -- fall
        enemy_move(o, true)
        play_smb_sound_object(o, "bowserDeath", true)
    end

    if o.oPosY < -32 * PX then
        obj_mark_for_deletion(o)
    end

    o.oInteractStatus = 0
end

id_bhvSmbBowserCutscene = hook_smb_behavior("bhvSmbBowserCutscene", OBJ_LIST_GENACTOR, smb_bowser_cutscene_init, smb_bowser_cutscene_loop)

-- fireballs
---@param o Object
function smb_fireball_init(o)
    o.oFlags = 0
    o.oGravity = 4
    o.oInteractType = INTERACT_DAMAGE
    o.hitboxRadius = 4 * PX -- 4 pixels on both sides from center
    o.hitboxHeight = 8 * PX
    o.hitboxDownOffset = 0
    o.oDeathSound = 0 -- used as "played death sound" flag
    o.oAnimState = 0
    o.oShellCombo = -2 -- 200 points by default when defeating an enemy
    o.oHurtsEnemies = 1
    o.oIntangibleToMarioTimer = -1
    network_init_object(o, true, {"oVelX", "oVelY"})
end

---@param o Object
function smb_fireball_loop(o)
    cur_obj_unhide()
    o.oShellCombo = -2 -- 200 points by default when defeating an enemy

    if o.oAction == 0 then
        if not obj_is_on_screen(o, 0, 0) then
            obj_mark_for_deletion(o)
        end

        if o.oTimer == 1 then
            o.oVelY = -TERMINAL_VEL
            play_smb_sound_object(o, "fireball")
        end

        o.oVelX = sign(o.oMoveAngleYaw) * 0x30
        cur_obj_become_tangible()
        local landed, hitWall = enemy_move_and_collide(o)
        if landed then
            o.oVelY = 0x30
        end
        if hitWall then
            play_smb_sound_object(o, "bump")
            cur_obj_change_action(1)
            o.oAnimState = 4
            return
        end

        if o.oTimer % 8 == 0 then
            o.oAnimState = (o.oAnimState + 1) % 3
        end

        if o.oInteractStatus & INT_STATUS_INTERACTED ~= 0 then
            cur_obj_change_action(1)
            o.oAnimState = 4
        end
    elseif o.oAction == 1 then
        cur_obj_become_intangible()
        if o.oTimer % 2 == 0 then
            o.oAnimState = (o.oAnimState - 3) % 3 + 4
        end
        if o.oTimer >= 6 then
            obj_mark_for_deletion(o)
        end
    end

    if o.oPosY < -16 * PX then
        obj_mark_for_deletion(o)
    end

    o.oInteractStatus = 0
end

id_bhvSmbFireball = hook_smb_behavior("bhvSmbFireball", OBJ_LIST_GENACTOR, smb_fireball_init, smb_fireball_loop)

-- bowser flame
---@param o Object
function smb_bowser_flame_init(o)
    o.oFlags = OBJ_FLAG_SET_FACE_YAW_TO_MOVE_YAW
    o.oGravity = 0
    o.oInteractType = INTERACT_DAMAGE
    o.hitboxRadius = 2 * PX -- 2 pixels on both sides from center
    o.hitboxHeight = 4 * PX
    o.hitboxDownOffset = 0
    o.oHitboxLeftOffset = 6 * PX
    o.oDeathSound = 0 -- used as "played sound" flag
    o.oAnimState = 1
    o.oMoveAngleYaw = -0x4000
    o.oSkipObjectInteraction = 1
    o.oRenderBehindFloor = 1
    if o.setHome == 0 then
        o.oHomeX = 0
        o.oHomeY = (NES_HEIGHT - cur_obj_random_from_table(FlameYPosData) - 8) * PX
        o.setHome = 1
    end
    network_init_object(o, true, {"oHomeX", "oHomeY", "oVelX", "oVelY"})
end

---@param o Object
function smb_bowser_flame_loop(o)
    cur_obj_unhide()

    -- don't render if we're doing the cutscene
    local localInCutscene = (gMarioStates[0].action == ACT_SMB_HIT_AXE or gMarioStates[0].action == ACT_SMB_WALK_CUTSCENE)

    if not obj_is_on_screen(o, BLOCK, 3 * BLOCK) then
        obj_mark_for_deletion(o)
    end

    if o.oTimer > 5 then
        o.oRenderBehindFloor = 0
    end

    if not localInCutscene then
        play_smb_sound_object(o, "bowserFlame", true)
    end

    -- move to selected Y position
    o.oVelX = -PX - 4
    local homeDiff = o.oHomeY - o.oPosY
    if homeDiff > PX then
        o.oVelY = PX
    elseif homeDiff < -PX then
        o.oVelY = -PX
    else
        o.oPosY = o.oHomeY
        o.oVelY = 0
    end
    cur_obj_become_tangible()
    enemy_move(o, false)

    if o.oTimer % 2 == 0 then
        o.oAnimState = (o.oAnimState + 2) % 4
    end

    if localInCutscene then
        cur_obj_become_intangible()
        cur_obj_hide()
    end

    o.oInteractStatus = 0
end

id_bhvSmbBowserFlame = hook_smb_behavior("bhvSmbBowserFlame", OBJ_LIST_GENACTOR, smb_bowser_flame_init, smb_bowser_flame_loop)

-- hammer
---@param o Object
function smb_hammer_init(o)
    o.oFlags = 0
    o.oGravity = 1
    o.oInteractType = INTERACT_DAMAGE

    -- this hitbox SUCKS
    o.hitboxRadius = 4 * PX -- 4 pixels on both sides from center
    o.hitboxHeight = 8 * PX
    o.hitboxDownOffset = -8 * PX
    o.oHitboxLeftOffset = 4 * PX

    -- so it doesn't appear not at the parent
    cur_obj_hide()
    cur_obj_become_intangible()

    o.oDeathSound = 0
    o.oAnimState = frame_based_on_level_theme(0, 1, 4)
    o.oSkipObjectInteraction = 1
    o.oRenderBehindFloor = 1
    network_init_object(o, true, {"oVelX", "oVelY"})
end

---@param o Object
function smb_hammer_loop(o)
    cur_obj_unhide()
    cur_obj_become_tangible()

    -- don't render if we're doing the cutscene
    local localInCutscene = (gMarioStates[0].action == ACT_SMB_HIT_AXE or gMarioStates[0].action == ACT_SMB_WALK_CUTSCENE)
    
    if o.oAction == 0 then
        if o.parentObj == o or o.parentObj == nil then
            o.oVelX = PX * sign(o.oMoveAngleYaw)
            o.oVelY = 2 * PX
            cur_obj_change_action(1)
            return
        end

        o.oMoveAngleYaw = o.parentObj.oFaceAngleYaw

        -- set position based on parent
        if obj_has_behavior_id(o.parentObj, id_bhvSmbBowser) ~= 0 then
            o.oPosX = o.parentObj.oPosX + (5 * PX) * sign(o.oMoveAngleYaw)
            o.oPosY = o.parentObj.oPosY + 26 * PX
        else
            o.oPosX = o.parentObj.oPosX - (3 * PX) * sign(o.oMoveAngleYaw)
            o.oPosY = o.parentObj.oPosY + (18 * PX)
            o.parentObj.usingObj = o -- so hammer bros animations work in multiplayer
        end

        o.oFaceAngleYaw = 0
        o.oRenderBehindFloor = 1
        if o.oTimer >= 0xE or o.parentObj.oFaceAnglePitch == 0x8000 then
            o.oVelX = PX * sign(o.oMoveAngleYaw)
            o.oVelY = 2 * PX
            cur_obj_change_action(1)
        end
    else
        o.oRenderBehindFloor = 0
        -- make all hammers always have the same orientation
        o.oFaceAngleYaw = ((get_global_timer() // 2) % 4) * 0x4000

        if not obj_is_on_screen(o, 0, 0) then
            obj_mark_for_deletion(o)
        end

        if o.oPosY < -16 * PX then
            obj_mark_for_deletion(o)
        end

        enemy_move(o, true)
        -- min y vel is -2 px
        if o.oVelY < -2 * PX then
            o.oVelY = -2 * PX
        end
    end

    if localInCutscene then
        cur_obj_hide()
        cur_obj_become_intangible()
    end

    o.oInteractStatus = 0
end

id_bhvSmbHammer = hook_smb_behavior("bhvSmbHammer", OBJ_LIST_GENACTOR, smb_hammer_init, smb_hammer_loop)

-- cheep cheeps/fish in underwater stages
---@param o Object
function smb_uw_cheep_cheep_init(o)
    o.oFlags = OBJ_FLAG_SET_FACE_YAW_TO_MOVE_YAW
    o.oGravity = 1
    o.oInteractType = INTERACT_BOUNCE_TOP2
    o.hitboxRadius = 5 * PX -- 5 pixels on both sides from center
    o.hitboxHeight = 6 * PX
    o.hitboxDownOffset = -4 * PX
    o.oDeathSound = 0 -- used as "played death sound" flag
    o.oDamageOrCoinValue = 2 -- 200 points by default (not possible anyway)

    o.oAnimState = 0
    if o.oBehParams & 0xFF ~= 0 then -- grey
        o.oAnimState = o.oAnimState + 2
    end

    o.oMoveAngleYaw = -0x4000
    cur_obj_set_home_once()
    network_init_object(o, true, {"oHomeX", "oHomeY", "oVelX", "oVelY"})
end

---@param o Object
function smb_uw_cheep_cheep_loop(o)
    cur_obj_unhide()

    if o.oAction == 0 then -- normal
        if not obj_is_on_screen(o, BLOCK, 3 * BLOCK) then
            obj_mark_for_deletion(o)
        end

        o.oFaceAnglePitch = 0
        o.oVelX = -4
        o.oVelY = 0
        if o.oBehParams & 0xFF == 0 then -- double speed for red
            o.oVelX = o.oVelX * 2
        end
        if o.oBehParams2ndByte ~= 0 then -- zig-zag
            if o.oSubAction == 0 then -- up
                o.oVelY = 2
                if o.oPosY >= o.oHomeY + BLOCK then
                    o.oSubAction = 1
                end
            else -- down
                o.oVelY = -2
                if o.oPosY <= o.oHomeY - BLOCK then
                    o.oSubAction = 0
                end
            end
        end
        cur_obj_become_tangible()
        enemy_move(o, false)

        if o.oTimer % 8 == 0 then
            o.oAnimState = (o.oAnimState + 1) % 2
            if o.oBehParams & 0xFF ~= 0 then -- grey
                o.oAnimState = o.oAnimState + 2
            end
        end

        if o.oInteractStatus & INT_STATUS_INTERACTED ~= 0 and o.oInteractStatus & INT_STATUS_WAS_ATTACKED ~= 0 then
            o.oVelX = 0
            o.oVelY = PX
            cur_obj_change_action(1)
        end
    else -- death
        o.oFaceAnglePitch = 0x8000 -- flip sprite
        cur_obj_become_intangible()
        play_smb_sound_object(o, "kick", true)
        enemy_move(o, true)
    end

    o.oInteractStatus = 0
end

id_bhvSmbUwCheepCheep = hook_smb_behavior("bhvSmbUwCheepCheep", OBJ_LIST_GENACTOR, smb_uw_cheep_cheep_init, smb_uw_cheep_cheep_loop)

-- flying cheep cheeps/fish featured in 2-3, 7-3, and 8-4
---@param o Object
function smb_flying_cheep_cheep_init(o)
    o.oFlags = OBJ_FLAG_SET_FACE_YAW_TO_MOVE_YAW
    o.oGravity = 1
    o.oInteractType = INTERACT_BOUNCE_TOP2
    o.hitboxRadius = 5 * PX -- 5 pixels on both sides from center
    o.hitboxHeight = 6 * PX
    o.hitboxDownOffset = -4 * PX
    o.oDeathSound = 0 -- used as "played death sound" flag
    o.oDamageOrCoinValue = 2 -- 200 points by default

    o.oAnimState = 0

    cur_obj_set_home_once()
    network_init_object(o, true, {"oHomeX", "oHomeY", "oVelX", "oVelY"})
end

---@param o Object
function smb_flying_cheep_cheep_loop(o)
    cur_obj_unhide()

    if o.oAction == 0 then -- normal
        if not obj_is_on_screen(o, 2 * BLOCK, 2 * BLOCK) then
            obj_mark_for_deletion(o)
        end

        o.oFaceAnglePitch = 0
        cur_obj_become_tangible()
        enemy_move(o, true)

        if o.oTimer % 8 == 0 then
            o.oAnimState = (o.oAnimState + 1) % 2
        end

        if o.oInteractStatus & INT_STATUS_INTERACTED ~= 0 and o.oInteractStatus & INT_STATUS_WAS_ATTACKED ~= 0 then
            o.oVelX = 0
            if o.oInteractStatus & INT_STATUS_ATTACK_MASK == ATTACK_FROM_ABOVE then
                o.oVelY = -PX
                cur_obj_change_action(1)
            else
                o.oVelY = PX
                cur_obj_change_action(2)
            end
        end
    elseif o.oAction == 1 then -- stomp death
        o.oFaceAnglePitch = 0x8000 -- flip sprite
        cur_obj_become_intangible()
        play_smb_sound_object(o, "stomp", true)
        enemy_move(o, true)
    elseif o.oAction == 2 then -- other death
        o.oFaceAnglePitch = 0x8000 -- flip sprite
        cur_obj_become_intangible()
        play_smb_sound_object(o, "kick", true)
        enemy_move(o, true)
    end

    if o.oPosY < -16 * PX then
        cur_obj_change_action(4)
    end

    o.oInteractStatus = 0
end

id_bhvSmbFlyingCheepCheep = hook_smb_behavior("bhvSmbFlyingCheepCheep", OBJ_LIST_GENACTOR, smb_flying_cheep_cheep_init, smb_flying_cheep_cheep_loop)

-- hammer bro
---@param o Object
function smb_hammer_bro_init(o)
    o.oFlags = 0
    o.oGravity = 4
    o.oInteractType = INTERACT_BOUNCE_TOP2
    o.hitboxRadius = 4.5 * PX -- 9 pixels wide
    o.hitboxHeight = 24 * PX
    o.hitboxDownOffset = 3 * PX -- extends BELOW hammer bro
    o.oHitboxLeftOffset = -0.5 * PX
    o.oDeathSound = 0 -- used as "played death sound" flag
    o.oDamageOrCoinValue = 6 -- 1000 points by default
    o.oFireballOverridePoints = 1 -- always 1000 points when defeated by a fireball
    o.oAnimState = frame_based_on_level_theme(0, 7, 2)

    o.oTurnTowardsOnHit = 1
    if o.setHome == 0 then
        o.oMoveAngleYaw = -0x4000
        o.oFaceAngleYaw = -0x4000
        -- initialize chase timer
        o.oSMBTimer2 = 0x80 * 21
        if do_hard_mode_objs() then
            o.oSMBTimer2 = 0x50 * 21
        end
    end
    if o.oHomeY == 0 then
        cur_obj_set_home_once()
    end
    o.setHome = 1
    -- oFlyGuyIdleTimer is the hammer throw timer, I ran out of fields
    network_init_object(o, true, {"oHomeX", "oHomeY", "oVelX", "oVelY", "oFlyGuyIdleTimer", "oSMBTimer2", "oBowserBroJumpTimer"})
end

---@param o Object
function smb_hammer_bro_loop(o)
    cur_obj_unhide()

    if o.oAction == 0 then
        if not obj_is_on_screen(o, 2 * BLOCK, 2 * BLOCK) then
            cur_obj_change_action(3)
            return
        end

        -- decrement chase timer
        if o.oSMBTimer2 ~= 0 then
            o.oSMBTimer2 = o.oSMBTimer2 - 1
        end

        local m = nearest_non_cutscene_mario_state_to_object(o)

        o.oFaceAnglePitch = 0
        if m then
            o.oFaceAngleYaw = 0x4000 * sign(m.pos.x - o.oPosX) -- face towards nearest mario
        end
        
        o.oVelX = sign(o.oMoveAngleYaw) * 4
        if o.oSMBTimer2 == 0 and o.oFaceAngleYaw < 0 then
            -- chase player
            o.oVelX = o.oVelX * 2
            o.oMoveAngleYaw = o.oFaceAngleYaw
        elseif o.oTimer % 64 == 0 then
            -- shimmy back and forth
            o.oMoveAngleYaw = -o.oMoveAngleYaw
        end

        -- handle hammer timer
        if o.oFlyGuyIdleTimer == 0 then
            o.oFlyGuyIdleTimer = 0x30
            if do_hard_mode_objs() then
                o.oFlyGuyIdleTimer = 0x1C
            end
            if m and m.playerIndex == 0 then
                o.usingObj = attempt_spawn_hammer(o)
            end
        else
            o.oFlyGuyIdleTimer = o.oFlyGuyIdleTimer - 1
        end

        -- handle jump timer
        if o.oBowserBroJumpTimer == 0 then
            local u8 = position_based_random_u16() & 0xFF
            o.oBowserBroJumpTimer = u8 | 192 -- 192-255 frames
        else
            o.oBowserBroJumpTimer = o.oBowserBroJumpTimer - 1
        end
        
        cur_obj_become_tangible()
        local landed, hitWall = false, false
        local doCollision = true
        if o.oSubAction == 1 then -- jumping up
            doCollision = (o.oVelY <= 0)
        elseif o.oSubAction == 2 then -- jumping down
            doCollision = (o.oVelY <= -TERMINAL_VEL)
        end
        if doCollision then
            landed, hitWall = enemy_move_and_collide(o)
        else
            enemy_move(o, true)
        end

        -- land from jump
        if landed then
            o.oSubAction = 0
        end

        local holdHammer = false
        if o.usingObj and o.usingObj ~= o then
            if o.usingObj.oAction == 0 then
                holdHammer = true
                if o.oAnimState % 7 < 2 then
                    o.oAnimState = o.oAnimState + 2
                end
            else
                o.usingObj = nil
                if o.oAnimState % 7 >= 2 then
                    o.oAnimState = o.oAnimState - 2
                end
            end
        end

        if o.oTimer % 8 == 0 then
            o.oAnimState = frame_based_on_level_theme((o.oAnimState % 7 + 1) % 2, 7, 2)
            if holdHammer then
                o.oAnimState = o.oAnimState + 2
            end
        end

        -- do jump
        if o.oSubAction == 0 and (o.oBowserBroJumpTimer == 0 or hitWall) then
            if o.oPosY < 5 * BLOCK then
                o.oSubAction = 1 -- jump up
            elseif o.oPosY >= 9 * BLOCK then
                o.oSubAction = 2 -- jump down
            else
                -- randomly jump either up or down
                o.oSubAction = (m and floor_or_ceil(m.pos.x)) or 0 -- to make a bit more random
                o.oSubAction = (position_based_random_u16() + o.oSubAction) % 2 + 1
            end

            if o.oSubAction == 1 then
                o.oVelY = 6 * PX
            else
                o.oVelY = 3 * PX
            end
        end

        if o.oInteractStatus & INT_STATUS_INTERACTED ~= 0 and o.oInteractStatus & INT_STATUS_WAS_ATTACKED ~= 0 then
            if o.oInteractStatus & INT_STATUS_ATTACK_MASK == ATTACK_FROM_ABOVE then
                o.oVelX = 0
                o.oVelY = -PX
                cur_obj_change_action(1)
            else
                o.oVelY = 4 * PX
                o.oVelX = sign(o.oVelX) * -8
                o.oFaceAngleYaw = -o.oFaceAngleYaw
                o.oMoveAngleYaw = o.oFaceAngleYaw
                cur_obj_change_action(2)
            end
        end
    elseif o.oAction == 1 then -- stomp death
        o.oFaceAnglePitch = 0x8000 -- flip sprite
        cur_obj_become_intangible()
        play_smb_sound_object(o, "stomp", true)
        enemy_move(o, true)
    elseif o.oAction == 2 then -- other death
        o.oFaceAnglePitch = 0x8000 -- flip sprite
        cur_obj_become_intangible()
        play_smb_sound_object(o, "kick", true)
        enemy_move(o, true)
    elseif o.oAction == 3 then -- off-screen
        cur_obj_hide()
        cur_obj_become_intangible()
        local homeOnScreen, side = point_is_on_screen(o.oHomeX, BLOCK, BLOCK)
        if o.oSubAction == 0 and not homeOnScreen then
            o.oPosX, o.oPosY = o.oHomeX, o.oHomeY
            o.oSubAction = 1
        elseif o.oSubAction == 1 and homeOnScreen then
            -- initialize chase and jump timer again
            o.oSMBTimer2 = 0x80 * 21
            oBowserBroJumpTimer = 0
            if do_hard_mode_objs() then
                o.oSMBTimer2 = 0x50 * 21
            end

            cur_obj_change_action(0)
        end
    end

    if o.oPosY < -16 * PX then
        if o.oAction == 1 or o.oAction == 2 then
            obj_mark_for_deletion(o)
        else
            cur_obj_change_action(3)
        end
    end

    o.oInteractStatus = 0
end

id_bhvSmbHammerBro = hook_smb_behavior("bhvSmbHammerBro", OBJ_LIST_GENACTOR, smb_hammer_bro_init, smb_hammer_bro_loop)

-- flag (on flagpole)
function smb_flag_init(o)
    o.oFlags = 0
    o.oGravity = 4
    o.oDeathSound = 0
    o.oAnimState = 0
    o.oPosX = o.oPosX - 8 * PX
    cur_obj_set_home_once()
end

function smb_flag_loop(o)
    local m = gMarioStates[0]
    if o.oAction == 0 then
        local FLAGPOLE_POINTS = {
            0x18, 0x22, 0x50, 0x68, 0x90
        }
        if m.action & ACT_GROUP_MASK ~= ACT_GROUP_CUTSCENE and math.abs(o.oPosX + 8 * PX - m.pos.x) <= 8 * PX and m.pos.y <= NES_HEIGHT * PX then
            -- award points based on where mario ends up on the flag
            local flagPoints = 1
            while flagPoints < #FLAGPOLE_POINTS and (o.oPosY - m.pos.y + BLOCK) // PX > FLAGPOLE_POINTS[flagPoints] do
                flagPoints = flagPoints + 1
            end
            award_points(m, (#FLAGPOLE_POINTS-flagPoints)*2+1, true, o.oPosX + BLOCK + 4 * PX, 3 * BLOCK, true)
            if m.pos.y < BLOCK * 3 then -- flag glitch
                set_mario_action(m, ACT_SMB_SLIDE_DOWN_FLAG, 1)
                cur_obj_change_action(2)
            else
                set_mario_action(m, ACT_SMB_SLIDE_DOWN_FLAG, 0)
                m.usedObj = o
                cur_obj_change_action(1)
            end
        end
    elseif o.oAction == 1 then
        o.oPosY = o.oPosY - 2 * PX
        if o.oPosY <= BLOCK * 3 + PX * 3 then
            o.oPosY = BLOCK * 3 + PX * 3
            cur_obj_change_action(2)
        end
    end
end

id_bhvSmbFlag = hook_smb_behavior("bhvSmbFlag", OBJ_LIST_LEVEL, smb_flag_init, smb_flag_loop)

-- castle flag; also handles ending sequence
function smb_castle_flag_init(o)
    o.oFlags = 0
    o.oGravity = 4
    o.oDeathSound = 0
    o.oRenderBehindFloor = 1
    o.oAnimState = 0
    cur_obj_set_home_once()
end

-- firework position relative to the flag (in pixels)
local FIREWORK_POS = {
    {-0x14, 0x45}, {-0x34, 0x15}, {0x2C, 0x35}, {0x2C, 0x5}, {-0x4, 0x35}, {-0x34, 0x15}
}
function smb_castle_flag_loop(o)
    local m = gMarioStates[0]
    local smb = gSMBStates[0]
    cur_obj_unhide()
    if o.oAction == 0 then
        cur_obj_hide()
        if m.action == ACT_SMB_WALK_CUTSCENE then
            if m.actionState == 0 then
                o.oBehParams = smb.timer % 10 -- store last digit of time completed
            elseif smb.timer == 0 then
                cur_obj_change_action(1)
            end
        end
    elseif o.oAction == 1 then
        o.oPosY = o.oPosY + PX
        if o.oPosY >= o.oHomeY + BLOCK * 2 then
            o.oPosY = o.oHomeY + BLOCK * 2
            cur_obj_change_action(2)
        end
    elseif o.oAction == 2 then
        -- trigger fireworks if last digit of time was 1, 3, or 6 (why???)
        if o.oBehParams == 1 or o.oBehParams == 3 or o.oBehParams == 6 then
            cur_obj_change_action(3)
        elseif m.actionTimer > 6 * 60 and o.oTimer >= 42 then
            entranceType = 0
            warp_to_level(LEVEL_SMB, 1, gNetworkPlayers[0].currActNum+1)
        end
    elseif o.oAction == 3 then -- fireworks sequence
        if o.oTimer == 1 then
            o.oAnimState = o.oAnimState + 1
            local x, y = FIREWORK_POS[o.oAnimState][1], FIREWORK_POS[o.oAnimState][2]
            spawn_object_no_rotate(id_bhvSmbFirework, E_MODEL_NONE, o.oPosX+x*PX, o.oPosY+y*PX, 0, false)
            o.oBehParams = o.oBehParams - 1
            o.oTimer = o.oTimer + 1
            if o.oBehParams == 0 then
                cur_obj_change_action(2)
            end
        elseif o.oTimer > 32 then
            o.oTimer = 0
        end
    end
end

id_bhvSmbCastleFlag = hook_smb_behavior("bhvSmbCastleFlag", OBJ_LIST_LEVEL, smb_castle_flag_init, smb_castle_flag_loop)

-- firework
---@param o Object
function smb_firework_init(o)
    o.oFlags = 0
    o.oAnimState = 4
end

---@param o Object
function smb_firework_loop(o)
    if o.oTimer % 8 == 0 then
        o.oAnimState = (o.oAnimState - 3) % 3 + 4
    end
    if o.oTimer >= 8 * 3 then
        play_smb_sound_object(o, "boom")
        local smb = gSMBStates[0]
        smb.points = smb.points + 500
        o.oTimer = 0
        obj_mark_for_deletion(o)
    end
end

id_bhvSmbFirework = hook_smb_behavior("bhvSmbFirework", OBJ_LIST_UNIMPORTANT, smb_firework_init, smb_firework_loop)

-- bouncing block (when hitting a block)
---@param o Object
function smb_bounce_block_init(o)
    o.oFlags = 0
    o.oGravity = 4
    o.oMoveAngleYaw = 0
    o.oVelY = 24
    cur_obj_set_home_once()
    o.oAnimState = get_tile_frame(o.oBehParams) or 3
end

---@param o Object
function smb_bounce_block_loop(o)
    enemy_move(o, true)
    if o.oTimer > 15 then
        local blockX, blockY = (o.oHomeX // BLOCK), (o.oHomeY // BLOCK)
        add_to_2d_table(loaded_tile_data, blockX, blockY, o.oBehParams)
        obj_mark_for_deletion(o)
    end
end

id_bhvSmbBounceBlock = hook_smb_behavior("bhvSmbBounceBlock", OBJ_LIST_LEVEL, smb_bounce_block_init, smb_bounce_block_loop)

-- coin particle effect
---@param o Object
function smb_coin_particle_init(o)
    o.oFlags = 0
    o.oGravity = 4
    o.oMoveAngleYaw = 0
    o.oVelY = 72
    cur_obj_set_home_once()
    o.oAnimState = 0
end

---@param o Object
function smb_coin_particle_loop(o)
    enemy_move(o, true)
    if o.oTimer % 2 ~= 0 then
        o.oAnimState = (o.oAnimState + 1) % 4
    end
    if o.oVelY <= 0 and o.oPosY < o.oHomeY + BLOCK then
        if o.oBehParams ~= 0 then
            spawn_object_no_rotate(id_bhvSmbPointNumber, E_MODEL_NONE, o.oPosX, o.oPosY, 0, false, function(o)
                o.oAnimState = 1
            end)
        end
        obj_mark_for_deletion(o)
    end
end

id_bhvSmbCoinParticle = hook_smb_behavior("bhvSmbCoinParticle", OBJ_LIST_UNIMPORTANT, smb_coin_particle_init, smb_coin_particle_loop)

-- brick particle ffect
---@param o Object
function smb_brick_particle_init(o)
    o.oFlags = 0
    o.oGravity = 4
    o.oFaceAngleYaw = 0
    o.oMoveAngleYaw = 0
    cur_obj_set_home_once()
    o.oAnimState = frame_based_on_level_theme(0, 1, 4)
end

-- 1 px on x axis per frame
-- lower bricks jump up 1 block and higher jump up 3
-- spawn in each corner of the brick
---@param o Object
function smb_brick_particle_loop(o)
    enemy_move(o, true)
    if o.oTimer % 4 == 0 then
        o.oFaceAngleYaw = limit_angle(o.oFaceAngleYaw + 0x4000)
    end
    if o.oPosY < -16 * PX then
        obj_mark_for_deletion(o)
    end
end

id_bhvSmbBrickParticle = hook_smb_behavior("bhvSmbBrickParticle", OBJ_LIST_UNIMPORTANT, smb_brick_particle_init, smb_brick_particle_loop)

-- point visual
---@param o Object
function smb_point_number_init(o)
    local smb = gSMBStates[0]
    o.oFlags = 0
    o.oHomeX = o.oPosX - (smb.cameraX * PX)
end

---@param o Object
function smb_point_number_loop(o)
    -- flagpole
    if o.oAction == 1 then
        if o.parentObj == o or o.parentObj == nil then
            o.parentObj = obj_get_first_with_behavior_id(id_bhvSmbFlag)
        end
        o.oPosY = 14.5 * BLOCK - o.parentObj.oPosY
        return
    end

    local smb = gSMBStates[0]
    o.oPosX = o.oHomeX + (smb.cameraX * PX)
    o.oPosY = o.oPosY + (PX // 2)
    if o.oTimer > 45 then
        obj_mark_for_deletion(o)
    end
end

id_bhvSmbPointNumber = hook_smb_behavior("bhvSmbPointNumber", OBJ_LIST_UNIMPORTANT, smb_point_number_init, smb_point_number_loop)

-- air bubble (in water levels)
---@param o Object
function smb_air_bubble_init(o)
    o.oFlags = 0
    cur_obj_unhide()
end

---@param o Object
function smb_air_bubble_loop(o)
    -- randomly move up one pixel
    if math.random(0, 1) == 0 then
        o.oPosY = o.oPosY + PX
    end

    if o.oPosY > (NES_HEIGHT - 8) * PX - 2 * BLOCK then
        obj_mark_for_deletion(o)
    end
end

id_bhvSmbAirBubble = hook_smb_behavior("bhvSmbAirBubble", OBJ_LIST_UNIMPORTANT, smb_air_bubble_init, smb_air_bubble_loop)

-- messenger (peach or toad)
-- they do literally nothing but stand there
---@param o Object
function smb_messenger_init(o)
    o.oFlags = 0
    o.oAnimState = 0
    if gNetworkPlayers[0].currActNum > 28 then
        o.oAnimState = 1
    end
end

---@param o Object
function smb_messenger_loop(o)
    -- LITERALLY NOTHING
end

id_bhvSmbMessenger = hook_smb_behavior("bhvSmbMessenger", OBJ_LIST_LEVEL, smb_messenger_init, smb_messenger_loop)

-- springboard (TODO: castle palette for 6-3)
---@param o Object
function smb_springboard_init(o)
    o.oFlags = 0
    o.oGravity = 0
    o.oInteractType = INTERACT_SPRINGBOARD
    o.hitboxRadius = 0.5 * BLOCK
    o.hitboxHeight = 2 * BLOCK
    o.oAnimState = frame_based_on_level_theme(0, 7, 3)
    o.allowRemoteInteractions = 1 -- so other players can visually bounce off of the spring
    if o.oHomeY == 0 then
        cur_obj_set_home_once()
    end
    --network_init_object(o, true, {})
end

---@param o Object
function smb_springboard_loop(o)
    cur_obj_unhide()
    cur_obj_become_tangible()

    if o.oAction == 0 then -- pre-bounce
        if o.oInteractStatus & INT_STATUS_INTERACTED ~= 0 then
            cur_obj_change_action(1)
        end
    else -- during bounce
        if o.oTimer % 4 == 1 then
            if o.oSubAction == 0 then
                o.oAnimState = (o.oAnimState % 7 + 1) % 4
                if o.oAnimState == 2 then
                    o.oSubAction = 1
                end
            else
                o.oAnimState = (o.oAnimState % 7 - 1) % 4
                if o.oAnimState == 0 then
                    cur_obj_change_action(0)
                end
            end
            o.oAnimState = frame_based_on_level_theme(o.oAnimState, 7, 3)
        end
    end

    o.oInteractStatus = 0
end

id_bhvSmbSpringboard = hook_smb_behavior("bhvSmbSpringboard", OBJ_LIST_LEVEL, smb_springboard_init, smb_springboard_loop)

-- vine
function smb_vine_init(o)
    o.oFlags = 0
    o.oGravity = 0
    o.oInteractType = INTERACT_POLE
    o.hitboxRadius = 4 * PX
    o.hitboxHeight = 0 -- increases as it grows
    o.oRenderBehindFloor = 1
    o.oAnimState = frame_based_on_level_theme(1, 2, 2)
    o.allowRemoteInteractions = 1 -- so other players will climb the vine
    if o.oHomeY == 0 then
        cur_obj_set_home_once()
    end
    if o.coopFlags & COOP_OBJ_FLAG_NON_SYNC == 0 then
        network_init_object(o, true, {})
    end
end

function smb_vine_loop(o)
    cur_obj_unhide()
    cur_obj_become_tangible()
    
    play_smb_sound_object(o, "vine", true)

    if o.oAction == 0 then -- growing
        o.oPosY = o.oPosY + 8
        if o.oPosY > NES_HEIGHT * PX or (o.oBehParams == 1 and o.oPosY > 5 * BLOCK) then
            cur_obj_change_action(1)
        end
    end

    -- extend hitbox from its home
    o.hitboxHeight = o.oPosY - o.oHomeY - 8 * PX
    o.hitboxDownOffset = o.hitboxHeight - 8 * PX

    o.oInteractStatus = 0
end

id_bhvSmbVine = hook_smb_behavior("bhvSmbVine", OBJ_LIST_LEVEL, smb_vine_init, smb_vine_loop)

-- controllers; invisible objects that do things like handle exits, frency objects, etc.
function smb_controller_init(o)
    cur_obj_become_intangible()
    cur_obj_hide()
end

-- exit controller; controls where pipes and vines go when loaded
function smb_exit_controller_loop(o)
    local loaded = obj_is_on_screen(o, 0, 0, true)
    if o.oAction == 0 and loaded then
        doUpwardPipe = ((o.oBehParams >> 24) ~= 0) -- first byte
        pipeEntranceArea = o.oBehParams2ndByte -- second byte
        pipeEntranceScreen = (o.oBehParams >> 8) & 0xFF -- third byte
        bonusExitScreen = o.oBehParams & 0xFF -- last byte
        cur_obj_change_action(1)
    elseif o.oAction == 1 and not loaded then
        cur_obj_change_action(0)
    end
end
id_bhvSmbExitController = hook_smb_behavior("bhvSmbExitController", OBJ_LIST_LEVEL, smb_controller_init, smb_exit_controller_loop)

-- warp zone controller; controls the warp zone active state and where the pipes go
function smb_warp_zone_controller_loop(o)
    if o.oAction == 0 then
        local loaded = obj_is_on_screen(o, 0, 0, true)
        if loaded then
            warpZoneActive = true
            warpZoneData = {o.oBehParams >> 16, (o.oBehParams >> 8) & 0xFF, o.oBehParams & 0xFF}
            cur_obj_change_action(1)
        end
    elseif o.oAction == 1 then
        -- only unload if the entire screen is unloaded
        local loaded = obj_is_on_screen(o, 0, NES_WIDTH * PX, true)
        if not loaded then
            warpZoneActive = false
            warpZoneData = {36, 5, 36} -- default
            cur_obj_change_action(0)
        end
    end
end
id_bhvSmbWarpZoneController = hook_smb_behavior("bhvSmbWarpZoneController", OBJ_LIST_LEVEL, smb_controller_init, smb_warp_zone_controller_loop)

-- loop controller; used for looping levels (4-4, 7-4, and 8-4)
function smb_loop_controller_loop(o) -- funny name
    local m = gMarioStates[0]
    local smb = gSMBStates[0]
    if o.oAction == 0 then
        if (not waitForLoop) and m.pos.x > o.oPosX then
            if m.action == ACT_SMB_IN_PIPE or m.action == ACT_SMB_BUBBLED then
                -- if we're coming out of a pipe, ignore (for 8-4)
                -- also ignore if we're in a bubble so that we can reach players out of the loop
                cur_obj_change_action(1)
                smb.loopScreenRight = 0
            elseif math.abs(o.oPosY - m.pos.y) <= 1.5 * BLOCK then -- check if mario's y pos within 1.5 blocks
                -- passed!
                play_smb_sound_object(o, "pause")
                cur_obj_change_action(1)
                smb.loopScreenRight = 0
            else -- fail, set loop
                play_smb_sound_object(o, "break")
                waitForLoop = true
                smb.loopScreenRight = o.oBehParams & 0xFF
                smb.loopScreenLeft = o.oBehParams2ndByte
            end
        end
    elseif waitForLoop and smb.loopScreenRight ~= 0 and smb.loopScreenLeft * 16 * BLOCK < o.oPosX then
        -- reset if our loop point got set behind this object (for chained loop points in 7-4)
        cur_obj_change_action(0)
    end
end
id_bhvSmbLoopController = hook_smb_behavior("bhvSmbLoopController", OBJ_LIST_LEVEL, smb_controller_init, smb_loop_controller_loop)

-- bowser flame frenzy
function smb_bowser_flame_frenzy_loop(o)
    local frenzyValid = frenzy_spawn_is_valid(o, 3 * BLOCK)

    if o.oAction == 0 then
        if frenzyValid then
            if o.oSMBTimer2 == 0 then
                o.oSMBTimer2 = 0x40
                if do_hard_mode_objs() then
                    o.oSMBTimer2 = o.oSMBTimer2 - 0x10 -- spit fire sooner in hard mode
                end
            end
            cur_obj_change_action(1)
        end
    elseif o.oAction == 1 then
        if not frenzyValid then
            cur_obj_change_action(0)
            return
        end

        if o.oSMBTimer2 ~= 0 then
            o.oSMBTimer2 = o.oSMBTimer2 - 1
        end
        if o.oSMBTimer2 == 0 then
            local smb = gSMBStates[0]
            local x = (smb.cameraX + NES_WIDTH + 16) * PX
            local y = (NES_HEIGHT - FlameYPosData[math.random(1, #FlameYPosData)] - 8) * PX
            spawn_object_no_rotate(id_bhvSmbBowserFlame, E_MODEL_NONE, x, y, 0, true, function(flame)
                flame.oHomeX = x
                flame.oHomeY = y
                flame.setHome = 1
            end)
            
            o.oBowserFlameProgress = o.oBowserFlameProgress + 1
            o.oSMBTimer2 = FlameTimerData[o.oBowserFlameProgress]
            if do_hard_mode_objs() then
                o.oSMBTimer2 = o.oSMBTimer2 - 0x10 -- spit fire sooner in hard mode
            end
            if o.oBowserFlameProgress >= #FlameTimerData then
                o.oBowserFlameProgress = 0
            end
        end
    end
end
id_bhvSmbBowserFlameFrenzy = hook_smb_behavior("bhvSmbBowserFlameFrenzy", OBJ_LIST_LEVEL, smb_controller_init, smb_bowser_flame_frenzy_loop)

-- uw fish frenzy, used in water levels
-- not to be confused with the flying fish frenzy used in 2-3, etc.

-- where fish can spawn
local FishYPosData = {0x40, 0x30, 0x90, 0x50, 0x20, 0x60, 0xa0, 0x70}
---@param o Object
function smb_uw_fish_frenzy_loop(o)
    local frenzyValid = frenzy_spawn_is_valid(o, 3 * BLOCK)

    if o.oAction == 0 then
        if frenzyValid then
            if o.oSMBTimer2 == 0 then
                o.oSMBTimer2 = 0x20
            end
            cur_obj_change_action(1)
        end
    elseif o.oAction == 1 then
        if not frenzyValid then
            cur_obj_change_action(0)
            return
        end

        if o.oSMBTimer2 ~= 0 then
            o.oSMBTimer2 = o.oSMBTimer2 - 1
        end
        if o.oSMBTimer2 == 0 then
            -- count total fish on our screen; only spawn if there aren't 3 yet
            local fishCount = obj_count_on_screen(id_bhvSmbUwCheepCheep, BLOCK, 3 * BLOCK)
            if fishCount < 3 then
                local smb = gSMBStates[0]
                local x = (smb.cameraX + NES_WIDTH + 16) * PX
                local y = (NES_HEIGHT - FishYPosData[math.random(1, #FishYPosData)] - 48) * PX

                -- 2/3 chance for red and 1/3 chance for grey (reversed if this is world 2)
                local world = ((gNetworkPlayers[0].currActNum-1) // 4) + 1
                local param1 = 0
                if (math.random(0, 2) == 0) ~= (world == 2) then
                    param1 = 1
                end
                -- 50/50 for each movement type
                local param2 = math.random(0, 1)

                spawn_object_no_rotate(id_bhvSmbUwCheepCheep, E_MODEL_NONE, x, y, 0, true, function(fish)
                    fish.oHomeX = x
                    fish.oHomeY = y
                    fish.oBehParams = param1
                    fish.oBehParams2ndByte = param2
                    fish.setHome = 1
                end)
            end

            o.oSMBTimer2 = 0x20
        end
    end
end

id_bhvSmbUwFishFrenzy = hook_smb_behavior("bhvSmbUwFishFrenzy", OBJ_LIST_LEVEL, smb_controller_init, smb_uw_fish_frenzy_loop)

-- flying fish frenzy, used in 2-3, 7-3, and 8-4

-- possible x speeds for cheep cheeps (in order: not moving, walk, run)
local FlyCCXSpeedData = {{0x0e, 0x05, 0x06, 0x0e}, {0x1c, 0x20, 0x10, 0x0c}, {0x1e, 0x22, 0x18, 0x14}}
-- possible times between cheep-cheep spawns
local FlyCCTimerData = {0x10, 0x60, 0x20, 0x48}
---@param o Object
function smb_flying_fish_frenzy_loop(o)
    local frenzyValid = frenzy_spawn_is_valid(o, 2 * BLOCK)

    if o.oAction == 0 then
        if frenzyValid then
            if o.oSMBTimer2 == 0 then
                o.oSMBTimer2 = FlyCCTimerData[math.random(1, #FlyCCTimerData)]
            end
            cur_obj_change_action(1)
        end
    elseif o.oAction == 1 then
        if not frenzyValid then
            cur_obj_change_action(0)
            return
        end

        if o.oSMBTimer2 ~= 0 then
            o.oSMBTimer2 = o.oSMBTimer2 - 1
        end
        if o.oSMBTimer2 == 0 then
            -- count total fish on our screen; only spawn if there aren't 3 yet
            local fishCount = obj_count_on_screen(id_bhvSmbFlyingCheepCheep, 2 * BLOCK, 2 * BLOCK)
            local maxFish = 3
            if do_hard_mode_objs() then
                maxFish = 4
            end

            if fishCount < maxFish then
                local m = gMarioStates[0]
                local smb = gSMBStates[0]

                -- Don't spawn near middle
                -- Not smb accurate but I don't care at this point
                local x = 0
                if math.random(0, 1) == 0 then
                    x = floor_or_ceil(smb.cameraX + NES_WIDTH // 10 * math.random(0, 3)) * PX -- left side
                else
                    x = floor_or_ceil(smb.cameraX + NES_WIDTH // 10 * math.random(7, 10)) * PX -- right side
                end

                local y = -8 * PX
                local side = 1
                local xSpeedData = FlyCCXSpeedData[1]
                if math.abs(m.vel.x) >= MAX_RUN_VEL then
                    xSpeedData = FlyCCXSpeedData[3] -- running cheep cheep speeds
                elseif math.abs(m.vel.x) >= MAX_WALK_VEL then
                    xSpeedData = FlyCCXSpeedData[2] -- walking cheep cheep speeds
                elseif x - m.pos.x > 0 then
                    -- if we're not moving very fast and we spawn to the right, face left towards mario
                    side = -1
                end
                
                local xSpeed = xSpeedData[math.random(1, #xSpeedData)]

                spawn_object_no_rotate(id_bhvSmbFlyingCheepCheep, E_MODEL_NONE, x, y, 0, true, function(fish)
                    fish.oHomeX = x
                    fish.oHomeY = y
                    fish.oMoveAngleYaw = side * 0x4000
                    fish.oVelX = xSpeed * side
                    fish.oVelY = 5 * PX
                    fish.setHome = 1
                end)
            end

            o.oSMBTimer2 = FlyCCTimerData[math.random(1, #FlyCCTimerData)]
        end
    end
end

id_bhvSmbFlyingFishFrenzy = hook_smb_behavior("bhvSmbFlyingFishFrenzy", OBJ_LIST_LEVEL, smb_controller_init, smb_flying_fish_frenzy_loop)

-- bullet bill frenzy, used in 5-3... and no other levels, surprisingly
-- in SMB, this is actually the same object as the uw fish frenzy, so they share a lot of code
---@param o Object
function smb_bullet_bill_frenzy_loop(o)
    local frenzyValid = frenzy_spawn_is_valid(o, 3 * BLOCK)

    if o.oAction == 0 then
        if frenzyValid then
            if o.oSMBTimer2 == 0 then
                o.oSMBTimer2 = 0x20
            end
            cur_obj_change_action(1)
        end
    elseif o.oAction == 1 then
        if not frenzyValid then
            cur_obj_change_action(0)
            return
        end

        if o.oSMBTimer2 ~= 0 then
            o.oSMBTimer2 = o.oSMBTimer2 - 1
        end
        if o.oSMBTimer2 == 0 then
            -- don't spawn if there is a bullet on screen
            local bulletCount = obj_count_on_screen(id_bhvSmbBulletBill, 4 * BLOCK, 4 * BLOCK)
            if bulletCount < 1 then
                local smb = gSMBStates[0]
                local x = (smb.cameraX + NES_WIDTH + 16) * PX
                local y = (NES_HEIGHT - FishYPosData[math.random(1, #FishYPosData)] - 48) * PX

                spawn_object_no_rotate(id_bhvSmbBulletBill, E_MODEL_NONE, x, y + PX, 0, true, function(bullet)
                    bullet.oMoveAngleYaw = -0x4000
                    bullet.oFaceAngleYaw = bullet.oMoveAngleYaw
                    bullet.oBehParams = 1
                end)

                o.oSMBTimer2 = 0x20
            end
        end
    end
end

id_bhvSmbBulletBillFrenzy = hook_smb_behavior("bhvSmbBulletBillFrenzy", OBJ_LIST_LEVEL, smb_controller_init, smb_bullet_bill_frenzy_loop)

-- stop frenzy controller (also affects lakitu)
function smb_stop_frenzy_loop(o)
    if o.oAction == 0 then
        local loaded = obj_is_on_screen(o, loaded_tile_data.width * BLOCK, 0, true)
        if loaded then
            gSMBStates[0].frenzyDisabled = true
            cur_obj_change_action(1)
        end
    elseif o.oAction == 1 then
        -- only unload to the left
        local loaded = obj_is_on_screen(o, loaded_tile_data.width * BLOCK, 0, true)
        if not loaded then
            gSMBStates[0].frenzyDisabled = false
            cur_obj_change_action(0)
        end
    end
end
id_bhvSmbStopFrenzy = hook_smb_behavior("bhvSmbStopFrenzy", OBJ_LIST_LEVEL, smb_controller_init, smb_stop_frenzy_loop)

-- whirlpool; sets mario's whirlpool state
function smb_whirlpool_loop(o)
    local lowerBound = o.oPosX - (o.oBehParams * BLOCK / 2) - 8 * PX
    local upperBound = o.oPosX + (o.oBehParams * BLOCK / 2) + 8 * PX
    for i=0,MAX_PLAYERS-1 do
        local m = gMarioStates[i]
        if is_player_active_and_synced(m) ~= 0 and m.pos.x >= lowerBound and m.pos.x <= upperBound then
            gSMBStates[i].inWhirlpool = true
            m.usedObj = o
        end
    end
end
id_bhvSmbWhirlpool = hook_smb_behavior("bhvSmbWhirlpool", OBJ_LIST_LEVEL, smb_controller_init, smb_whirlpool_loop)

-- bullet bill blaster (the spawner, not the bullet itself)
function smb_bill_blaster_init(o)
    cur_obj_hide()
    cur_obj_become_intangible()
    cur_obj_set_home_once()
    
    network_init_object(o, false, {"oSMBTimer2"})
end

function smb_bill_blaster_loop(o)
    -- only run for nearest mario so randomization is good
    local m = nearest_non_cutscene_mario_state_to_object(o)
    if (not m) or m.playerIndex ~= 0 then return end

    local bulletCount = obj_count_on_screen(id_bhvSmbBulletBill, 4 * BLOCK, 4 * BLOCK)
    if bulletCount >= 3 then return end -- wait to fire

    local sync = false
    if o.oSMBTimer2 ~= 0 then
        local chance = 32
        if do_hard_mode_objs() then
            chance = 16
        end
        -- count down at random
        -- this is actually around 1/4 as likely as in SMB
        -- however SMB's randomization function made it less likely in that game,
        -- so presumably, it balances out
        if math.random(1, chance) <= 3 then
            o.oSMBTimer2 = o.oSMBTimer2 - 1
            sync = true
        end
    else
        o.oSMBTimer2 = 0xE
        sync = false
        network_send_object(o, false)
        -- don't fire if off screen or a mario is close
        if not obj_is_on_screen(o, -8 * PX, -8 * PX) then return end
        
        if get_horizontal_dist_between_objects(m.marioObj, o) < 2.5 * BLOCK then return end
        
        spawn_object_no_rotate(id_bhvSmbBulletBill, E_MODEL_NONE, o.oPosX, o.oPosY, 0, true, function(bullet)
            bullet.oRenderBehindFloor = 1
            bullet.oMoveAngleYaw = sign(m.pos.x - o.oPosX) * 0x4000 -- towards mario
            bullet.oFaceAngleYaw = bullet.oMoveAngleYaw
        end)
    end

    if sync then
        network_send_object(o, false)
    end
end
id_bhvSmbBillBlaster = hook_smb_behavior("bhvSmbBillBlaster", OBJ_LIST_LEVEL, smb_bill_blaster_init, smb_bill_blaster_loop)

-- mushrooms (super and 1up)
---@param o Object
function smb_mushroom_init(o)
    o.oFlags = OBJ_FLAG_SET_FACE_YAW_TO_MOVE_YAW
    o.oGravity = 4
    o.oMoveAngleYaw = 0x4000
    o.oInteractType = INTERACT_POWERUP
    o.oDamageOrCoinValue = ((o.oBehParams == 0) and 0) or 3 -- determines powerup type
    o.hitboxRadius = 6 * PX -- 6 pixels on both sides from center
    o.hitboxHeight = 12 * PX
    o.hitboxDownOffset = -3 * PX
    o.oAnimState = 0
    o.oRenderBehindFloor = 1
    if o.oBehParams ~= 0 then
        o.oAnimState = frame_based_on_level_theme(1, 1, 3)
    end
    cur_obj_become_intangible()
    cur_obj_set_home_once()
    network_init_object(o, true, {"oHomeX", "oHomeY", "oVelX", "oVelY"})
end

---@param o Object
function smb_mushroom_loop(o)
    cur_obj_unhide()

    if o.oAction == 0 then
        -- rise from block
        if o.oTimer > 15 then   
            cur_obj_become_intangible()
            o.oRenderBehindFloor = 1
            o.oPosY = o.oPosY + 4
            if o.oPosY >= o.oHomeY + BLOCK then
                cur_obj_change_action(1)
                o.oRenderBehindFloor = 0
            end
        else
            cur_obj_hide()
        end
    else
        o.oRenderBehindFloor = 0
        o.oVelX = sign(o.oMoveAngleYaw) * 16
        cur_obj_become_tangible()
        local landed, hitWall = enemy_move_and_collide(o)

        if hitWall then
            o.oMoveAngleYaw = -o.oMoveAngleYaw
        end

        if o.oInteractStatus & INT_STATUS_INTERACTED ~= 0 then
            if o.oInteractStatus & INT_STATUS_ATTACK_MASK ~= ATTACK_FROM_BELOW then
                obj_mark_for_deletion(o)
                return
            else
                o.oVelY = 4 * PX
            end
        end

        if not obj_is_on_screen(o, 4 * BLOCK, 4 * BLOCK) then
            obj_mark_for_deletion(o)
            return
        end
    end

    if o.oPosY < -16 * PX then
        obj_mark_for_deletion(o)
    end

    o.oInteractStatus = 0
end

id_bhvSmbMushroom = hook_smb_behavior("bhvSmbMushroom", OBJ_LIST_LEVEL, smb_mushroom_init, smb_mushroom_loop)

-- fire flower
---@param o Object
function smb_fire_flower_init(o)
    o.oFlags = OBJ_FLAG_SET_FACE_YAW_TO_MOVE_YAW
    o.oGravity = 4
    o.oMoveAngleYaw = 0x4000
    o.oInteractType = INTERACT_POWERUP
    o.oDamageOrCoinValue = 1 -- determines powerup type
    o.hitboxRadius = 6 * PX -- 6 pixels on both sides from center
    o.hitboxHeight = 12 * PX
    o.hitboxDownOffset = -3 * PX
    o.oAnimState = 0
    o.oRenderBehindFloor = 1
    o.oAnimState = frame_based_on_level_theme(0, 4, 4)
    cur_obj_become_intangible()
    cur_obj_set_home_once()
    network_init_object(o, true, {"oHomeX", "oHomeY"})
end

---@param o Object
function smb_fire_flower_loop(o)
    cur_obj_unhide()

    if o.oTimer % 2 ~= 0 then
        o.oAnimState = frame_based_on_level_theme((o.oAnimState + 1) % 4, 4, 4)
    end

    if o.oAction == 0 then
        -- rise from block
        if o.oTimer > 15 then   
            cur_obj_become_intangible()
            o.oRenderBehindFloor = 1
            o.oPosY = o.oPosY + 4
            if o.oPosY >= o.oHomeY + BLOCK then
                cur_obj_change_action(1)
                o.oRenderBehindFloor = 0
            end
        else
            cur_obj_hide()
        end
    else
        o.oRenderBehindFloor = 0
        cur_obj_become_tangible()

        if o.oInteractStatus & INT_STATUS_INTERACTED ~= 0 and o.oInteractStatus & INT_STATUS_ATTACK_MASK ~= ATTACK_FROM_BELOW then
            obj_mark_for_deletion(o)
            return
        end

        if not obj_is_on_screen(o, 4 * BLOCK, 4 * BLOCK) then
            obj_mark_for_deletion(o)
            return
        end
    end

    if o.oPosY < -16 * PX then
        obj_mark_for_deletion(o)
    end

    o.oInteractStatus = 0
end

id_bhvSmbFireFlower = hook_smb_behavior("bhvSmbFireFlower", OBJ_LIST_LEVEL, smb_fire_flower_init, smb_fire_flower_loop)

-- super star
---@param o Object
function smb_super_star_init(o)
    o.oFlags = OBJ_FLAG_SET_FACE_YAW_TO_MOVE_YAW
    o.oGravity = 2
    o.oMoveAngleYaw = 0x4000
    o.oInteractType = INTERACT_POWERUP
    o.oDamageOrCoinValue = 2 -- determines powerup type
    o.hitboxRadius = 6 * PX -- 6 pixels on both sides from center
    o.hitboxHeight = 12 * PX
    o.hitboxDownOffset = -3 * PX
    o.oAnimState = 0
    o.oRenderBehindFloor = 1
    o.oAnimState = frame_based_on_level_theme(0, 4, 4)
    cur_obj_become_intangible()
    cur_obj_set_home_once()
    network_init_object(o, true, {"oHomeX", "oHomeY", "oVelX", "oVelY"})
end

---@param o Object
function smb_super_star_loop(o)
    cur_obj_unhide()

    if o.oTimer % 2 ~= 0 then
        o.oAnimState = frame_based_on_level_theme((o.oAnimState + 1) % 4, 4, 4)
    end

    if o.oAction == 0 then
        -- rise from block
        if o.oTimer > 15 then   
            cur_obj_become_intangible()
            o.oRenderBehindFloor = 1
            o.oPosY = o.oPosY + 4
            if o.oPosY >= o.oHomeY + BLOCK then
                cur_obj_change_action(1)
                o.oRenderBehindFloor = 0
            end
        else
            cur_obj_hide()
        end
    else
        o.oRenderBehindFloor = 0
        o.oVelX = sign(o.oMoveAngleYaw) * 16
        cur_obj_become_tangible()
        local landed, hitWall = enemy_move_and_collide(o)

        if o.oTimer > 10 or o.oVelY <= -TERMINAL_VEL + PX then
            o.oSubAction = 1
        end
        
        if landed and o.oSubAction == 1 then
            o.oVelY = 0x38
        end
        if hitWall then
            o.oMoveAngleYaw = -o.oMoveAngleYaw
        end

        if o.oInteractStatus & INT_STATUS_INTERACTED ~= 0 then
            if o.oInteractStatus & INT_STATUS_ATTACK_MASK ~= ATTACK_FROM_BELOW then
                obj_mark_for_deletion(o)
                return
            else
                o.oVelY = 4 * PX
            end
        end

        if not obj_is_on_screen(o, 4 * BLOCK, 4 * BLOCK) then
            obj_mark_for_deletion(o)
            return
        end
    end

    if o.oPosY < -16 * PX then
        obj_mark_for_deletion(o)
    end

    o.oInteractStatus = 0
end

id_bhvSmbSuperStar = hook_smb_behavior("bhvSmbSuperStar", OBJ_LIST_LEVEL, smb_super_star_init, smb_super_star_loop)

-- looping platforms (1-2, for example)
---@param o Object
function smb_general_platform_init(o)
    o.oFlags = 0
    o.hitboxRadius = 1.5 * BLOCK -- 3 blocks wide
    o.hitboxHeight = BLOCK -- 1 block tall, despite their appearance
    o.hitboxDownOffset = 0.5 * BLOCK
    o.oAnimState = 0
    o.header.gfx.scale.x = 6 -- used to render this with a width of 3 blocks

    if o.oBehParams2ndByte ~= 0 then
        -- change width based on param
        o.header.gfx.scale.x = o.oBehParams2ndByte
        o.hitboxRadius = BLOCK * o.oBehParams2ndByte / 4
        if o.setHome == 0 and o.oBehParams2ndByte % 4 == 0 then
            o.oPosX = o.oPosX - 8 * PX
        end
    elseif do_hard_mode_objs() then
        -- change to width of 2 blocks in hard mode
        o.header.gfx.scale.x = 4
        o.hitboxRadius = BLOCK
        if o.setHome == 0 and obj_has_behavior_id(o, id_bhvSmbPulleyPlatform) == 0 then
            o.oPosX = o.oPosX - 8 * PX
        end
    end
    
    cur_obj_set_home_once()
    network_init_object(o, true, {"oVelX", "oVelY", "oHomeX", "oHomeY"})
end

---@param o Object
function smb_looping_platform_loop(o)
    if o.oAction == 0 then -- main action
        if not obj_is_on_screen(o, 6 * BLOCK, 6 * BLOCK) then
            cur_obj_hide()
            cur_obj_change_action(1)
            return
        end
        cur_obj_unhide()

        if o.oBehParams & 0xFF == 0 then
            -- downward
            o.oVelY = -PX
        else
            -- upward
            o.oVelY = PX
        end
        o.oVelX = 0

        enemy_move(o, false)

        local topHeight = (NES_HEIGHT - 8) * PX
        local bottomHeight = -16 * PX
        -- castle looping lifts loop differently
        if o.oBehParams2ndByte == 3 then
            topHeight = topHeight - 2 * BLOCK
            bottomHeight = bottomHeight - 2 * BLOCK
        end

        if o.oPosY < bottomHeight then
            o.oPosY = topHeight
        elseif o.oPosY > topHeight then
            o.oPosY = bottomHeight
        end

        smb_handle_surface_obj(o)
    else -- off-screen
        cur_obj_hide()
        if obj_is_on_screen(o, 6 * BLOCK, 6 * BLOCK) then
            o.oPosX, o.oPosY = o.oHomeX, o.oHomeY
            cur_obj_change_action(0)
        end
    end
end

id_bhvSmbLoopingPlatform = hook_smb_behavior("bhvSmbLoopingPlatform", OBJ_LIST_SURFACE, smb_general_platform_init, smb_looping_platform_loop)

-- moving platforms that don't loop
---@param o Object
function smb_moving_platform_loop(o)
    if o.oAction == 0 then -- main action
        if not obj_is_on_screen(o, 6 * BLOCK, 6 * BLOCK) then
            cur_obj_hide()
            cur_obj_change_action(1)
            return
        end
        cur_obj_unhide()

        -- similar to flying paratroopas
        if o.oBehParams & 0xFF == 0 then -- left-right
            o.oVelY = 0
            if o.oSubAction == 0 then -- going left
                o.oVelX = o.oVelX - 4 * SSPX
                if o.oVelX < -PX then
                    o.oVelX = -PX
                    o.oSubAction = 1
                end
            elseif o.oSubAction == 1 then -- going right
                o.oVelX = o.oVelX + 4 * SSPX
                if o.oVelX > PX then
                    o.oVelX = PX
                    o.oSubAction = 0
                end
            end
        else -- up-down
            o.oVelX = 0
            if o.oSubAction == 0 then -- going down
                o.oVelY = o.oVelY - 5 * SSPX
                if o.oVelY < -(0x19A * SSPX) then
                    o.oVelY = -(0x19A * SSPX)
                    o.oSubAction = 1
                end
            elseif o.oSubAction == 1 then -- going up
                o.oVelY = o.oVelY + 5 * SSPX
                if o.oVelY > (0x19A * SSPX) then
                    o.oVelY = (0x19A * SSPX)
                    o.oSubAction = 0
                end
            end
        end

        enemy_move(o, false)

        smb_handle_surface_obj(o)
    else -- off-screen
        cur_obj_hide()
        if obj_is_on_screen(o, 6 * BLOCK, 6 * BLOCK) then
            o.oPosX, o.oPosY = o.oHomeX, o.oHomeY
            o.oVelX, o.oVelY = 0, 0
            cur_obj_change_action(0)
        end
    end
end

id_bhvSmbMovingPlatform = hook_smb_behavior("bhvSmbMovingPlatform", OBJ_LIST_SURFACE, smb_general_platform_init, smb_moving_platform_loop)

-- cloud platform in bonus areas
---@param o Object
function smb_cloud_platform_loop(o)
    cur_obj_unhide()

    if o.oAction == 0 then
        -- start moving if anyone steps on the platform
        for i=0,MAX_PLAYERS-1 do
            local m = gMarioStates[i]
            local smb = gSMBStates[i]
            if is_player_active_and_synced(m) ~= 0 and smb.platform == o then
                cur_obj_change_action(1)
                break
            end
        end
    else
        if o.oPosX - o.hitboxRadius > loaded_tile_data.width * BLOCK then
            obj_mark_for_deletion(o)
            return
        end

        o.oVelX = PX
        enemy_move(o, false)

        smb_handle_surface_obj(o)
    end
end

id_bhvSmbCloudPlatform = hook_smb_behavior("bhvSmbCloudPlatform", OBJ_LIST_SURFACE, smb_general_platform_init, smb_cloud_platform_loop)

-- falling platform; falls if mario stands on it
---@param o Object
function smb_falling_platform_loop(o)
    if o.oAction == 0 then -- main action
        if not obj_is_on_screen(o, 6 * BLOCK, 6 * BLOCK) then
            cur_obj_hide()
            cur_obj_change_action(1)
            return
        end
        cur_obj_unhide()

        o.oVelX = 0
        -- return to home position so multiplayer doesn't suck
        if o.oSubAction ~= 0 then
            o.oVelY = 2 * PX
            if o.oPosY >= o.oHomeY then
                o.oPosY = o.oHomeY
                o.oVelY = 0
                o.oSubAction = 0
                o.oTimer = 0
            end
        end
        enemy_move(o, false)

        local count = smb_handle_surface_obj(o)
        if count ~= 0 then
            o.oVelY = -2 * PX
            o.oTimer = 0
            o.oSubAction = 0
        elseif o.oSubAction == 0 then
            o.oVelY = 0
        end

        -- return home after 8 seconds of no one using it
        if o.oSubAction == 0 and o.oTimer > 8 * 60 and o.oPosY ~= o.oHomeY then
            o.oSubAction = 1
        end
    elseif o.oAction == 1 then -- off-screen
        cur_obj_hide()
        if obj_is_on_screen(o, 6 * BLOCK, 6 * BLOCK) then
            o.oPosX, o.oPosY = o.oHomeX, o.oHomeY
            o.oVelX, o.oVelY = 0, 0
            cur_obj_change_action(0)
        end
    end
end

id_bhvSmbFallingPlatform = hook_smb_behavior("bhvSmbFallingPlatform", OBJ_LIST_SURFACE, smb_general_platform_init, smb_falling_platform_loop)

-- pulley platform; affected by mario's weight. If it or its partner gets too high up, both of them fall
---@param o Object
function smb_pulley_platform_loop(o)
    -- find our partner (uses oBehParams)
    local partner = o.usingObj
    if partner == o or partner == nil then
        partner = obj_get_first_with_behavior_id_and_field_s32(id_bhvSmbPulleyPlatform, 0x40, o.oBehParams)
        if partner == o then partner = obj_get_next_with_same_behavior_id_and_field_s32(partner, 0x40, o.oBehParams) end
        -- if no partner found, hide until partner exists
        if not partner then
            cur_obj_hide()
            return
        end
    end

    if o.oAction == 0 then -- main action
        local onScreen = true
        if not obj_is_on_screen(o, 6 * BLOCK, 6 * BLOCK) then
            onScreen = false
            cur_obj_hide()
        else
            cur_obj_unhide()
        end
        
        o.oVelX = 0
        o.oGravity = 0x0A * SSPX
        enemy_move(o, false)
        local tile = get_tile_at_coordinate_pos(o.oPosX, o.oPosY + 7 * PX, true)
        if partner.oAction == 1 or tile == TILE_ROPE_PULLEY_L or tile == TILE_ROPE_PULLEY_R then
            -- fall
            o.oVelY = 0
            cur_obj_change_action(1)
        end
        
        if loaded_tile_data then -- draw rope tiles
            local height = 13
            local blockX = o.oPosX // BLOCK
            local blockY = (o.oPosY + 7 * PX) // BLOCK
            for y=0,height - 1 do
                local tile2 = get_tile_at_block(blockX, y, true)
                -- stop drawing at top of pulley
                if tile2 == TILE_ROPE_PULLEY_L or tile2 == TILE_ROPE_PULLEY_R then break end

                if y < blockY then
                    add_to_2d_table(loaded_tile_data, blockX, y, 0) -- air below our block
                elseif y == blockY and ((o.oPosY + 7 * PX) // (BLOCK // 2)) % 2 ~= 0 then
                    add_to_2d_table(loaded_tile_data, blockX, y, TILE_ROPE_V_HALF) -- half rope
                else
                    add_to_2d_table(loaded_tile_data, blockX, y, TILE_ROPE_V) -- rope above tile
                end
            end
        end

        local count = 0
        if onScreen then
            count = smb_handle_surface_obj(o)
        end
        o.oSMBTimer2 = count -- store here for partner to read

        -- set velocities
        if o.oSMBTimer2 == partner.oSMBTimer2 then
            -- balanced
            o.oVelY = approach_f32(o.oVelY, 0, o.oGravity, o.oGravity)
        elseif o.oSMBTimer2 > partner.oSMBTimer2 then
            -- we have more weight
            o.oVelY = o.oVelY - o.oGravity
            if o.oVelY < -3 * PX then
                o.oVelY = -3 * PX
            end
        elseif o.oSMBTimer2 < partner.oSMBTimer2 then
            -- our partner has more weight
            o.oVelY = o.oVelY + o.oGravity
            if o.oVelY > 3 * PX then
                o.oVelY = 3 * PX
            end
        end
    elseif o.oAction == 1 then -- falling
        cur_obj_unhide()
        o.oVelX = 0
        o.oGravity = 2
        enemy_move(o, true)

        -- award points
        if o.oDeathSound == 0 and o.oTimer <= 10 and gSMBStates[0].platform == o then
            local m = gMarioStates[0]
            award_points(m, 6, true, m.pos.x, m.pos.y + BLOCK)
            o.oDeathSound = 1 -- never award points again
        end

        if obj_is_on_screen(o, 6 * BLOCK, 6 * BLOCK) then
            smb_handle_surface_obj(o)
        end
    
        if o.oPosY < -16 * PX then
            cur_obj_change_action(2)
        end
    elseif o.oAction == 2 then -- at screen bottom
        cur_obj_hide()
        if partner.oAction == 1 then o.oTimer = 0 end

        -- respawn after 8 seconds
        if partner.oAction == 0 or o.oTimer > 8 * 60 then
            o.oPosX, o.oPosY = o.oHomeX, o.oHomeY
            o.oVelY = 0
            cur_obj_change_action(0)
        end
    end
end

id_bhvSmbPulleyPlatform = hook_smb_behavior("bhvSmbPulleyPlatform", OBJ_LIST_SURFACE, smb_general_platform_init, smb_pulley_platform_loop)

-- axe
function smb_axe_init(o)
    o.oFlags = 0
    o.oInteractType = INTERACT_POWERUP
    o.oDamageOrCoinValue = 4
    o.hitboxRadius = 4 * PX
    o.hitboxHeight = BLOCK
    o.hitboxDownOffset = 0
    o.oAnimState = 0
    o.oRenderBehindFloor = 1
    cur_obj_set_home_once()
end

function smb_axe_loop(o)
    local smb = gSMBStates[0]
    local m = gMarioStates[0]

    cur_obj_become_tangible()

    if o.oAction == 0 and obj_is_on_screen(o, 0, 0) then
        smb.tempScrollStop = o.oPosX
    end

    -- animate like a coin
    local timer = get_global_timer() % 24
    if timer < 12 then
        o.oAnimState = 0
    elseif timer >= 16 and timer < 20 then
        o.oAnimState = 2
    else
        o.oAnimState = 1
    end

    if o.oInteractStatus & INT_STATUS_INTERACTED ~= 0 and o.oAction == 0 then
        local bowser = obj_get_first_with_behavior_id(id_bhvSmbBowser)
        if bowser and bowser.oAction < 2 then
            set_mario_action(m, ACT_SMB_HIT_AXE, 0)
            spawn_object_no_rotate(id_bhvSmbBowserCutscene, E_MODEL_NONE, bowser.oPosX, bowser.oPosY, 0, false)
        else
            set_mario_action(m, ACT_SMB_WALK_CUTSCENE, 1)
        end
        smb.tempScrollStop = 0
        if gNetworkPlayers[0].currActNum > 28 then
            local name = network_get_player_text_color_string(0)..gNetworkPlayers[0].name
            djui_popup_create_global(name.."\\#ffff50\\ finished the game!", 2)
        end
        cur_obj_change_action(1)
        obj_mark_for_deletion(o)
    end
    
    o.oInteractStatus = 0
end

id_bhvSmbAxe = hook_smb_behavior("bhvSmbAxe", OBJ_LIST_LEVEL, smb_axe_init, smb_axe_loop)

-- sprite info (texture, width, height, spaceBetweenFrames, flipType, noSync, xOffset, yOffset)
SPRITE_INFO_NO_FLIP = 0
SPRITE_INFO_FLIP = 1
SPRITE_INFO_ROTATE = 2
SPRITE_INFO_1_FRAME = 3
SPRITE_INFO_FIREBALL_SPECIAL = 4
SPRITE_INFO_TILED = 5
SPRITE_INFO_BOWSER_SPECIAL = 6
SPRITE_INFO_VINE_SPECIAL = 7
OBJECT_SPRITE_INFO = {
    -- enemies and other obstacles
    [id_bhvSmbGoomba] = {get_texture_info("goomba"), 16, 16, 0, SPRITE_INFO_NO_FLIP},
    [id_bhvSmbKoopa] = {get_texture_info("koopa"), 16, 32, 1, SPRITE_INFO_FLIP},
    [id_bhvSmbPiranhaPlant] = {get_texture_info("piranhaPlant"), 16, 32, 0, SPRITE_INFO_NO_FLIP, false, 8},
    [id_bhvSmbFlyingParatroopa] = {get_texture_info("koopa"), 16, 32, 1, SPRITE_INFO_FLIP},
    [id_bhvSmbBouncingParatroopa] = {get_texture_info("koopa"), 16, 32, 1, SPRITE_INFO_FLIP},
    [id_bhvSmbBlooper] = {get_texture_info("blooper"), 16, 24, 1, SPRITE_INFO_NO_FLIP},
    [id_bhvSmbPodoboo] = {get_texture_info("podoboo"), 16, 16, 0, SPRITE_INFO_1_FRAME},
    [id_bhvSmbHammerBro] = {get_texture_info("hammerBro"), 16, 32, 1, SPRITE_INFO_FLIP},
    [id_bhvSmbBowser] = {get_texture_info("bowser"), 32, 32, 1, SPRITE_INFO_BOWSER_SPECIAL},
    [id_bhvSmbFireBarBase] = {get_texture_info("fireBall"), 8, 8, 0, SPRITE_INFO_FIREBALL_SPECIAL, false, 0, 5},
    [id_bhvSmbFireBarPart] = {get_texture_info("fireBall"), 8, 8, 0, SPRITE_INFO_FIREBALL_SPECIAL, true},
    [id_bhvSmbSpiny] = {get_texture_info("spiny"), 16, 16, 0, SPRITE_INFO_FLIP},
    [id_bhvSmbLakitu] = {get_texture_info("lakitu"), 16, 32, 0, SPRITE_INFO_FLIP},
    [id_bhvSmbBuzzyBeetle] = {get_texture_info("buzzyBeetle"), 16, 16, 0, SPRITE_INFO_FLIP},

    -- level objects (platforms, axe, etc.)
    [id_bhvSmbLoopingPlatform] = {get_texture_info("platform"), 8, 8, 0, SPRITE_INFO_1_FRAME, false, 0, 8},
    [id_bhvSmbMovingPlatform] = {get_texture_info("platform"), 8, 8, 0, SPRITE_INFO_1_FRAME, false, 0, 8},
    [id_bhvSmbCloudPlatform] = {get_texture_info("cloudPlatform"), 8, 8, 0, SPRITE_INFO_1_FRAME, false, 0, 8},
    [id_bhvSmbFallingPlatform] = {get_texture_info("platform"), 8, 8, 0, SPRITE_INFO_1_FRAME, false, 0, 8},
    [id_bhvSmbPulleyPlatform] = {get_texture_info("platform"), 8, 8, 0, SPRITE_INFO_1_FRAME, false, 0, 8},
    [id_bhvSmbSpringboard] = {get_texture_info("springboard"), 16, 32, 1, SPRITE_INFO_NO_FLIP, true, 0, 1},
    [id_bhvSmbAxe] = {get_texture_info("axe"), 16, 16, 1, SPRITE_INFO_NO_FLIP, true, 0, 1},
    [id_bhvSmbFlag] = {get_texture_info("goalFlag"), 16, 16, 0, SPRITE_INFO_1_FRAME, true},
    [id_bhvSmbCastleFlag] = {get_texture_info("castleFlag"), 16, 16, 0, SPRITE_INFO_1_FRAME, true},
    [id_bhvSmbVine] = {get_texture_info("vine"), 16, 16, 0, SPRITE_INFO_VINE_SPECIAL},

    -- controllers / spawners
    [id_bhvSmbExitController] = {nil, "EX", 0, 0, 0, true}, -- "Ex" is what the text is used to represent this object in the editor
    [id_bhvSmbWarpZoneController] = {nil, "WZ", 0, 0, 0, true},
    [id_bhvSmbLoopController] = {nil, "LO", 0, 0, 0, true},
    [id_bhvSmbBowserFlameFrenzy] = {nil, "BF", 0, 0, 0, true},
    [id_bhvSmbUwFishFrenzy] = {nil, "UF", 0, 0, 0, true},
    [id_bhvSmbFlyingFishFrenzy] = {nil, "FF", 0, 0, 0, true},
    [id_bhvSmbWhirlpool] = {nil, "WP", 0, 0, 0, true},
    [id_bhvSmbBulletBillFrenzy] = {nil, "BB", 0, 0, 0, true},
    [id_bhvSmbStopFrenzy] = {nil, "$F", 0, 0, 0, true},
    [id_bhvSmbBillBlaster] = {nil, "BI", 0, 0, 0, false, 0, 1},

    -- power-ups
    [id_bhvSmbMushroom] = {get_texture_info("mushroom"), 16, 16, 0, SPRITE_INFO_NO_FLIP},
    [id_bhvSmbFireFlower] = {get_texture_info("fireFlower"), 16, 16, 0, SPRITE_INFO_NO_FLIP},
    [id_bhvSmbSuperStar] = {get_texture_info("star"), 16, 16, 0, SPRITE_INFO_NO_FLIP},
    [id_bhvSmbFireball] = {get_texture_info("fireBall"), 8, 8, 0, SPRITE_INFO_FIREBALL_SPECIAL},
    
    -- projectiles
    [id_bhvSmbBowserFlame] = {get_texture_info("bowserFire"), 32, 8, 0, SPRITE_INFO_NO_FLIP},
    [id_bhvSmbHammer] = {get_texture_info("hammer"), 16, 16, 0, SPRITE_INFO_ROTATE},
    [id_bhvSmbUwCheepCheep] = {get_texture_info("cheepCheep"), 16, 16, 0, SPRITE_INFO_FLIP}, -- yes the fish are projectiles
    [id_bhvSmbFlyingCheepCheep] = {get_texture_info("cheepCheep"), 16, 16, 0, SPRITE_INFO_FLIP},
    [id_bhvSmbBulletBill] = {get_texture_info("bulletBill"), 16, 16, 0, SPRITE_INFO_FLIP},

    -- particles / other visual objects
    [id_bhvSmbFirework] = {get_texture_info("fireBall"), 8, 8, 0, SPRITE_INFO_FIREBALL_SPECIAL, true},
    [id_bhvSmbBounceBlock] = {"smb_tileset", 16, 16, 0, SPRITE_INFO_TILED, true},
    [id_bhvSmbCoinParticle] = {get_texture_info("coin"), 16, 16, 0, SPRITE_INFO_NO_FLIP, true},
    [id_bhvSmbPointNumber] = {get_texture_info("points"), 16, 8, 1, SPRITE_INFO_NO_FLIP, true},
    [id_bhvSmbMessenger] = {get_texture_info("messenger"), 16, 32, 0, SPRITE_INFO_NO_FLIP, true},
    [id_bhvSmbBrickParticle] = {get_texture_info("brick"), 8, 8, 0, SPRITE_INFO_ROTATE, true},
    [id_bhvSmbAirBubble] = {get_texture_info("airBubble"), 8, 8, 0, SPRITE_INFO_1_FRAME, true},
    [id_bhvSmbBowserCutscene] = {get_texture_info("bowser"), 32, 32, 1, SPRITE_INFO_FLIP, false},
}

-- object lists that will be iterated through for rendering
VALID_OBJ_LISTS = {OBJ_LIST_UNIMPORTANT, OBJ_LIST_SURFACE, OBJ_LIST_GENACTOR, OBJ_LIST_LEVEL}

-- object lists that will be iterated through for collision with each other
VALID_OBJ_LISTS_COL = {OBJ_LIST_GENACTOR}

-- jank method of getting the pause menu to work
function pause_controller_loop(o)
    o.activeFlags = o.activeFlags | ACTIVE_FLAG_UNIMPORTANT
    if gSMBStates[0].paused and timeStopEnabled then
        before_mario_update(gMarioStates[0], true)
    end
end
id_bhvPauseController = hook_behavior(nil, OBJ_LIST_DEFAULT, false, nil, pause_controller_loop, "bhvPauseController")

-- the graphics for each Fake Bowser. Args: Enemy, Frame, Gravity
FakeBowserData = {
    {id_bhvSmbGoomba, 8, 4},
    {id_bhvSmbKoopa, 4, 4},
    {id_bhvSmbBuzzyBeetle, 10, 4},
    {id_bhvSmbSpiny, 1, 4},
    {id_bhvSmbLakitu, 2, 4},
    {id_bhvSmbBlooper, 1, 1},
    {id_bhvSmbHammerBro, 1, 4},
}